(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 * )

Require Import Base.

Record joinable (t: Type) : Type := Joinable {
   is_empty: t -> Prop;
   join: t -> t -> t -> Prop;
   join_com: forall a b c, join a b c -> join b a c;
   join_empty: forall e a b, is_empty e -> join e a b -> a=b;
   exists_empty: forall a, exists e, is_empty e /\ join e a a;
   join_empty2: forall a b c, join a b c ->  is_empty c -> is_empty a;
   join_empty3: forall e a, join e a a -> is_empty e;
   join_assoc: forall a b c d e, join a b d -> join d c e ->
                    exists f, join b c f /\ join a f e;
   join_eq: forall x y z z', join x y z -> join x y z' -> z = z';
   cancellation: forall a1 a2 b c, join a1 b c -> join a2 b c -> a1=a2
}.

Implicit Arguments is_empty.
Implicit Arguments join.
Implicit Arguments join_com.
Implicit Arguments join_empty.
Implicit Arguments exists_empty.
Implicit Arguments join_empty2.
Implicit Arguments join_empty3.
Implicit Arguments join_assoc.
Implicit Arguments join_eq.
Implicit Arguments cancellation.

Definition joins (A : Type) (j: joinable A) (e1 e2 : A) : Prop :=
  exists e, join j e1 e2 e.
Implicit Arguments joins.

Lemma joins_com: forall A j (e1 e2 : A), 
  joins j e1 e2 ->
  joins j e2 e1.
Proof.
unfold joins.
intros.
destruct H.
exists x.
apply join_com.
trivial.
Qed.
Implicit Arguments joins_com.

Definition sub (A : Type) (j: joinable A) (e1 e : A) : Prop :=
  exists e2, join j e1 e2 e.
Implicit Arguments sub.

Lemma is_empty_join:
 forall (A: Type) (j: joinable A) (e: A), 
     is_empty j e -> join j e e e.
Proof.
intros.
destruct (exists_empty j e) as [? [? ?]].
destruct (join_assoc j e x x e e) as [f [? ?]].
apply join_com; auto.
apply join_com; auto.
generalize (cancellation _ _ _ _ _ H1 (join_com _ _ _ _ H3)); intro.
subst.
generalize (join_empty _ _ _ _ H H3); intro.
subst. auto.
Qed.

Record joinmap (key: Type) (t: Type) (j : joinable t) : Type 
 := Joinmap {
    jm_t : Type;
    jm_j : joinable jm_t;
    lookup: jm_t -> key -> t;
    prim : forall (f: key -> t) (e: t),
                  (forall k, j.(join) e (f k) (f k)) -> 
                  jm_t;
    join_rule: forall s1 s2 s,
                jm_j.(join) s1 s2 s <-> 
                forall x, j.(join) (lookup s1 x) (lookup s2 x) (lookup s x);
    empty_rule: forall e x, jm_j.(is_empty) e -> j.(is_empty) (lookup e x);
    prim_rule: forall f e pf k, lookup (prim f e pf) k = f k;
    ext: forall s1 s2, 
        (forall x, lookup s1 x = lookup s2 x) <-> s1 = s2;
    can_join: forall s1 s2,
            (forall x, exists v, 
                j.(join) (lookup s1 x) (lookup s2 x) v) ->
            exists s3, jm_j.(join) s1 s2 s3;
    can_split: forall s1 s3,
            (forall x, exists v, 
                j.(join) (lookup s1 x) v (lookup s3 x)) ->
            exists s2, jm_j.(join) s1 s2 s3
}. 

Module Type JOINMAP.
Parameter mkJoinmap : 
  forall (key: Type) (t: Type) (j: joinable t), joinmap key t j.
End JOINMAP.

Module Joinmap : JOINMAP.
(* We do it with a Module to preserve abstraction, so the client
 * doesn't know the representation of jm_j.
 *)

Section JJ.
Variable key: Type.
Variable t' : Type.
Variable range: joinable t'.

Definition valid (f: key -> t') : Prop := 
  exists e, forall k, range.(join) e (f k) (f k).

Inductive lis : Type :=
 make: forall (f: key -> t'), valid f -> lis.

Definition fn (a: lis) : key -> t' := 
 match a with make f _ => f end.


Lemma prim_valid:
  forall (f: key -> t') (e: t') (pf: forall k, range.(join) e (f k) (f k)),
     valid f.
Proof.
intros.
unfold valid.
exists e.
exact pf.
Qed.

Definition prim (f: key -> t') (e: t')
                (pf: forall k, range.(join) e (f k) (f k)) : lis
  := make f (prim_valid f e pf).

Definition lookup (l : lis) (x : key) : t' := fn l x.

Definition m_is_empty (a: lis) : Prop :=  forall x, range.(is_empty) (fn a x).

Notation "a '@' b '=R' c" := (range.(join) a b c) (at level 70, b at next level).

Definition m_join (a b c : lis) : Prop :=
  forall x, range.(join) (fn a x) (fn b x) (fn c x).

Notation "a '@' b '=J' c" := (m_join a b c) (at level 70, b at next level).

Lemma empty_rule: forall e x, m_is_empty e -> range.(is_empty) (lookup e x).
Proof.
intros.
generalize (H x).
simpl; auto.
Qed.

Lemma m_join_com: forall a b c, m_join a b c -> m_join b a c.
Proof.
unfold m_join; intros.
apply range.(join_com).
auto.
Qed.

Lemma m_join_empty: forall e a b, m_is_empty e -> m_join e a b -> a=b.
Proof.
unfold m_join, m_is_empty; intros.
destruct a; destruct b.
assert(f=f0).
extensionality x.
spec H x. spec H0 x.
exact (range.(join_empty) _ _ _ H H0).
subst f0.
rewrite (proof_irr v v0); auto.
Qed.

Lemma m_exists_empty: forall a, exists e, m_is_empty e /\ m_join e a a.
Proof.
destruct a.
unfold valid in v.
generalize v; intro v'.
destruct v as [e1 ?].
assert (valid (fun _ => e1)).
unfold valid.
exists e1.
intro.
generalize (range.(join_empty3) _ _ (H k)); intro.
destruct (range.(exists_empty) e1) as [e2 [? ?]].
generalize (range.(join_empty) _ _ _ H0 (range.(join_com) _ _ _ H2)); intro.
subst e2.
auto.
exists (make _ H0).
split.
unfold is_empty.
intro.
simpl.
spec H x.
apply range.(join_empty3) with (f x); auto.
intro.
simpl.
apply H.
Qed.

Lemma ext:
   forall s1 s2, (forall x, lookup s1 x = lookup s2 x) <-> s1 = s2.
Proof.
unfold lookup; destruct s1; destruct s2; intros; split; auto; intro.
assert (f=f0).
extensionality x; auto.
subst f.
rewrite (proof_irr v v0); auto.
rewrite <- H.
intros; auto.
Qed.

Lemma join_rule: 
  forall a b c,
   a @ b =J c  <->  forall x, lookup a x @ lookup b x =R lookup c x.
Proof.
unfold join; intros; split.
auto.
auto.
Qed.

Lemma m_join_eq: forall x y z z', 
     m_join x y z -> m_join x y z' -> z = z'.
Proof.
destruct x; destruct y; destruct z; destruct z'.
unfold join.
intros.
assert (f1=f2).
extensionality i.
spec H i.
spec H0 i.
simpl in H.
simpl in H0.
eapply range.(join_eq).
apply H.
auto.
subst.
rewrite (proof_irr v1 v2); auto.
Qed.

Lemma m_cancellation: forall a1 a2 b c, 
             m_join a1 b c -> m_join a2 b c -> a1=a2.
Proof.
intros.
destruct a1; destruct a2.
assert (f=f0).
2: subst; rewrite (proof_irr v v0); auto.
extensionality x.
spec H x.
spec H0 x.
simpl in H, H0.
apply (range.(cancellation) _ _ _ _ H H0).
Qed.

Lemma prim_rule: forall (f: key -> t') (e: t') 
                (pf: forall k, range.(join) e (f k) (f k))
                (k: key),
                 lookup (prim f e pf) k = f k.
Proof.
intros.
unfold prim.
simpl.
auto.
Qed.

Lemma m_join_empty2: forall a b c, 
             m_join a b c ->  m_is_empty c -> m_is_empty a.
Proof.
intros.
intro i.
spec H0 i.
spec H i.
eapply range.(join_empty2).
apply H.
auto.
Qed.

Lemma m_join_empty3: forall e a, m_join e a a -> m_is_empty e.
Proof.
intros.
intro i.
spec H i.
eapply range.(join_empty3).
apply H.
Qed.

Lemma m_join_assoc:
  forall a b c d e, m_join a b d -> m_join d c e ->
                    exists f, m_join b c f /\ m_join a f e.
Proof.
destruct a as [a va];
destruct b as [b vb];
destruct c as [c vc]; 
destruct d as [d vd]; 
destruct e as [e ve].
unfold join; intros.
assert (forall x : key, exists y : t', b x @ c x =R y /\ a x @ y =R e x).
intro.
apply (range.(join_assoc) (a x) (b x) (c x) (d x) (e x) (H x) (H0 x)).
generalize (choice (fun y z => (b y @ c y =R z /\ a y @ z =R e y)) H1); intro.
destruct H2 as [f H2].
assert (valid f).
unfold valid.
generalize vb; intro G.
destruct G as [emp ?].
exists emp.
intro.
spec H3 k.
spec H2 k.
destruct H2.
destruct (range.(join_assoc) _ _ _ _ _ H3 H2) as [f' [? ?]].
generalize (range.(join_eq) _ _ _ _ H5 H2); intro.
subst f'. auto.
simpl.
exists (make f H3).
split; intro; simpl; spec H2 x; destruct H2; auto.
Qed.

Lemma can_join: 
  forall a b,
         (forall x, exists v, range.(join) (lookup a x) (lookup b x) v) ->
        exists c, m_join a b c.
Proof.
destruct a as [a va]; destruct b as [b vb].
intros.
unfold m_join.
simpl.
assert (exists c, forall x, a x @ b x =R c x).
apply (choice (fun x y => a x @ b x =R y)).
simpl in H.
auto.
destruct H0 as [c ?].
assert (valid c).
unfold valid.
generalize va; intro G; destruct G as [e ?].
exists e.
intro.
spec H1 k; spec H0 k.
destruct (range.(join_assoc) _ _ _ _ _ H1 H0) as [f' [? ?]].
generalize (range.(join_eq) _ _ _ _ H2 H0); intro.
subst f'.
auto.
exists (make c H1).
intro. simpl.
auto.
Qed.

Lemma can_split:
  forall a c,
         (forall x, exists v, range.(join) (lookup a x) v (lookup c x)) ->
        exists b, m_join a b c.
Proof.
intros.
destruct a as [a va]; destruct c as [c vc].
intros.
unfold join.
simpl; simpl in H.
assert (exists b, forall x, a x @ b x =R c x).
apply (choice (fun x y => a x @ y =R c x)); auto.
destruct H0 as [b ?].
assert (valid b).
unfold valid.
destruct (classic (exists x: key, True)).
destruct H1 as [x _].
generalize vc; intro G; destruct G as [emp ?].
exists emp.
intro.
spec H1 k; spec H0 k.
destruct (range.(join_assoc) _ _ _ _ _ H0 (range.(join_com) _ _ _ H1)) as [f' [? ?]].
generalize (range.(join_empty3) _ _ H1); intro.
generalize (range.(join_empty) _ _ _ H4 (range.(join_com) _ _ _ H2)); intro.
subst f'.
apply range.(join_com); auto.
destruct va as [e ?].
exists e.
intro.
contradiction H1.
exists k.
auto.
exists (make b H1).
intros.
simpl.
auto.
Qed.

Definition mkJoinmap  : joinmap key t' range :=
  Joinmap key t' range lis
   (Joinable lis m_is_empty m_join m_join_com m_join_empty m_exists_empty
        m_join_empty2 m_join_empty3 m_join_assoc m_join_eq m_cancellation )
   lookup prim join_rule empty_rule prim_rule ext can_join can_split.

End JJ.
End Joinmap.

Definition mkJoinmap := Joinmap.mkJoinmap.

Implicit Arguments jm_t.
Implicit Arguments jm_j.
Implicit Arguments lookup.
Implicit Arguments prim.
Implicit Arguments join_rule.
Implicit Arguments empty_rule.
Implicit Arguments prim_rule.
Implicit Arguments ext.
Implicit Arguments can_join.
Implicit Arguments can_split.
