(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import Setoid.
Require Import Base.
Require Import List.

Require Import Maps.
Require Import Machine.
Require Import ModelSpec.
Require Import LMModel.
Require Import Listmachine.

Module Type LM_HOARE_RULES.
  Declare Module LM : LIST_MACHINE.
  Export LM.

  Open Local Scope pred_scope.
  
  Axiom with_step_intro : forall p k k' s s' q G,
    Machine.step p s k s' k' ->
    [ G ] |-- projecting Sel_S s ->
    [ G ] |-- with_subst (U Sel_S s'::nil) q ->
    [ G ] |-- with_step p k k' q.

  Section HoareRules.
    Variable p : program.
    Variable pt : map pred.

    Axiom check_block_weaken : forall e e' i,
      |> (pgm_typ pt p) |-- e >=> e' ->
      check_block p pt e' i ->
      check_block p pt e  i.

    Axiom check_instr_weaken : forall P P' Q Q' i,
      |> (pgm_typ pt p) |-- P  >=> P' ->
      |> (pgm_typ pt p) |-- Q' >=> Q ->

      check_instr p pt P' i Q' ->
      check_instr p pt P  i Q.


    Axiom check_block_halt: forall e,
   (*----------------------------------------------------*)
      check_block p pt e instr_halt.

    Axiom check_block_seq: forall e i1 i2 e',
      check_instr p pt e  i1 e' ->
      check_block p pt e' i2 ->
   (* ----------------------------------------------------*)
      check_block p pt e (i1 ;; i2).

    Axiom check_block_jump : forall v t,
      check_block p pt (t && (t_slot v (t_cont p t))) (instr_jump v).

    Axiom check_instr_seq : forall e e' e'' i1 i2,

      check_instr p pt e  i1 e'  ->
      check_instr p pt e' i2 e'' ->
   (* -------------------------------------------------- *)
      check_instr p pt e (i1;;i2) e''.

    Axiom check_instr_getlabel : forall l e v t t',
      pt#l = Some t' ->
      |-- t >=> t' ->
      not_in_dom v ([]e) ->
   (* -------------------------------------------------- *)
      check_instr p pt e (instr_getlabel l v) (e && (t_slot v (t_cont p t))).

    Axiom check_instr_getnil : forall e v,
      not_in_dom v ([]e) ->
   (* -------------------------------------------------- *)
      check_instr p pt e (instr_getlabel (L 0) v) (e && (t_slot v t_nil)).

    Axiom check_instr_branch_nil : forall x e e' v l,
      pt#l = Some x ->
      |-- e >=> x ->
   (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v t_nil))
                       (instr_branch_if_nil v l)
                       e'.

   Axiom check_instr_branch_pair: forall e v l,
    (*----------------------------------------------------*)
      check_instr p pt
                     (e && (t_slot v t_anypair))
                     (instr_branch_if_nil v l)
                     e.

   Axiom check_instr_branch: forall e x v l,
      pt#l = Some x ->
      |-- (e && t_slot v t_nil) >=> x ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v (t_nil || t_anypair)))
                       (instr_branch_if_nil v l)
                       (e && (t_slot v t_anypair)).


   Axiom check_instr_fetch_0 : forall e v1 v2 t0 t1,
      not_in_dom v2 e ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v1 (t_pair t0 t1)))
                       (instr_fetch_field v1 0 v2)
                       (e && (t_slot v2 t0)).

   Axiom check_instr_fetch_1 : forall e v1 v2 t0 t1,
      not_in_dom v2 e ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v1 (t_pair t0 t1)))
                       (instr_fetch_field v1 1 v2)
                       (e && (t_slot v2 t1)).

   Axiom check_instr_cons: forall e v1 v2 v3 t0 t1,
      not_in_dom v3 e ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v1 ([](valuepred t0))) && (t_slot v2 ([](valuepred t1))))
                       (instr_cons v1 v2 v3)
                       (e && (t_slot v3 (t_pair t0 t1))).
  End HoareRules.
End LM_HOARE_RULES.

Module LMHoareRules (LM':LIST_MACHINE) : LM_HOARE_RULES.
  Module LM := LM'.
  Import LM.

  Open Local Scope pred_scope.

  Lemma with_step_intro : forall p k k' s s' q G,
    Machine.step p s k s' k' ->
    [ G ] |-- projecting Sel_S s ->
    [ G ] |-- with_subst (U Sel_S s'::nil) q ->
    [ G ] |-- with_step p k k' q.
      
    intros.
    cut_with (with_subst (U Sel_S s' :: nil) q); auto.
    cut_with (projecting Sel_S s); auto.
    clear H0 H1.
    red; simpl; intuition.
    rewrite with_step_spec.
    exists (world_update (U Sel_S s'::nil) w).
    simpl; intuition.
    destruct w; compute; intuition.
    compute in H1.
    subst s; auto.
    rewrite with_subst_spec in H0.
    compute in H0.
    destruct w; simpl in *.
    apply H0; auto.
   Qed.

  Ltac prove_with_step C :=
     let S := fresh "s" in
     apply (with_projecting Sel_S); intro S;
     refine (with_step_intro _ _ _ _ _ _ _ _ _ _);
       [ apply C | exact_at 0 | ].

  Lemma with_subst_shift : forall x (a:field_type x) p,
     projecting x a |-- p ->
     |-- with_subst (U x a :: nil) p.

     intros x a z.
     repeat rewrite with_subst_spec.
     unfold appPosMode, with_subst_rel.
     unfold derives; simpl; intuition.
     apply H.
     subst w'.
     unfold projecting; intuition.
     apply proj_upd_same.
    Qed.

  Lemma with_subst_discard : forall x (a:field_type x) z G p,
     [ G ] |-- projecting x a ->
     [ G ] |-- with_subst (filter_subst (x::nil) z) p ->
     [ G ] |-- with_subst (U x a :: z) p.
     
     intros.
     cut_with (with_subst (filter_subst (x::nil) z) p); auto.
     cut_with (projecting x a); auto.
     clear H H0.
     rewrite with_subst_spec.
     rewrite with_subst_spec.
     unfold appPosMode, with_subst_rel.
     red; simpl; intuition.
     apply H.
     rewrite <- H1.
     red in H0.
     subst a.
     clear; induction z; simpl.
     rewrite upd_proj_same; auto.
     destruct a.
     destruct (field_eq_dec x x0).
     subst x0.
     rewrite IHz.
     rewrite upd_collapse.
     auto.
     simpl.
     rewrite IHz.
     rewrite upd_commute; auto.
    Qed.

  Section HoareRules.
    Variable p : program.
    Variable pt : map pred.

    Lemma lpt_nelpt : 
       |> (pgm_typ pt p) |-- []!|>(pgm_typ pt p).

       rewrite every_later_commute.
       rewrite nec_later_later.
       apply derives_modal.
       apply prove_adj with ubiquitous.
       unfold pgm_typ.
       apply_d adjective_forall; intuition_d.
       apply_d adjective_forall; intuition_d.
       unfold t_labelblock.
       apply_d adjective_impl_const.
       apply_d adjective_exist; intuition_d.
       apply_d adjective_conj.
       apply_d adjective_const.
       unfold labelguard.
       apply_d adjective_forall; intuition_d.
       unfold t_labelblock.
       apply_d adjective_impl_const.
       unfold guards.
       unfold sub.
       rewrite <- every_nec_commute.
       apply adjective_modal; auto.
      Qed.

    Lemma guards_neguards : forall p i e,
      guards p i e |-- []!(guards p i e).

       intros; unfold guards, sub.
       rewrite every_nec_commute.
       rewrite modal_idem; auto.
      Qed.

    Lemma guards_later : forall p i e,
      guards p i e |-- |> (guards p i e).

      unfold guards, sub; intros.
      rewrite later_nec_later; apply nec_later.
     Qed.

    Hint Resolve lpt_nelpt guards_neguards guards_later.

    Lemma check_block_weaken : forall e e' i,
      |> (pgm_typ pt p) |-- e >=> e' ->
      check_block p pt e' i ->
      check_block p pt e  i.

      unfold check_block, guards; intros.
      apply_d sub_trans; [ | apply_d H0; auto ].
      apply_d sub_nec.
      apply H.
     Qed.

    Lemma check_instr_weaken : forall P P' Q Q' i,
      |> (pgm_typ pt p) |-- P  >=> P' ->
      |> (pgm_typ pt p) |-- Q' >=> Q ->

      check_instr p pt P' i Q' ->
      check_instr p pt P  i Q.

      unfold check_instr, guards; intros.
      cut_using (H1 H2); auto.
      intuition_d.
      apply_d sub_trans.
      apply_d sub_nec.
      apply_d H; auto.
      spec_at 1 x.
      apply_at 0.
      apply_d sub_trans.
      apply_d sub_nec.
      apply_d H0; auto.
      auto.
     Qed.

    Lemma check_block_halt: forall e,
   (*----------------------------------------------------*)
      check_block p pt e instr_halt.

      unfold check_block; intros.
      unfold guards, sub.
      repeat apply_d modal_N.
      intro_d; clear_at 0.
      red; simpl; intuition.
      unfold safe, safeN.
      simpl; intros.
      inversion H2; clear H2; subst.
      red; left.
      compute; auto.
      compute in H3; intuition.
      inversion H2.
     Qed.

    Lemma check_block_seq: forall e i1 i2 e',
      check_instr p pt e  i1 e' ->
      check_block p pt e' i2 ->
   (* ----------------------------------------------------*)
      check_block p pt e (i1 ;; i2).

      unfold check_instr, check_block; intros.
      cut_using (H H1); auto.
      spec_at 0 i2.
      apply_at 0.
      apply_d H0; auto.
     Qed.

    Lemma check_block_jump : forall v t,
      check_block p pt (t && (t_slot v (t_cont p t))) (instr_jump v).

      unfold check_block; intros.
      apply step_guard.
      simpl; intuition; subst; auto.
      rewrite pmodal_conj; auto; intuition_d.
      apply (with_projecting Sel_S); intro s.
      elim_at 2.
      unfold t_slot.
      apply (field_pred_elim _ s); auto.
      case_eq (s#v); intros; [ | apply bot_elim ].
      unfold t_cont.
      rewrite <- nmodal_exist; auto.
      elim_at 0; intro l.
      rewrite <- nmodal_exist; auto.
      elim_at 0; intro i.
      repeat rewrite pmodal_conj; auto.
      intuition_d.
      unfold t_labelblock.
      rewrite bimodal_const.
      elim_at 0; intros.
      unfold just.     
      rewrite with_subst_projecting.
      elim_at 0; intros; subst v0.
      cut_with (|>([]t >=> safe p i)).
      clear_at 1.
      exist_d ([]t).
      exist_d i.
      split_d.
      coalesce_modal.
      apply with_step_intro with s s; auto.
      apply step_jump with l; auto.
      apply with_subst_discard; auto.
      simpl; rewrite with_subst_nil.
      rewrite later_nec_later; apply_d nec_later; auto.
      apply derives_weaken_head.
      rewrite with_subst_later_commute.
      apply derives_modal.
      unfold sub.
      rewrite <- every_nec_commute.
      rewrite every_eats_with_subst.
      auto.
     Qed.

    Lemma check_instr_seq : forall e e' e'' i1 i2,

      check_instr p pt e  i1 e'  ->
      check_instr p pt e' i2 e'' ->
   (* -------------------------------------------------- *)
      check_instr p pt e (i1;;i2) e''.

      unfold check_instr; intros.
      intuition_d.
      cut_with (|> guards p (i1 ;; i2 ;; x) ([]e) --> guards p ((i1;;i2);;x) ([]e)).
      apply_at 0; clear_at 0.
      cut_with (guards p (i1;;i2;;x) ([]e)).
      unfold guards, sub.
      rewrite later_nec_later.
      apply_d nec_later; auto.
      cut_using (H H1); auto.
      apply_at 0; clear_at 0.
      cut_using (H0 H1); auto.
      apply_at 0; clear_at 0.
      auto.

      clear_at 0.
      intro_d.
      apply step_guard;
       [ simpl; intuition; subst t; auto
       |].
      rewrite every_later_commute.
      rewrite nec_later_commute.
      apply derives_modal; auto.
      apply exist_intro with ([]e).
      apply exist_intro with (i1;;i2;;x).
      intuition_d.
      rewrite later_nec_later.
      cut_with (|> e); [ | apply_d nec_later; auto ].
      apply derives_weaken_head.
      prove_with_step step_seq.
      apply with_subst_discard; auto.
      simpl.
      rewrite with_subst_nil; auto.
     Qed.

    Lemma check_instr_getnil : forall e v,
      not_in_dom v ([]e) ->
   (* -------------------------------------------------- *)
      check_instr p pt e (instr_getlabel (L 0) v) (e && (t_slot v t_nil)).

      unfold check_instr; intros; intuition_d.
      apply step_guard.
      simpl; intuition; subst; auto.
      exist_d ([](e && t_slot v t_nil)).
      exist_d x.
      intuition_d.
      apply_d guards_later; auto.
      prove_with_step step_getlabel.
      clear_at 2.    
      rewrite later_nec_later.
      in_goal rewrite pmodal_conj; auto.
      in_goal rewrite pmodal_conj; auto.
      split_d.
      red in H.
      unfold with_var_upd in H.
      cut_using (H (value_label (L 0))).
      apply (field_pred_elim _ s); auto.
      simpl.
      coalesce_modal.
      auto.
      apply derives_axiom.
      unfold t_slot.
      rewrite with_subst_later_commute.
      apply modal_N; auto.
      rewrite with_subst_field_pred.
      rewrite get_set_same.
      apply modal_N; auto.
      unfold t_nil, just.
      rewrite with_subst_projecting.
      apply const_intro; auto.
     Qed.

    Lemma check_instr_getlabel : forall l e v t t',
      pt#l = Some t' ->
      |-- t >=> t' ->
      not_in_dom v ([]e) ->
   (* -------------------------------------------------- *)
      check_instr p pt e (instr_getlabel l v) (e && (t_slot v (t_cont p t))).

      unfold check_instr; intros; intuition_d.
      apply step_guard.
      simpl; intuition; subst; auto.
      exist_d ([](e && t_slot v (t_cont p t))).      
      exist_d x.
      intuition_d.
      apply_d guards_later; auto.
      prove_with_step step_getlabel.
      clear_at 2.
      rewrite later_nec_later.
      in_goal rewrite pmodal_conj; auto.
      in_goal rewrite pmodal_conj; auto.
      split_d.
      red in H1.
      unfold with_var_upd in H1.
      cut_using (H1 (value_label l)); auto.
      apply (field_pred_elim _ s); auto.
      coalesce_modal.
      move_at 2; apply derives_weaken_head.
      rewrite with_subst_later_commute.
      apply derives_modal.
      clear x.
      unfold pgm_typ.
      spec_at 0 l.
      destruct (H2 l).
      exists t'; auto.
      spec_at 0 x.
      elim_at 0.
      unfold t_labelblock.
      apply const_intro; auto.
      elim_at 0; intro q.
      intuition_d.
      unfold labelguard.
      spec_at 0 i.
      elim_at 0.
      unfold t_labelblock.
      apply const_intro; auto.
      replace q with t'; [ | congruence ].
      unfold t_slot.
      rewrite with_subst_field_pred.
      rewrite get_set_same.
      cut_with (with_subst (U Sel_S s#v <- (value_label l) :: nil) (guards p x ([]t))).
      apply derives_weaken_head.
      apply derives_modal.
      unfold t_cont.
      apply_d modal_distrib_exist.
      exist_d l.
      apply_d modal_distrib_exist.
      exist_d x.
      repeat rewrite pmodal_conj; auto.
      intuition_d.
(****)
      split; auto.
(****)      
      unfold t_labelblock.
      rewrite bimodal_const.
      apply const_intro; auto.
      unfold just.
      rewrite with_subst_projecting.
      apply const_intro; auto.
      rewrite with_subst_later_commute.
      rewrite sub_eats_with_subst.
      apply_d nec_later.
      apply pdrop_modal; auto.
      simpl; intuition; subst; auto.

      apply_d every_with_subst.
      unfold guards, sub.
      rewrite every_nec_commute.
      rewrite modal_idem; auto.
      cut_using H0.
      unfold sub.
      rewrite <- modal_idem; auto.
      rewrite <- every_nec_commute.
      do 2 coalesce_modal.
      intro_d.
      apply_at 2.
      coalesce_modal.
     Qed.


   Lemma check_instr_branch_nil : forall x e e' v l,
      pt#l = Some x ->
      |-- e >=> x ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v t_nil))
                       (instr_branch_if_nil v l)
                       e'.

      unfold check_instr; intros.
      intro_d; intro i'; intro_d.
      apply step_guard.
      simpl; intuition; subst; auto.
      rewrite pmodal_conj; auto; elim_at 0.
      exist_d ([]e).
      destruct (H1 l).
      exists x; auto.
      exist_d x0.
      split_d.
      coalesce_modal.      
      unfold guards, pgm_typ.
      spec_at 0 l.
      spec_at 0 x0.
      elim_at 0.
      unfold t_labelblock; apply const_intro; auto.
      elim_at 0; intro q.
      intuition_d.
      replace q with x; [ | congruence ].
      unfold labelguard.
      spec_at 0 x0.
      elim_at 0.
      unfold t_labelblock; apply const_intro; auto.
      unfold guards.
      apply_d sub_trans; [ | exact_at 0 ].
      apply_d sub_nec; auto.

      apply (with_projecting Sel_S); intro s.
      elim_at 2.
      unfold t_slot.
      apply (field_pred_elim _ s); auto.
      case_eq (s#v); intros; auto.
      unfold t_nil, just.
      rewrite with_subst_projecting.
      elim_at 0; intros; subst v0.
      apply with_step_intro with s s; auto.
      apply step_branch_taken; auto.
      apply with_subst_discard; simpl; auto.
      rewrite with_subst_nil.
      rewrite later_nec_later; apply_d nec_later; auto.
     Qed.      

   Lemma check_instr_branch_pair: forall e v l,
    (*----------------------------------------------------*)
      check_instr p pt
                     (e && (t_slot v t_anypair))
                     (instr_branch_if_nil v l)
                     e.
      
      unfold check_instr; intros.
      intuition_d.
      apply step_guard.
      simpl; intuition; subst; auto.
      exist_d ([]e).
      exist_d x.
      split_d.
      apply_d nec_later.
      unfold guards, sub.
      rewrite modal_idem; auto.
      apply (with_projecting Sel_S); intro s.
      rewrite pmodal_conj; auto.
      elim_at 1.
      elim_at 1.
      unfold t_slot.
      apply (field_pred_elim _ s); auto.
      case_eq (s#v); intros; simpl; auto.
      unfold t_anypair.
      rewrite <- nmodal_exist; auto.
      intuition_d.
      rewrite <- nmodal_exist; auto.
      intuition_d.
      unfold t_pair.
      rewrite with_subst_field_pred.
      destruct v0.
      change bot with (const False).
      rewrite bimodal_const; auto.
      apply with_step_intro with s s; auto.
      apply step_branch_not_taken with v0_1 v0_2; auto.
      apply with_subst_discard; simpl; auto.
      rewrite with_subst_nil.
      apply_d nec_later; auto.
      rewrite modal_idem; auto.
     Qed.

   Lemma check_instr_branch: forall e x v l,
      pt#l = Some x ->
      |-- (e && t_slot v t_nil) >=> x ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v (t_nil || t_anypair)))
                       (instr_branch_if_nil v l)
                       (e && (t_slot v t_anypair)).

      unfold check_instr; intros.
      intuition_d.
      apply step_guard.      
      simpl; intuition; subst; auto.
      rewrite pmodal_conj; auto.
      elim_at 0.
      apply (with_projecting Sel_S); intro s.
      unfold t_slot at 1.
      elim_at 2.
      apply (field_pred_elim _ s); auto.
      case_eq (s#v); intros; auto.
      rewrite nmodal_disj; auto.
      elim_at 0.

      (* nil case *)
      unfold t_nil, just.
      rewrite with_subst_projecting.
      elim_at 0; intros.
      subst v0.
      exist_d ([](e && t_slot v t_nil)).
      destruct (H1 l).
      exists x; auto.
      exist_d x1.
      split_d.
      move_at 3.
      apply derives_weaken_head.
      apply derives_modal.
      unfold pgm_typ, guards.
      spec_at 0 l.
      spec_at 0 x1.
      elim_at 0.
      unfold t_labelblock; apply const_intro; auto.
      intuition_d.
      unfold labelguard.
      spec_at 0 x1.
      elim_at 0.
      unfold t_labelblock; apply const_intro; auto.
      unfold guards.
      replace x2 with x; [ | congruence ].
      apply_d sub_trans; [ | exact_at 0 ].
      apply_d sub_nec; auto.

      apply with_step_intro with s s; auto.
      apply step_branch_taken; auto.
      apply with_subst_discard; simpl; auto.
      rewrite with_subst_nil.
      rewrite later_nec_later.
      in_goal rewrite pmodal_conj; auto.
      split_d.
      apply_d nec_later; auto.
      apply_d nec_later.
      apply (remove_adj_goal _ _ necessary).
      apply_d slot_necessary.
      apply_d nil_necessary.      
      in_goal unfold t_slot.
      apply (field_pred_intro _ s); auto.
      rewrite H2.
      unfold t_nil, just.
      rewrite with_subst_projecting.
      apply const_intro; auto.
      
      (* cons case *)
      unfold t_anypair at 1.
      rewrite <- nmodal_exist; auto; intuition_d.
      rewrite <- nmodal_exist; auto; intuition_d.
      unfold t_pair at 1.
      rewrite with_subst_field_pred.
      destruct v0.
      change bot with (const False).
      rewrite bimodal_const; auto.
      clear_at 0.
      exist_d ([](e && t_slot v t_anypair)).
      exist_d x0.
      split_d.
      apply_d nec_later.
      unfold guards, sub.
      rewrite modal_idem; auto.
      apply with_step_intro with s s; auto.
      apply step_branch_not_taken with v0_1 v0_2; auto.
      rewrite later_nec_later.
      apply with_subst_discard; auto.
      simpl; rewrite with_subst_nil.
      in_goal rewrite pmodal_conj; auto.
      split_d.
      apply_d nec_later; auto.
      apply derives_weaken_head.
      apply_d nec_later.
      apply pdrop_modal; auto.
      simpl; intuition; subst; auto.
      apply prove_adj with necessary.
      apply projecting_necessary.
      unfold t_slot.
      apply (field_pred_intro _ s); auto.
      rewrite H2.
      unfold t_anypair.
      rewrite <- nmodal_exist; auto.
      exist_d top.
      rewrite <- nmodal_exist; auto.
      exist_d top.
      unfold t_pair.
      rewrite with_subst_field_pred.
      apply derives_axiom.
      apply modal_N; auto.
      split_d; apply modal_N; auto.
     Qed.


    Lemma check_instr_fetch_0 : forall e v1 v2 t0 t1,
      not_in_dom v2 e ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v1 (t_pair t0 t1)))
                       (instr_fetch_field v1 0 v2)
                       (e && (t_slot v2 t0)).

      unfold check_instr; intros.
      intuition_d.
      apply step_guard.
      simpl; intuition; subst; auto.      
      exist_d ([](e && (t_slot v2 t0))).
      exist_d x.
      split_d.
      apply_d nec_later.      
      unfold guards, sub.
      rewrite modal_idem; auto.
      apply (with_projecting Sel_S); intro s.
      rewrite pmodal_conj; auto.
      elim_at 1.
      elim_at 1.
      unfold t_slot at 1.
      apply (field_pred_elim _ s); auto.
      case_eq (s#v1); intros; auto.
      unfold t_pair.      
      rewrite with_subst_field_pred.
      destruct v.
      change bot with (const False).
      rewrite bimodal_const; auto.
      apply with_step_intro with s (s#v2 <- v3); auto.
      apply step_fetch_field_0 with v4; auto.
      rewrite later_nec_later.
      repeat rewrite pmodal_conj; auto.
      elim_at 0.
      clear_at 1.
      split_d.
      in_goal rewrite with_subst_later_commute.
      apply_d nec_later.
      move_at 1.
      cut_with ([](projecting Sel_S s)).
      do 4 clear_at 2.
      coalesce_modal.      
      cut_using (H v3); auto.
      unfold with_var_upd.
      apply (field_pred_elim _ s); auto.
      move_at 2.
      apply derives_weaken_head.
      apply prove_adj with necessary.
      apply projecting_necessary.
      in_goal rewrite with_subst_later_commute.
      in_goal unfold t_slot.
      rewrite with_subst_field_pred.
      rewrite get_set_same.
      rewrite with_subst_later_commute.
      rewrite with_subst_later_commute.
      apply derives_weaken_head.
      apply derives_modal.
      rewrite <- with_subst_app.
      rewrite with_subst_normalize.
      simpl.
      rewrite with_subst_commute.
      2: simpl; intuition; subst; discriminate.
      apply derives_modal.
      unfold valuepred.
      apply respecting_with_subst.
      simpl; intuition; subst; discriminate.
     Qed.

    Lemma check_instr_fetch_1 : forall e v1 v2 t0 t1,
      not_in_dom v2 e ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v1 (t_pair t0 t1)))
                       (instr_fetch_field v1 1 v2)
                       (e && (t_slot v2 t1)).

      unfold check_instr; intros.
      intuition_d.
      apply step_guard.
      simpl; intuition; subst; auto.      
      exist_d ([](e && (t_slot v2 t1))).
      exist_d x.
      split_d.
      apply_d nec_later.      
      unfold guards, sub.
      rewrite modal_idem; auto.
      apply (with_projecting Sel_S); intro s.
      rewrite pmodal_conj; auto.
      elim_at 1.
      elim_at 1.
      unfold t_slot at 1.
      apply (field_pred_elim _ s); auto.
      case_eq (s#v1); intros; auto.
      unfold t_pair.      
      rewrite with_subst_field_pred.
      destruct v.
      change bot with (const False).
      rewrite bimodal_const; auto.
      apply with_step_intro with s (s#v2 <- v4); auto.
      apply step_fetch_field_1 with v3; auto.
      rewrite later_nec_later.
      repeat rewrite pmodal_conj; auto.
      elim_at 0.
      clear_at 0.
      split_d.
      in_goal rewrite with_subst_later_commute.
      apply_d nec_later.
      move_at 1.
      cut_with ([](projecting Sel_S s)).
      do 4 clear_at 2.
      coalesce_modal.      
      cut_using (H v4); auto.
      unfold with_var_upd.
      apply (field_pred_elim _ s); auto.
      move_at 2.
      apply derives_weaken_head.
      apply prove_adj with necessary.
      apply projecting_necessary.
      in_goal rewrite with_subst_later_commute.
      in_goal unfold t_slot.
      rewrite with_subst_field_pred.
      rewrite get_set_same.
      rewrite with_subst_later_commute.
      rewrite with_subst_later_commute.
      apply derives_weaken_head.
      apply derives_modal.
      rewrite <- with_subst_app.
      rewrite with_subst_normalize.
      simpl.
      rewrite with_subst_commute.
      2: simpl; intuition; subst; discriminate.
      apply derives_modal.
      unfold valuepred.
      apply respecting_with_subst.
      simpl; intuition; subst; discriminate.
     Qed.


     Lemma check_instr_cons: forall e v1 v2 v3 t0 t1,
      not_in_dom v3 e ->
    (*----------------------------------------------------*)
      check_instr p pt (e && (t_slot v1 ([](valuepred t0))) && (t_slot v2 ([](valuepred t1))))
                       (instr_cons v1 v2 v3)
                       (e && (t_slot v3 (t_pair t0 t1))).

       unfold check_instr; intros.
       intuition_d.
       apply step_guard.
       simpl; intuition; subst; auto.
       exist_d ([](e && (t_slot v3 (t_pair ([](valuepred t0)) ([](valuepred t1)))))).
       exist_d x.
       split_d.
       apply_d nec_later.
       unfold guards, sub.
       in_goal rewrite modal_idem; auto.
       rewrite pair_eat_nec.
       rewrite pair_eat_valuepred.
       auto; fail.

       clear_at 1.
       rewrite pmodal_conj; auto.
       elim_at 0.
       rewrite pmodal_conj; auto.
       elim_at 0.
       apply (with_projecting Sel_S); intro s.       
       elim_at 2.
       elim_at 3.
       unfold t_slot at 1 2.
       apply (field_pred_elim _ s); auto.
       case_eq (s#v2); simpl; intros; auto.
       move_at 1.
       apply (field_pred_elim _ s); auto.
       case_eq (s#v1); simpl; intros; auto.
       apply with_step_intro with s (s#v3 <- (value_cons v0 v)); auto.
       apply step_cons; auto.
       rewrite later_nec_later.
       repeat rewrite pmodal_conj; auto.
       split_d.
       rewrite with_subst_later_commute; auto.
       apply_d nec_later.
       cut_with ([] (projecting Sel_S s)).
       coalesce_modal.
       cut_using (H (value_cons v0 v)); auto.
       unfold with_var_upd.
       apply (field_pred_elim _ s); auto.
       move_at 2; apply derives_weaken_head.
       apply prove_adj with necessary.
       apply projecting_necessary.
       rewrite with_subst_later_commute.
       unfold t_slot.
       rewrite with_subst_field_pred.
       rewrite get_set_same.
       rewrite pair_eat_nec.
       unfold t_pair.         
       rewrite with_subst_field_pred.
       repeat rewrite pmodal_conj; auto; split_d.
       in_goal repeat rewrite <- with_subst_app.
       in_goal rewrite with_subst_normalize.
       simpl.
       rewrite <- with_subst_later_commute.
       in_goal rewrite with_subst_split_one.
       coalesce_modal.
       repeat rewrite with_subst_later_commute.
       apply_d modal_4; auto.
       apply_d nec_later.
       apply derives_modal.
       unfold valuepred.
       apply_d (respecting_with_subst (Sel_V::nil)).
       apply_d modal_4; auto.
       simpl; intuition; subst; discriminate.
       
       in_goal repeat rewrite <- with_subst_app.
       in_goal rewrite with_subst_normalize.
       simpl.
       rewrite <- with_subst_later_commute.
       in_goal rewrite with_subst_split_one.
       coalesce_modal.
       repeat rewrite with_subst_later_commute.
       apply_d modal_4; auto.
       apply_d nec_later.
       apply derives_modal.
       unfold valuepred.
       apply_d (respecting_with_subst (Sel_V::nil)).
       apply_d modal_4; auto.
       simpl; intuition; subst; discriminate.
      Qed.

  End HoareRules.
End LMHoareRules.
