(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import Setoid.
Require Import Base.
Require Import List.

Require Import Maps.
Require Import Machine.
Require Import ModelSpec.
Require Import LMModel.

Module Type LIST_MACHINE.
  Declare Module MM : MODAL_MODEL
    with Module W := LMSimpleWorlds
    with Module WS := LMStep.

  Export LMWorldCarrier.
  Export LMStep.
  Export MM.

  Open Local Scope pred_scope.
 
  Definition just := projecting Sel_V.

  Definition t_labelblock (p:program) (l:label) (i:instr) : pred :=
      const (p#l = Some i).

  Definition labelguard (p:program) (l: label) (e: pred) : pred :=
    All_ i:instr, (t_labelblock p l i) --> (guards p i ([]e)).

  Definition pt_domain_ok (pt:map pred) (p:program) : Prop :=
    forall l, (exists t, pt#l = Some t) -> (exists i, p#l = Some i).

  Definition pgm_typ (m:map pred) (p:program) : pred :=
    All_ l:label, All_ i:instr,
      t_labelblock p l i -->
      Ex_ e:pred,
         (const (m#l = Some e)) &&
         (labelguard p l e).

  Definition check_instr (p:program) (pt: map pred) (pre: pred) (i: instr) (post: pred) : Prop :=
     pt_domain_ok pt p ->
     [ |> (pgm_typ pt p) :: nil ] |-- All_ i':instr, (guards p i' ([]post)) --> (guards p (i;;i') ([]pre)).

  Definition check_block (p:program) (pt: map pred) (pre: pred) (i: instr) : Prop :=
     pt_domain_ok pt p ->
     [ |> (pgm_typ pt p) :: nil ] |-- guards p i ([] pre).

  Definition check_blocks (m:map pred) (p:program) :=
    forall l i, p#l = Some i ->
      exists e, m#l = Some e /\ check_block p m e i.

  Definition valuepred := respecting (Sel_V::nil).

  Definition t_slot (v:var) (p:pred) : pred :=
    field_pred Sel_S (fun s =>
      match s#v with
      | Some x => with_subst (U Sel_V x :: nil) p
      | None   => bot
      end).

  Definition t_pair (t1 t2:pred) : pred :=
    field_pred Sel_V (fun v =>
      match v with
      | value_cons a b => with_subst (U Sel_V a :: nil) (|> (valuepred t1))
                       && with_subst (U Sel_V b :: nil) (|> (valuepred t2))
      | _              => bot
      end).

  Definition t_nil : pred := just (value_label (L 0)).

  Definition t_anypair : pred :=
    Ex_ t1 : pred, Ex_ t2 : pred, t_pair t1 t2.

  Definition t_cont (p:program) (t: pred) : pred :=
    Ex_ l : label, Ex_ i:instr,
       (t_labelblock p l i) &&
       (just (value_label l)) &&
       (|> ([]t >=> safe p i)).

  Definition with_var_upd i v t :=
     field_pred Sel_S (fun s => with_subst (U Sel_S (s#i <- v) :: nil) t).

  Definition not_in_dom (i: var) (e: pred) : Prop :=
    forall v, e |-- with_var_upd i v e.


  Axiom nil_necessary : |-- necessary t_nil.
  Axiom nil_respects : forall l,
    In Sel_V l ->
    |-- respects l t_nil.

  Axiom sub_pair_later : forall f f' g g',
    [ |>(f >=> f') :: |>(g >=> g') :: nil ] |-- t_pair f g >=> t_pair f' g'.

  Axiom equ_pair_later : forall f f' g g',
    [ |>(f <=> f') :: |>(g <=> g') :: nil ] |-- t_pair f g <=> t_pair f' g'.

  Axiom sub_pair : forall f f' g g',
    [ (f >=> f') :: (g >=> g') :: nil ] |-- t_pair f g >=> t_pair f' g'.

  Axiom equ_pair : forall f f' g g',
    [ (f <=> f') :: (g <=> g') :: nil ] |-- t_pair f g <=> t_pair f' g'.

  Axiom pair_necessary : forall t1 t2,
    |-- necessary (t_pair t1 t2).

  Axiom adj_later_commute : forall z M (J:adjective z M) p,
    (forall p, M (|> p) = |> (M p)) ->
    |> (J p) |-- J (|> p).

  Axiom pair_respects : forall l t1 t2,
    In Sel_V l ->
    |-- respects l (t_pair t1 t2).

  Axiom cont_necessary : forall p t,
    |-- necessary (t_cont p t).

  Axiom cont_respects : forall l p t,
    In Sel_V l ->
    |-- respects l (t_cont p t).

  Axiom contractive_pair : forall f g,
    nonexpansive f ->
    nonexpansive g ->
    contractive (fun x => t_pair (f x) (g x)).

  Axiom nonexpansive_pair: forall f g,
    nonexpansive f ->
    nonexpansive g ->
    nonexpansive (fun x => t_pair (f x) (g x)).

  Axiom slot_necessary : forall p v,
    necessary p |-- necessary (t_slot v p).

  Axiom slot_intersection : forall v p q,
     t_slot v p && t_slot v q = t_slot v (p && q).

  Axiom pair_eat_valuepred : forall p q,
    t_pair (valuepred p) (valuepred q) = t_pair p q.

  Axiom pair_eat_nec : forall p q,
    t_pair ([]p) ([]q) = t_pair p q.

  Axiom sub_slot : forall v p p',
    p >=> p' |-- t_slot v p >=> t_slot v p'.

  Axiom sub_valuepred : forall p p',
    p >=> p' |-- valuepred p >=> valuepred p'.

  Definition listF (t:pred) (X:pred) := t_nil || (t_pair t X).
  Definition t_list (t: pred) := rec (listF t).
  Definition t_listcons (t: pred)  := t_pair t (t_list t).

  Axiom listF_contractive : forall t:pred, contractive (listF t).
  Axiom list_unfold: forall t, t_list t = t_nil || (t_pair t (t_list t)).

  Axiom list_necessary : forall t,
    |-- necessary (t_list t).

  Axiom list_respects : forall l t,
    In Sel_V l ->
    |-- respects l (t_list t).

  Axiom list_valuepred : forall t,
    valuepred (t_list t) = t_list t.
 
  Axiom nil_valuepred :
     valuepred t_nil = t_nil.

  Axiom listcons_valuepred : forall t,
     valuepred (t_listcons t) = t_listcons t.

  Axiom list_nec : forall t,
     [](t_list t) = t_list t.

  Axiom listcons_nec : forall t,
     [](t_listcons t) = t_listcons t.

  Axiom nil_nec : 
     []t_nil = t_nil.

  Axiom cont_valuepred : forall p e,
    valuepred (t_cont p e) = t_cont p e.

  Axiom cont_nec: forall p e,
    [] (t_cont p e) = t_cont p e.

  Axiom anypair_nec:  |-- necessary t_anypair.

  Axiom cont_sub : forall p t1 t2,
    t2 >=> t1 |-- t_cont p t1 >=> t_cont p t2.

  Axiom guard_sub : forall p e e' i,
    e >=> e' |-- guards p i e' --> guards p i e.

  Axiom subtype_nil: forall t, |-- t_nil >=> t_list t.

  Axiom subtype_list: forall t t', 
    t >=> t' |-- t_list t >=> t_list t'.

  Axiom subtype_listcons: forall t t',
    t >=> t' |-- t_listcons t >=> t_listcons t'.

  Axiom subtype_listmixed: forall t t',
    t >=> t' |-- t_listcons t >=> t_list t'.

  Axiom sub_list_nil_anypair : forall t,
    |-- t_list t >=> t_nil || t_anypair.

  Axiom sub_pair_listcons : forall t,
    |-- t_pair t (t_list t) >=> t_listcons t.

  Axiom nil_pair_bot : forall p q,
    [ t_nil :: t_pair p q :: nil ] |-- bot.

  Axiom sub_list_anypair_listcons : forall t,
    |-- t_list t && t_anypair >=> t_listcons t.

  Axiom sub_listcons_anypair : forall t,
    |-- t_listcons t >=> t_anypair.

  Axiom program_safety: forall p pt e i,
    check_blocks pt p ->
    pt_domain_ok pt p ->

    pt#entry_label = Some e ->
    p#entry_label = Some i ->
    e = top ->

    |-- safe p i.

End LIST_MACHINE.

Module ListMachine (MM': MODAL_MODEL
    with Module W := LMSimpleWorlds
    with Module WS := LMStep) : LIST_MACHINE.

  Import LMWorldCarrier.
  Import LMStep.
  Import MM'.

  Module MM := MM'.

  Open Local Scope pred_scope.
 
  Definition just := projecting Sel_V.

  Definition t_labelblock (p:program) (l:label) (i:instr) : pred :=
      const (p#l = Some i).

  Definition labelguard (p:program) (l: label) (e: pred) : pred :=
    All_ i:instr, (t_labelblock p l i) --> (guards p i ([]e)).

  Definition pt_domain_ok (pt:map pred) (p:program) : Prop :=
    forall l, (exists t, pt#l = Some t) -> (exists i, p#l = Some i).

  Definition pgm_typ (m:map pred) (p:program) : pred :=
    All_ l:label, All_ i:instr,
      t_labelblock p l i -->
      Ex_ e:pred,
         (const (m#l = Some e)) &&
         (labelguard p l e).

  Definition check_instr (p:program) (pt: map pred) (pre: pred) (i: instr) (post: pred) : Prop :=
     pt_domain_ok pt p ->
     [ |> (pgm_typ pt p) :: nil ] |-- All_ i':instr, (guards p i' ([]post)) --> (guards p (i;;i') ([]pre)).

  Definition check_block (p:program) (pt: map pred) (pre: pred) (i: instr) : Prop :=
     pt_domain_ok pt p ->
     [ |> (pgm_typ pt p) :: nil ] |-- guards p i ([] pre).

  Definition check_blocks (m:map pred) (p:program) :=
    forall l i, p#l = Some i ->
      exists e, m#l = Some e /\ check_block p m e i.

  Definition valuepred := respecting (Sel_V::nil).

  Definition t_slot (v:var) (p:pred) : pred :=
    field_pred Sel_S (fun s =>
      match s#v with
      | Some x => with_subst (U Sel_V x :: nil) p
      | None   => bot
      end).

  Definition t_pair (t1 t2:pred) : pred :=
    field_pred Sel_V (fun v =>
      match v with
      | value_cons a b => with_subst (U Sel_V a :: nil) (|> (valuepred t1))
                       && with_subst (U Sel_V b :: nil) (|> (valuepred t2))
      | _              => bot
      end).

  Definition t_nil : pred := just (value_label (L 0)).

  Definition t_anypair : pred :=
    Ex_ t1 : pred, Ex_ t2 : pred, t_pair t1 t2.

  Definition t_cont (p:program) (t: pred) : pred :=
    Ex_ l : label, Ex_ i:instr,
       (t_labelblock p l i) &&
       (just (value_label l)) &&
       (|> ([]t >=> safe p i)).

  Definition with_var_upd i v t :=
     field_pred Sel_S (fun s => with_subst (U Sel_S (s#i <- v) :: nil) t).

  Definition not_in_dom (i: var) (e: pred) : Prop :=
    forall v, e |-- with_var_upd i v e.

  Lemma nil_necessary : |-- necessary t_nil.
    unfold t_nil, just.

    apply projecting_necessary.
   Qed.

  Lemma nil_respects : forall l,
    In Sel_V l ->
    |-- respects l t_nil.

    unfold t_nil, just; intros.
    apply projecting_respects; auto.
   Qed.

  Lemma sub_pair_later : forall f f' g g',
    [ |>(f >=> f') :: |>(g >=> g') :: nil ] |-- t_pair f g >=> t_pair f' g'.

    intros.
    in_goal unfold sub.
    apply pdrop_modal; auto.
    simpl; intuition; subst t'; unfold sub; auto.
    apply pdrop_modal; auto.
    simpl; intuition; subst t'; unfold sub;
       repeat rewrite later_nec_later;
       rewrite every_later_commute; auto.
    unfold t_pair.
    apply (with_projecting Sel_V); intro v.
    intro_d.
    apply (field_pred_elim Sel_V v); [ assumed | ].
    apply field_pred_intro with v; [ assumed | ].
    destruct v; auto.
    intuition_d.
    cut_with (with_subst (U Sel_V v1::nil) (|> (valuepred (f --> f')))).
    coalesce_modal.
    coalesce_modal.
    coalesce_modal.
    apply_at 1; assumed.
    move_at 3; apply derives_weaken_head.
    apply_d every_with_subst.
    rewrite every_later_commute.
    apply derives_modal.
    unfold sub.
    unfold valuepred.
    elim_at 0.
    rewrite <- modal_idem; auto.
    apply derives_modal.
    apply everywhere_respecting.
    unfold sub; rewrite later_nec_later;
        rewrite <- every_later_commute; auto.
    cut_with (with_subst (U Sel_V v2::nil) (|> (valuepred (g --> g')))).
    coalesce_modal.
    coalesce_modal.
    coalesce_modal.
    apply_at 1; assumed.
    move_at 4; apply derives_weaken_head.
    apply_d every_with_subst.
    rewrite every_later_commute.
    apply derives_modal.
    elim_at 0.
    rewrite <- modal_idem; auto.
    apply derives_modal.
    unfold valuepred.
    apply everywhere_respecting.
   Qed.

  Lemma equ_pair_later : forall f f' g g',
    [ |>(f <=> f') :: |>(g <=> g') :: nil ] |-- t_pair f g <=> t_pair f' g'.

    intros; unfold equ;
       repeat rewrite pmodal_conj; auto;
       intuition_d; apply_d sub_pair_later; auto.
  Qed.

  Lemma sub_pair : forall f f' g g',
    [ (f >=> f') :: (g >=> g') :: nil ] |-- t_pair f g >=> t_pair f' g'.

    intros; apply_d sub_pair_later; apply_d nec_later;
       unfold sub; rewrite modal_idem; auto.
   Qed.

  Lemma equ_pair : forall f f' g g',
    [ (f <=> f') :: (g <=> g') :: nil ] |-- t_pair f g <=> t_pair f' g'.

    intros; unfold equ;
       repeat rewrite pmodal_conj; auto;
       intuition_d; apply_d sub_pair; auto.
   Qed.

  Lemma pair_necessary : forall t1 t2,
    |-- necessary (t_pair t1 t2).

    intros; unfold t_pair.
    apply_d field_pred_necessary; auto.
    intro_d; intro x.
    destruct x.
    apply_d adjective_bot.
    apply_d adjective_conj.
    apply_d with_subst_necessary; auto.
    apply_d later_necessary.
    apply_d with_subst_necessary; auto.
    apply_d later_necessary.
   Qed.  

  Lemma adj_later_commute : forall z M (J:adjective z M) p,
    (forall p, M (|> p) = |> (M p)) ->
    |> (J p) |-- J (|> p).

    intros.
    rewrite adj_pred_spec.
    rewrite adj_pred_spec.
    unfold equ.
    repeat rewrite pmodal_conj; auto.
    intuition_d.
    apply derives_weaken_head.
    rewrite H.
    unfold sub.
    rewrite <- nec_later_commute.
    apply derives_modal.
    rewrite <- every_later_commute.
    apply derives_modal.
    apply modal_K; auto.
    move_at 1; apply derives_weaken_head.    
    rewrite H.
    unfold sub.
    rewrite <- nec_later_commute.
    apply derives_modal.
    rewrite <- every_later_commute.
    apply derives_modal.
    apply modal_K; auto.
   Qed.

  Lemma pair_respects : forall l t1 t2,
    In Sel_V l ->
    |-- respects l (t_pair t1 t2).

    intros.
    unfold t_pair.
    apply_d field_pred_respects; auto.
    intro_d; intro x.
    destruct x.
    apply_d adjective_bot.
    apply_d adjective_conj.
    apply_d with_subst_respects; auto.
    apply_d later_respects.
    unfold valuepred.
    apply respecting_respects.
    red; simpl; intuition (subst; auto).
    apply_d with_subst_respects; auto.
    apply_d later_respects.
    unfold valuepred.
    apply respecting_respects.
    red; simpl; intuition (subst; auto).
  Qed.

  Lemma cont_necessary : forall p t,
    |-- necessary (t_cont p t).
    
    intros; unfold t_cont.
    apply_d adjective_exist; intro_d; intro l.
    apply_d adjective_exist; intro_d; intro i.
    apply_d adjective_conj.
    apply_d adjective_conj.
    unfold t_labelblock.
    apply_d adjective_const.
    unfold just.
    apply_d projecting_necessary; auto.
    apply_d later_necessary.
  Qed.

  Lemma cont_respects : forall l p t,
    In Sel_V l ->
    |-- respects l (t_cont p t).

    intros.
    unfold t_cont.
    apply_d adjective_exist; intro_d; intro lab.
    apply_d adjective_exist; intro_d; intro i.
    apply_d adjective_conj.
    apply_d adjective_conj.
    unfold t_labelblock.
    apply_d adjective_const.
    unfold just.
    apply_d projecting_respects; auto.
    apply_d later_respects.
    unfold sub.
    apply_d nec_respects.
    rewrite everywhere_respecting_nil.
    apply_d respecting_respects.
    red; simpl; intuition.
  Qed.

  Lemma contractive_pair : forall f g,
    nonexpansive f ->
    nonexpansive g ->
    contractive (fun x => t_pair (f x) (g x)).
    
    unfold contractive; intros.
    apply_d equ_pair_later.
    apply derives_modal; apply H.
    apply derives_modal; apply H0.   
   Qed.

  Lemma nonexpansive_pair: forall f g,
    nonexpansive f ->
    nonexpansive g ->
    nonexpansive (fun x => t_pair (f x) (g x)).
    
    unfold nonexpansive; intros.
    apply_d equ_pair; auto.
  Qed.

  Lemma slot_necessary : forall p v,
    necessary p |-- necessary (t_slot v p).

    intros.
    unfold t_slot.
    apply_d field_pred_necessary.
    intro_d; intro s.
    destruct (s#v).
    apply with_subst_necessary.
    apply_d adjective_bot.
   Qed.

  Lemma pair_eat_valuepred : forall p q,
    t_pair (valuepred p) (valuepred q) = t_pair p q.

    intros; unfold t_pair, valuepred.
    repeat rewrite modal_idem; auto.
   Qed.

  Lemma pair_eat_nec : forall p q,
    t_pair ([]p) ([]q) = t_pair p q.

    unfold t_pair.
    unfold valuepred.
    intros.
    repeat rewrite <- respecting_later_commute.   
    repeat rewrite later_nec_later.
    auto.
   Qed.

  Lemma sub_slot : forall v p p',
    p >=> p' |-- t_slot v p >=> t_slot v p'.

    intros; unfold t_slot.
    cut_with ([]!(p >=> p')).
    apply derives_weaken_head.
    unfold sub; do 2 apply derives_modal.
    apply (with_projecting Sel_S); intro s.
    intuition_d.
    apply (field_pred_elim _ s); auto.
    apply (field_pred_intro _ s); auto.
    destruct (s#v); auto.
    elim_at 2.
    cut_with (with_subst (U Sel_V v0 :: nil) (p --> p')).
    coalesce_modal.
    apply_at 1; auto.
    apply_d every_with_subst; auto.
    unfold sub; rewrite every_nec_commute; auto.
    rewrite modal_idem; auto.
   Qed.

  Lemma sub_valuepred : forall p p',
    p >=> p' |-- valuepred p >=> valuepred p'.

    intros.
    unfold valuepred.
    cut_with ([]!(p >=> p')).
    apply derives_weaken_head.
    unfold sub; do 2 apply derives_modal.
    cut_with (respecting (Sel_V::nil) (p --> p')).
    intuition_d; coalesce_modal.
    apply_d everywhere_respecting.
    elim_at 0; auto.
    unfold sub; rewrite every_nec_commute; auto.
    rewrite modal_idem; auto.
   Qed.

  Definition listF (t:pred) (X:pred) := t_nil || (t_pair t X).
  Definition t_list (t: pred) := rec (listF t).
  Definition t_listcons (t: pred)  := t_pair t (t_list t).

  Lemma listF_contractive : forall t:pred, contractive (listF t).
    intros; unfold listF.
    apply contractive_disj.
    apply contractive_const.
    change (t_pair t) with (fun x => t_pair ((fun _ => t) x) x).
    apply contractive_pair.
    apply contractive_nonexpansive.
    apply contractive_const.
    apply nonexpansive_id.
  Qed.

  Lemma list_unfold: forall t, t_list t = t_nil || (t_pair t (t_list t)).
    intro t.
    unfold t_list at 1.
    rewrite fold_unfold; auto.
    apply listF_contractive.
  Qed.

  Lemma list_necessary : forall t,
    |-- necessary (t_list t).
    
    intros.
    unfold t_list.
    apply_d (adjective_recursive _ _ necessary (listF t) (listF_contractive _)); auto.
    intro_d; intro p.
    apply modal_N; auto.
    intro_d; apply derives_axiom.
    unfold listF, equ; split_d;
      unfold sub; do 2 (apply modal_N; auto); intuition_d.
    apply_d modal_distrib_disj.
    left_d.
    apply (remove_adj_goal _ _ necessary); auto.
    apply_d nil_necessary.
    apply_d modal_distrib_disj.
    right_d.
    apply (remove_adj_goal _ _ necessary); auto.    
    apply_d pair_necessary.
    apply_d sub_project_herenow; [ | exact_at 0 ].
    apply_d sub_pair.
    auto.
    unfold sub.
    apply_d modal_N.
    
    elim_at 0. 
    elim_at 0; auto.
    right_d.
    unfold t_pair.
    unfold valuepred.
    rewrite respecting_nec_commute.
    rewrite later_nec_later.
    auto.
   Qed.

  Lemma list_respects : forall l t,
    In Sel_V l ->
    |-- respects l (t_list t).

    intros.
    unfold t_list.
    apply_d (adjective_recursive _ _ (respects l) (listF t) (listF_contractive _)); auto.
    intro_d; intro p.
    rewrite adj_nec.
    apply modal_N; auto.
    intro_d.
    apply derives_axiom.
    unfold equ; split_d; unfold sub;
       do 2 (apply modal_N; auto); intro_d.
    unfold listF; intuition_d.
    apply_d modal_distrib_disj.
    left_d.
    apply (remove_adj_goal _ _ (respects l)); auto.
    apply_d nil_respects.
    apply_d modal_distrib_disj.
    right_d.
    apply (remove_adj_goal _ _ (respects l)); auto.
    apply_d pair_respects.
    apply_d sub_project_herenow; [ | exact_at 0 ].
    apply_d sub_pair; auto.
    apply derives_axiom; unfold sub; do 2 (apply modal_N; auto).

    elim_at 0.
    unfold listF.
    intuition_d.
    right_d.
    unfold t_pair.    
    unfold valuepred.
    rewrite (respecting_compose (Sel_V::nil) (Sel_V::nil) l p); auto.
    red; simpl; intuition.
    red; simpl; intuition; subst; auto.
    simpl; intuition.
   Qed.

  Lemma list_valuepred : forall t,
    valuepred (t_list t) = t_list t.
 
    intros; unfold valuepred; apply derives_equal.
    elim_at 0; auto.
    apply prove_adj with (respects (Sel_V::nil)).
    apply list_respects.
    simpl; auto.
   Qed.

  Lemma nil_valuepred :
     valuepred t_nil = t_nil.

     intros.
     apply equ_equal.
     unfold valuepred.
     cut_with (respects (Sel_V::nil) t_nil).
     apply adj_elim.
     apply equ_sym.
     apply nil_respects.
     simpl; auto.
    Qed.

  Lemma listcons_valuepred : forall t,
     valuepred (t_listcons t) = t_listcons t.

     intros; unfold t_listcons.
     unfold valuepred.
     apply derives_equal.
     elim_at 0; auto.
     apply prove_adj with (respects (Sel_V::nil)).
     apply pair_respects.
     simpl; auto.
    Qed.

  Lemma list_nec : forall t,
     [](t_list t) = t_list t.

     intros; apply equ_equal.
     apply_d equ_sym.
     rewrite <- (adj_pred_spec _ _ necessary).
     apply list_necessary.
    Qed.

  Lemma listcons_nec : forall t,
     [](t_listcons t) = t_listcons t.

     intros; apply equ_equal.
     apply_d equ_sym.
     rewrite <- (adj_pred_spec _ _ necessary).
     unfold t_listcons.
     apply pair_necessary.
    Qed.

  Lemma nil_nec : 
     []t_nil = t_nil.

     intros; apply equ_equal.
     apply_d equ_sym.
     rewrite <- (adj_pred_spec _ _ necessary).
     apply nil_necessary.
    Qed.

  Lemma cont_valuepred : forall p e,
    valuepred (t_cont p e) = t_cont p e.

    intros; apply derives_equal.
    unfold valuepred.
    elim_at 0; auto.
    unfold valuepred.
    apply prove_adj with (respects (Sel_V::nil)).
    apply_d cont_respects; auto.
   Qed.

  Lemma cont_nec: forall p e,
    [] (t_cont p e) = t_cont p e.

    intros; apply derives_equal.
    elim_at 0; auto.
    apply prove_adj with necessary.
    apply_d cont_necessary; auto.
   Qed.

  Lemma slot_intersection : forall v p q,
     t_slot v p && t_slot v q = t_slot v (p && q).

     intros; apply derives_equal; unfold t_slot.
     apply (with_projecting Sel_S); intro s.
     elim_at 1.
     apply (field_pred_elim _ s); auto.
     move_at 1.
     apply (field_pred_elim _ s); auto.
     apply (field_pred_intro _ s); auto.
     destruct (s#v); auto.
     coalesce_modal; auto.
     apply (with_projecting Sel_S); intro s.
     move_at 1.
     apply (field_pred_elim _ s); auto.
     split_d; apply (field_pred_intro _ s); auto;
       destruct (s#v); auto;
       coalesce_modal; auto.
    Qed.

  Lemma anypair_nec:  |-- necessary t_anypair.
    unfold t_anypair.
    apply_d adjective_exist; intro_d; intro p.
    apply_d adjective_exist; intro_d; intro q.
    apply_d pair_necessary.
   Qed.

  Lemma cont_sub : forall p t1 t2,
    t2 >=> t1 |-- t_cont p t1 >=> t_cont p t2.
    
    intros.
    unfold t_cont, sub.
    cut_with ([]!|>![](t2 --> t1)).
    apply derives_weaken_head.
    do 2 apply derives_modal.
    intuition_d.
    exist_d x.
    exist_d x0.
    split_d.
    split_d; auto.
    rewrite later_nec_later.
    rewrite later_nec_later.
    intuition_d; do 2 coalesce_modal.
    intro_d.
    apply_at 1; auto.
    apply_at 2; auto.
    rewrite every_later_commute.    
    rewrite modal_idem; auto.
    rewrite nec_later_later.
    rewrite <- every_later_commute.
    rewrite later_nec_later.
    rewrite every_later_commute.
    apply nec_later.
  Qed.

  Lemma guard_sub : forall p e e' i,
    e >=> e' |-- guards p i e' --> guards p i e.

    unfold guards, sub; intros.
    pullout_modal 2.
    do 2 apply derives_modal.
    intuition_d.
    apply_at 1.
    apply_at 2; auto.
  Qed.

  Lemma subtype_nil: forall t, |-- t_nil >=> t_list t.
    intros.
    rewrite list_unfold.
    unfold sub; intuition.
  Qed.

  Lemma subtype_list: forall t t', 
    t >=> t' |-- t_list t >=> t_list t'.

    intros.
    apply generalized_loeb_rule.
    apply_d nec_later.
    unfold sub; rewrite modal_idem; auto.
    in_goal rewrite list_unfold.
    in_goal rewrite (list_unfold t').
    unfold sub.
    apply pdrop_modal; auto.
    simpl; intuition; subst t'0; auto.
    apply pdrop_modal; auto.  
    simpl; intuition; subst t'0; auto with modal.
    rewrite every_nec_commute.
    rewrite modal_idem; auto.
    repeat rewrite later_nec_later.
    rewrite every_later_commute.
    rewrite modal_idem; auto.
    intuition_d.
    right_d.
    apply_d sub_project_herenow; [ | exact_at 0 ].
    apply_d sub_pair_later; auto.
    unfold sub.
    move_at 1; apply derives_weaken_head.
    rewrite later_nec_later.
    apply nec_later.
  Qed.

  Lemma subtype_listcons: forall t t',
    t >=> t' |-- t_listcons t >=> t_listcons t'.

    unfold t_listcons.
    intros.
    apply_d sub_pair; auto.
    apply_d subtype_list; auto.
  Qed.

  Lemma subtype_listmixed: forall t t',
    t >=> t' |-- t_listcons t >=> t_list t'.

    intros.
    unfold t_listcons, sub.
    cut_with ([]![]!(t --> t')).
    do 2 coalesce_modal.
    intro_d.
    in_goal rewrite list_unfold.
    right_d.
    apply_d sub_project_herenow; [ | exact_at 0 ].
    apply_d sub_pair; auto.
    apply_d subtype_list; auto.
    rewrite every_nec_commute.
    rewrite modal_idem; auto.
  Qed.

  Lemma sub_list_nil_anypair : forall t,
    |-- t_list t >=> t_nil || t_anypair.
    
    intros.
    rewrite list_unfold.   
    unfold sub; repeat apply modal_N; auto.
    intuition_d.
    right_d.
    unfold t_anypair.
    apply exist_intro with t.
    apply exist_intro with (t_list t).
    auto.
   Qed.  

  Lemma sub_pair_listcons : forall t,
    |-- t_pair t (t_list t) >=> t_listcons t.

    intros.
    unfold t_listcons.
    apply_d sub_pair; auto.
  Qed.

  Lemma nil_pair_bot : forall p q,
    [ t_nil :: t_pair p q :: nil ] |-- bot.
    
    intros.
    apply (with_projecting Sel_V); intro v.
    unfold t_pair, t_nil.
    move_at 2; apply (field_pred_elim _ v); auto.
    destruct v; auto.
    unfold just.
    clear_at 0.
    refine (projecting_eq _ _ _ _ _ _ _ _).
    exact_at 0.
    exact_at 1.
    elim_at 0; intros.
    discriminate.
  Qed.

  Lemma sub_list_anypair_listcons : forall t,
    |-- t_list t && t_anypair >=> t_listcons t.

    intros.
    unfold t_listcons.
    unfold sub; repeat apply modal_N; auto.
    intuition_d.
    rewrite list_unfold.
    elim_at 0.
    unfold t_anypair.
    intuition_d.
    cut_with bot; auto.
    apply nil_pair_bot.
    apply_d sub_project_herenow.
    2: exact_at 0.
    apply_d sub_pair; auto.
    apply derives_axiom.
    unfold sub.
    repeat apply modal_N; auto.
    rewrite list_unfold.
    intuition_d.
    right_d.
    rewrite <- list_unfold.
    auto.
  Qed.

  Lemma sub_listcons_anypair : forall t,
    |-- t_listcons t >=> t_anypair.

    intros.
    unfold t_listcons, t_anypair.
    unfold sub; repeat apply modal_N; auto.
    intro_d.
    apply exist_intro with t.
    apply exist_intro with (t_list t).
    auto.
   Qed.

  Lemma pgm_typ_check_blocks_valid : forall m p,
    pt_domain_ok m p ->
    check_blocks m p ->

    |> (pgm_typ m p) |-- pgm_typ m p.

    unfold check_blocks.
    intros.
    in_goal unfold pgm_typ.
    intuition_d.
    unfold t_labelblock.
    elim_at 0; intros.
    destruct (H0 x x0) as [t [? ?]]; auto.
    apply exist_intro with t.
    split_d.
    apply const_intro; auto.
    unfold labelguard.
    unfold check_block in H3.
    intuition_d.
    unfold t_labelblock.
    elim_at 0; intros.
    replace x1 with x0; [ | congruence ].
    apply H3.
    auto.
   Qed.

  Lemma program_safety: forall p pt e i,
    check_blocks pt p ->
    pt_domain_ok pt p ->

    pt#entry_label = Some e ->
    p#entry_label = Some i ->
    e = top ->

    |-- safe p i.

    intros.    
    cut_with (pgm_typ pt p).
    unfold pgm_typ.
    spec_at 0 entry_label.
    spec_at 0 i.
    elim_at 0.
    unfold t_labelblock.
    apply const_intro; auto.
    elim_at 0; intro e0.
    elim_at 0.
    elim_at 0; intros.
    replace e0 with e; [ | congruence ].
    unfold labelguard.
    spec_at 0 i.
    elim_at 0.
    unfold t_labelblock; apply const_intro; auto.
    unfold guards.
    apply_d sub_project_herenow; [ exact_at 0 | ].
    subst e.
    apply_d modal_N.
    apply loeb_rule.
    apply (pgm_typ_check_blocks_valid _ _ H0 H).
   Qed.

End ListMachine.
