(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Require Import Base.
Require Import Bool.

Require Import CompleteInduction.
Require Export Maps.
Require Export Machine.

Module AdvancedTC.
  Inductive typ : Set :=
  | ty_top : typ
  | ty_bot : typ
  | ty_nil : typ
  | ty_list: typ -> typ
  | ty_listcons: typ -> typ
  | ty_cont : map typ -> typ.

  Definition env := map typ.
  Definition program_typing := map env.

  Fixpoint emptymap A (m:map A) :=
    match m with
      | Leaf => true
      | Node l None r => emptymap A l && emptymap A r
      | _ => false
    end.
  Implicit Arguments emptymap.

  Section check_sub_env.
    Variable f : typ -> typ -> bool.

    Definition check_node (o1 o2:option typ) :=
      match o1, o2 with
        | _, None => true
        | Some t1, Some t2 => f t1 t2
        | _, _ => false
      end.

    Fixpoint check_env_sub (m1 m2:env) { struct m1 } : bool :=
    match m1, m2 with
      | _, Leaf => true
      | Leaf, _ => emptymap m2
      | Node l1 o1 r1,
        Node l2 o2 r2 =>
        check_env_sub l1 l2 &&
        check_env_sub r1 r2 &&
        check_node o1 o2
    end.

    Fixpoint check_env_sub_contra (m1 m2:env) { struct m2 } : bool :=
    match m1, m2 with
      | _, Leaf => true
      | Leaf, _ => emptymap m2
      | Node l1 o1 r1,
        Node l2 o2 r2 =>
        check_env_sub_contra l1 l2 &&
        check_env_sub_contra r1 r2 &&
        check_node o1 o2
    end.
  End check_sub_env.

  Fixpoint check_subtype (t1 t2:typ) { struct t1 } : bool :=
    match t1, t2 with
      | _, ty_top =>  true
      | ty_bot, _ =>  true
      | ty_nil, ty_nil => true
      | ty_nil, ty_list t1 => true
      | ty_list s1, ty_list t1 => check_subtype s1 t1
      | ty_listcons s1, ty_list t1 => check_subtype s1 t1
      | ty_listcons s1, ty_listcons t1 => check_subtype s1 t1
      | ty_cont m1, ty_cont m2 => check_env_sub_contra check_subtype_contra m2 m1
      | _, _ => false
    end
  with check_subtype_contra (t1 t2:typ) { struct t2 } : bool :=
    match t1, t2 with
      | _, ty_top =>  true
      | ty_bot, _ =>  true
      | ty_nil, ty_nil => true
      | ty_nil, ty_list t1 => true
      | ty_list s1, ty_list t1 => check_subtype_contra s1 t1
      | ty_listcons s1, ty_list t1 => check_subtype_contra s1 t1
      | ty_listcons s1, ty_listcons t1 => check_subtype_contra s1 t1
      | ty_cont m1, ty_cont m2 => check_env_sub check_subtype m2 m1
      | _, _ => false
    end.

  Section env_bounds.
    Variable lub : typ -> typ -> typ.
    Variable glb : typ -> typ -> typ.
    
    Definition node_lub (o1 o2:option typ) : option typ :=
      match o1, o2 with
        | None, _ => None
        | _, None => None
        | Some t1, Some t2 => Some (lub t1 t2)
      end.

    Definition node_glb (o1 o2:option typ) : option typ :=
      match o1, o2 with
        | None, _ => o2
        | _, None => o1
        | Some t1, Some t2 => Some (glb t1 t2)
      end.

    Fixpoint env_lub (m1 m2:env) { struct m1 } : env :=
      match m1, m2 with
        | Leaf, _ => Leaf _
        | _, Leaf => Leaf _
        | Node l1 o1 r1,
          Node l2 o2 r2 => Node _ (env_lub l1 l2) (node_lub o1 o2) (env_lub r1 r2)
      end.

    Fixpoint env_glb (m1 m2:env) { struct m1 } : env :=
      match m1, m2 with
        | Leaf, _ => m2
        | _, Leaf => m1
        | Node l1 o1 r1,
          Node l2 o2 r2 => Node _ (env_glb l1 l2) (node_glb o1 o2) (env_glb r1 r2)
      end.
  End env_bounds.

  Fixpoint lub (s t:typ) { struct s } : typ :=
    match s, t with
      | _, ty_bot => s
      | ty_bot, _ => t

      | ty_nil, ty_nil => ty_nil
      | ty_nil, ty_list t1 => (ty_list t1)
      | ty_nil, ty_listcons t1 => (ty_list t1)

      | ty_list s1, ty_nil => (ty_list s1)
      | ty_list s1, ty_list t1 => ty_list (lub s1 t1)
      | ty_list s1, ty_listcons t1 => ty_list (lub s1 t1)

      | ty_listcons s1, ty_nil => ty_list s1
      | ty_listcons s1, ty_list t1 => ty_list (lub s1 t1)
      | ty_listcons s1, ty_listcons t1 => ty_listcons (lub s1 t1)

      | ty_cont m1, ty_cont m2 => ty_cont (env_glb glb m1 m2)

      | _, _ => ty_top
    end

  with glb (s t:typ) { struct s } : typ :=
    match s, t with
      | ty_top, _ => t
      | _, ty_top => s

      | ty_nil, ty_nil => ty_nil
      | ty_nil, ty_list _ => ty_nil

      | ty_list s1, ty_nil => ty_nil
      | ty_list s1, ty_list t1 => ty_list (glb s1 t1)
      | ty_list s1, ty_listcons t1 => ty_listcons (glb s1 t1)

      | ty_listcons s1, ty_list t1 => ty_listcons (glb s1 t1)
      | ty_listcons s1, ty_listcons t1 => ty_listcons (glb s1 t1)

      | ty_cont m1, ty_cont m2 => ty_cont (env_lub lub m1 m2)

      | _, _ => ty_bot
    end.

  Definition typecheck_branch
    (pt: program_typing) (vt: env) (l: label): bool :=
    match pt#l with
      | None => false
      | Some vt' => check_env_sub check_subtype vt vt'
    end.


  Fixpoint typecheck_instr (pt: program_typing) (e: env) (i: instr)
    {struct i}: option env :=
    match i with
      | i1 ;; i2 =>
        match typecheck_instr pt e i1 with
          | None    => None
          | Some e1 => typecheck_instr pt e1 i2
        end
      | instr_branch_if_nil v l =>
        match e#v with
          | Some (ty_list t) =>
            if typecheck_branch pt (e#v <- ty_nil) l
              then Some (e#v <- (ty_listcons t))
              else None
          | Some (ty_listcons t) =>
            Some e
          | Some ty_nil =>
            if typecheck_branch pt e l
              then Some e
              else None
          | _ => None
        end
      | instr_fetch_field v1 0 v2 =>
        match e#v1 with
          | Some (ty_listcons t) => Some (e#v2 <- t)
          | _ => None
        end
      | instr_fetch_field v1 1 v2 =>
        match e#v1 with
          | Some (ty_listcons t) => Some (e#v2 <- (ty_list t))
          | _ => None
        end
      | instr_cons v1 v2 v3 =>
        match e#v1, e#v2 with
          | Some t1, Some t2 =>
            match lub (ty_list t1) t2 with
              | ty_list t3 => Some (e#v3 <- (ty_listcons t3))
              | _ => None
            end
          | _, _ => None
        end
      | instr_getlabel 1%positive v => Some (e#v <- ty_nil)
      | instr_getlabel x v =>
        match pt#x with
          | Some cont_x => Some (e#v <- (ty_cont cont_x))
          | _ => None
        end
      | _ =>
        None
    end.


  Fixpoint typecheck_block (pt: program_typing) (e: env) (i: instr)
    {struct i}: bool :=
    match i with
      | instr_halt => true
      | i1 ;; i2 =>
        match typecheck_instr pt e i1 with
          | None => false
          | Some e1 => typecheck_block pt e1 i2
        end
      | instr_jump v =>
        match e#v with
          | Some (ty_cont v_cont) => check_env_sub check_subtype e v_cont
          | _ => false
        end
      | _ => false
    end.
  
  Definition check_one_block (pt:program_typing) (i:instr) (oe:option env) :=
    match oe with
      | Some e => typecheck_block pt e i
      | None   => false
    end.

  Definition typecheck_blocks (p:program) (pt:program_typing) :=
    map_forall2 _ _ (check_one_block pt) p pt.

  Definition check_one_domain (e:env) (oi:option instr) :=
    match oi with
      | Some i => true
      | None   => false
    end.

  Definition check_pt_domain (p:program) (pt:program_typing) : bool :=
    map_forall2 _ _ check_one_domain pt p.

  Definition typecheck (p:program) (pt:program_typing) : bool :=
    (typecheck_blocks p pt &&
      check_pt_domain p pt &&
      match pt#entry_label with
        | Some Leaf => true
        | _   => false
      end) % bool.
End AdvancedTC.
