(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Require Import Base.
Require Import Bool.

Require Export Maps.
Require Export Machine.
Require Export advanced_tc.

Module Type ADVANCED_TC_SAFETY.
  Parameter ty : Type.

  Parameter t_top      : ty.
  Parameter t_bot      : ty.
  Parameter t_nil      : ty.
  Parameter t_list     : ty -> ty.
  Parameter t_listcons : ty -> ty.
  Parameter t_cont     : program -> map ty -> ty.

  Definition env := map ty.
  Definition program_typing := map env.

  (* The subtyping relation for types,
     together with the properties required
     to prove soundness of the checking
     algorithm.
   *)
  Parameter subtype : ty -> ty -> Prop.
  
  (* Subtyping for environments is defined in
     a straightforward way from regular subtyping.
   *)
  Definition env_sub (e e':env) :=
    forall v t', e'#v = Some t' ->
       exists t, e#v = Some t /\ subtype t t'.

  Axiom sub_refl : forall t, subtype t t.

  Axiom sub_trans : forall t1 t2 t3,
     subtype t1 t2 ->
     subtype t2 t3 ->
     subtype t1 t3.

  Axiom sub_top  : forall t, subtype t t_top.
  Axiom sub_bot  : forall t, subtype t_bot t.

  Axiom sub_nil  : forall t,
     subtype t_nil (t_list t).

  Axiom sub_list : forall t1 t2,
     subtype t1 t2 ->
     subtype (t_list t1) (t_list t2).

  Axiom sub_listmixed : forall t1 t2,
     subtype t1 t2 ->
     subtype (t_listcons t1) (t_list t2).

  Axiom sub_listcons : forall t1 t2,
     subtype t1 t2 ->
     subtype (t_listcons t1) (t_listcons t2).
  
  Axiom sub_cont : forall p e1 e2,
     env_sub e2 e1 ->
     subtype (t_cont p e1) (t_cont p e2).

  Hint Resolve sub_refl sub_trans sub_nil
     sub_list sub_listmixed sub_listcons
     sub_top sub_bot
     : typing.

  (* The Hoare "triple" relation specialized to
     the types for this typechecker.  
   *)
  Parameter check_instr :
    program -> program_typing -> env -> instr -> env -> Prop.

  Axiom check_instr_seq : forall p pt e e' e'' i1 i2,
      check_instr p pt e  i1 e' ->
      check_instr p pt e' i2 e'' ->
      check_instr p pt e  (i1;;i2) e''.

  Axiom check_instr_branch_list : forall p pt e e1 v l t,
      e#v = Some (t_list t) ->
      pt#l = Some e1 ->
      env_sub (e#v <- t_nil) e1 ->
      check_instr p pt
        e (instr_branch_if_nil v l) (e#v <- (t_listcons t)).

  Axiom check_instr_branch_listcons : forall p pt e l v t,
      e#v = Some (t_listcons t) ->
      check_instr p pt
         e (instr_branch_if_nil v l) e.

  Axiom check_instr_branch_nil : forall p pt e v l e1 e2,
      e#v = Some t_nil ->
      pt#l = Some e1 ->
      env_sub e e1 ->
      check_instr p pt
         e (instr_branch_if_nil v l) e2.

  Axiom check_instr_fetch_0: forall p pt e v1 v2 t,
      e#v1 = Some (t_listcons t) ->
      check_instr p pt
         e (instr_fetch_field v1 0 v2) (e#v2 <- t).

  Axiom check_instr_fetch_1: forall p pt e v1 v2 t,
      e#v1 = Some (t_listcons t) ->
      check_instr p pt
         e (instr_fetch_field v1 1 v2) (e#v2 <- (t_list t)).

  Axiom check_instr_cons: forall p pt e v1 v2 v3 t1 t2 t3,
      e#v1 = Some t1 ->
      e#v2 = Some t2 ->
      subtype t1 t3 ->
      subtype t2 (t_list t3) ->
      check_instr p pt
         e (instr_cons v1 v2 v3) (e#v3 <- (t_listcons t3)).

  Axiom check_instr_getnil: forall p pt e v,
      check_instr p pt
         e (instr_getlabel (L 0) v) (e#v <- t_nil).

  Axiom check_instr_getlabel: forall p pt e e' v l,
      pt#l = Some e' ->
      check_instr p pt
         e (instr_getlabel l v) (e#v <- (t_cont p e')).

  (* A relation expressing sufficent preconditions for
     program safety of instruction blocks.
   *)
  Parameter check_block :
    program -> program_typing -> env -> instr -> Prop.

  Axiom check_block_halt: forall p pt e,
      check_block p pt e instr_halt.

  Axiom check_block_seq: forall p pt e i1 i2 e',
      check_instr p pt e i1 e' ->
      check_block p pt e' i2 ->
      check_block p pt e (i1 ;; i2).

  Axiom check_block_jump : forall p pt v e e',
      e#v = Some (t_cont p e') ->
      env_sub e e' ->
      check_block p pt e (instr_jump v).

  (* This relation expresses the fact that the program typing
     accuratly describes the given program.  That is, the sufficent
     conditions given in the program typing actually reflect the
     code found in the program.
   *)
  Definition check_blocks (p:program) (pt:program_typing) :=
   forall l i,
     p#l = Some i ->
     exists e,
       pt#l = Some e /\
       check_block p pt e i.

  (* This relation states that every label found in the program typing
     maps to some actual label in the program.
   *)

  Definition typing_dom_match (p: program) (pt: program_typing) : Prop :=
    forall l e, pt#l = Some e -> exists i, p#l = Some i.

  (* This is the main safety lemma.  It says that if all blocks
     check out with relation to the program typing, and the program
     typing domain checks out, and the entry label has no precondition,
     then the program is safe to run by jumping to the entry label.
   *)
  Axiom safety : forall (p:program) (pt: program_typing),
     check_blocks p pt ->
     typing_dom_match p pt ->
     pt#entry_label = Some (empty ty) ->
     safe_prog p.

End ADVANCED_TC_SAFETY.


(* This module signature expresses the goal theorem.
   It states that if a program passes the typechecker,
   then it is safe to run.  
 *)
Module Type ADVANCED_TC_SAFETY_THEOREM.
  Import AdvancedTC.

  Axiom typecheck_safeprog : forall
     (p:program) (pt:program_typing),
     
     typecheck p pt = true -> safe_prog p.

End ADVANCED_TC_SAFETY_THEOREM.


(* This module functor proves the main safety theorem when
   given an implemenation of SIMPLE_TC_SAFETY.
   The proofs in here are basically just a straightforward
   unrolling of the typechecking algorithm to show that
   it correctly applies the rules.  The proofs from
   SIMPLE_TC_SAFETY do the work of showing that these
   rules are sufficent for safety.
 *)

Module AdvancedTCSafetyTheorem (X:ADVANCED_TC_SAFETY)
  : ADVANCED_TC_SAFETY_THEOREM.

  Import AdvancedTC.

  Fixpoint interp_ty (p:program) (t:typ) : X.ty :=
    match t with
      | ty_top => X.t_top
      | ty_bot => X.t_bot
      | ty_nil => X.t_nil
      | ty_list t => X.t_list (interp_ty p t)
      | ty_listcons t => X.t_listcons (interp_ty p t)
      | ty_cont e => X.t_cont p (map_fmap typ X.ty (interp_ty p) e)
    end.

  Definition interp_env (p:program) (e:env) : X.env := map_fmap typ X.ty (interp_ty p) e.
  Definition interp_pt (p:program) (pt:program_typing) : X.program_typing
      := map_fmap env X.env (interp_env p) pt.

  Lemma ty_full_rect :
     forall P : typ -> Type,
       P ty_top ->
       P ty_bot ->
       P ty_nil ->
       (forall t : typ, P t -> P (ty_list t)) ->
       (forall t : typ, P t -> P (ty_listcons t)) ->
       (forall m : map typ, (forall i t, m#i = Some t -> P t) -> P (ty_cont m)) ->
       forall t : typ, P t.

    intros P Htop Hbot Hnil Hlist Hlistcons Hmap.
    fix 1; intro t; destruct t.
    apply Htop.
    apply Hbot.
    apply Hnil.    
    apply Hlist; apply ty_full_rect.
    apply Hlistcons; apply ty_full_rect.
    apply Hmap; induction m; simpl; intros;
      destruct i; simpl in *; try discriminate.
    apply IHm2 with i; apply H.
    apply IHm1 with i; apply H.
    destruct o; try discriminate.
    replace t with t0.
    apply ty_full_rect.
    congruence.
   Defined.

  Lemma check_env_eq : check_env_sub = check_env_sub_contra.
    extensionality f.
    extensionality m1.
    extensionality m2.
    revert m2; induction m1.
    simpl; intros; destruct m2; simpl; auto.
    intro m2; destruct m2; simpl; auto.
    rewrite IHm1_1.
    rewrite IHm1_2.
    auto.
   Qed.

  Lemma check_subtype_eq' : forall t1 t2,
    check_subtype t1 t2 = check_subtype_contra t1 t2 /\
    check_subtype t2 t1 = check_subtype_contra t2 t1.

    induction t1 using ty_full_rect;
      simpl; intros; destruct t2; simpl; auto.
    split; auto.
    destruct (IHt1 t2); auto.
    destruct (IHt1 t2); auto.
    revert m H.
    induction m0; simpl; intros; destruct m; simpl; auto.
    destruct (IHm0_1 m1).
    intros; apply H with (xO i); auto.
    destruct (IHm0_2 m2).
    intros; apply H with (xI i); auto.
    rewrite H0.
    rewrite H1.
    rewrite H2.
    rewrite H3.
    clear H0 H1 H2 H3.
    replace (check_node check_subtype_contra o o0)
       with (check_node check_subtype o o0).
    replace (check_node check_subtype_contra o0 o)
       with (check_node check_subtype o0 o).
    auto.
    destruct o; destruct o0; simpl; auto.
    destruct (H xH t0 (refl_equal _) t); auto.
    destruct o; destruct o0; simpl; auto.
    destruct (H xH t0 (refl_equal _) t); auto.
   Qed.

  Lemma check_subtype_eq :
    check_subtype = check_subtype_contra.

    extensionality t1.
    extensionality t2.
    destruct (check_subtype_eq' t1 t2); auto.
   Qed.

  Lemma check_sub_fixpoint : forall t1 t2,
    check_subtype t1 t2 =
    match t1, t2 with
      | _, ty_top =>  true
      | ty_bot, _ =>  true
      | ty_nil, ty_nil => true
      | ty_nil, ty_list t1 => true
      | ty_list s1, ty_list t1 => check_subtype s1 t1
      | ty_listcons s1, ty_list t1 => check_subtype s1 t1
      | ty_listcons s1, ty_listcons t1 => check_subtype s1 t1
      | ty_cont m1, ty_cont m2 => check_env_sub check_subtype m2 m1
      | _, _ => false
    end.

    intros t1 t2; destruct t1; destruct t2; simpl; auto.
    rewrite check_subtype_eq.
    rewrite check_env_eq; auto.
   Qed.

  Lemma emptymap_get_false : forall A (m:map A) v t,
    emptymap m = true ->
    m#v = Some t ->
    False.

    intros A m; induction m; intros.
    destruct v; discriminate.
    simpl in H; destruct o; try discriminate.
    destruct (andb_prop _ _ H).
    destruct v; try discriminate.
    apply IHm2 with v t; auto.
    apply IHm1 with v t; auto.
   Qed.

  Lemma check_env_sub_sound' : forall p (m1 m2:env),
    (forall i t1, m1#i = Some t1 -> forall t2,
      (check_subtype t1 t2 = true -> X.subtype (interp_ty p t1) (interp_ty p t2)) /\
      (check_subtype t2 t1 = true -> X.subtype (interp_ty p t2) (interp_ty p t1))) ->

    (check_env_sub check_subtype m1 m2 = true -> X.env_sub (interp_env p m1) (interp_env p m2)) /\
    (check_env_sub check_subtype m2 m1 = true -> X.env_sub (interp_env p m2) (interp_env p m1)).

    split; red; intros.
    revert H m2 H0 v t' H1.
    induction m1; simpl; intros.
    destruct m2; try discriminate.
    destruct v; discriminate.
    elimtype False.
    destruct v; simpl in *.
    destruct o; try discriminate.
    destruct (andb_prop _ _ H0).
    destruct (fmap_eqn2 _ _ m2_2 (interp_ty p) v t') as [x [? ?]]; auto.
    apply emptymap_get_false with typ m2_2 v x; auto.
    destruct o; try discriminate.
    destruct (andb_prop _ _ H0).
    destruct (fmap_eqn2 _ _ m2_1 (interp_ty p) v t') as [x [? ?]]; auto.
    apply emptymap_get_false with typ m2_1 v x; auto.
    destruct o; discriminate.
    destruct m2.
    destruct v; discriminate.
    destruct (andb_prop _ _ H0); clear H0.
    destruct (andb_prop _ _ H2); clear H2.
    destruct v; simpl in *.
    apply IHm1_2 with m2_2; auto; intros.
    apply H with (xI i); simpl; auto.
    apply IHm1_1 with m2_1; auto; intros.
    apply H with (xO i); simpl; auto.
    destruct o0; simpl in *; try discriminate.
    destruct o; simpl in *; try discriminate.
    exists (interp_ty p t0); split; auto.
    replace t' with (interp_ty p t); [ | congruence ].
    destruct (H xH t0 (refl_equal _) t).
    apply H2; auto.

    revert H m2 H0 v t' H1.
    induction m1; simpl; intros.
    destruct v; try discriminate.
    destruct m2; try discriminate.
    destruct o; try discriminate.
    elimtype False.    
    simpl in *.
    destruct (andb_prop _ _ H0).
    destruct v; try discriminate; simpl in *.
    destruct (fmap_eqn2 _ _ m1_2 (interp_ty p) v t') as [x [? ?]]; auto.
    apply emptymap_get_false with typ m1_2 v x; auto.
    destruct (fmap_eqn2 _ _ m1_1 (interp_ty p) v t') as [x [? ?]]; auto.
    apply emptymap_get_false with typ m1_1 v x; auto.
    simpl in *.
    destruct (andb_prop _ _ H0); clear H0.
    destruct (andb_prop _ _ H2); clear H2.
    destruct v; simpl in *.
    apply IHm1_2; auto; intros.
    apply H with (xI i); simpl; auto.
    apply IHm1_1; auto; intros.
    apply H with (xO i); simpl; auto.
    destruct o; try discriminate.
    destruct o0; simpl in *; try discriminate.
    exists (interp_ty p t0); split; auto.
    replace t' with (interp_ty p t); [ | congruence ].
    destruct (H xH t (refl_equal _) t0).
    apply H5; auto.
   Qed.

  Lemma check_sub_sound' : forall p t1 t2,
    (check_subtype t1 t2 = true -> X.subtype (interp_ty p t1) (interp_ty p t2)) /\
    (check_subtype t2 t1 = true -> X.subtype (interp_ty p t2) (interp_ty p t1)).

    intros p t1; induction t1 using ty_full_rect;
      intros; destruct t2; simpl in *; split; intros;
      try discriminate; auto with typing.
    apply X.sub_list; destruct (IHt1 t2); auto.
    apply X.sub_list; destruct (IHt1 t2); auto.
    apply X.sub_listmixed; destruct (IHt1 t2); auto.
    apply X.sub_listmixed; destruct (IHt1 t2); auto.
    apply X.sub_listcons; destruct (IHt1 t2); auto.
    apply X.sub_listcons; destruct (IHt1 t2); auto.
    apply X.sub_cont.
    destruct (check_env_sub_sound' p m m0); auto.
    apply H2; auto.
    rewrite check_env_eq.
    rewrite check_subtype_eq.
    auto.
    apply X.sub_cont.
    destruct (check_env_sub_sound' p m m0); auto.
    apply H1.
    rewrite check_env_eq.
    rewrite check_subtype_eq.
    auto.
   Qed.

  Lemma check_env_sub_sound : forall p e1 e2,
    check_env_sub check_subtype e1 e2 = true ->
    X.env_sub (interp_env p e1) (interp_env p e2).

    intros p e1 e2.
    destruct (check_env_sub_sound' p e1 e2); auto.
    intros; apply check_sub_sound'.
   Qed.

  Lemma check_sub_sound : forall p t1 t2,
    check_subtype t1 t2 = true ->
    X.subtype (interp_ty p t1) (interp_ty p t2).

    intros p t1 t2.
    destruct (check_sub_sound' p t1 t2); auto.
   Qed.

  Lemma env_lub_glb_commute' : forall (m1 m2:map typ),
    (forall i t, m1#i = Some t -> forall t2,
      (lub t t2 = lub t2 t /\ glb t t2 = glb t2 t)) ->

      (env_lub lub m1 m2 = env_lub lub m2 m1) /\
      (env_glb glb m1 m2 = env_glb glb m2 m1).

    intro m1; induction m1; intros; destruct m2; simpl; auto.
    destruct (IHm1_1 m2_1); simpl; intros.
    apply H with (xO i); auto.
    rewrite H0.
    rewrite H1.
    destruct (IHm1_2 m2_2); simpl; intros.
    apply H with (xI i); auto.
    rewrite H2.
    rewrite H3.
    replace (node_lub lub o o0) with (node_lub lub o0 o).
    replace (node_glb glb o o0) with (node_glb glb o0 o).
    auto.
    destruct o; destruct o0; simpl; auto.
    destruct (H xH t (refl_equal _) t0); congruence.
    destruct o; destruct o0; simpl; auto.
    destruct (H xH t (refl_equal _) t0); congruence.
   Qed.

  Lemma lub_glb_commute : forall t1 t2,
    (lub t1 t2 = lub t2 t1) /\
    (glb t1 t2 = glb t2 t1).

    induction t1 using ty_full_rect; intros; destruct t2; split;
      simpl; auto; try (destruct (IHt1 t2); congruence).
    destruct (env_lub_glb_commute' m m0 H); congruence.
    destruct (env_lub_glb_commute' m m0 H); congruence.
   Qed.

  Lemma env_lub_glb_commute : forall (m1 m2:map typ),
      (env_lub lub m1 m2 = env_lub lub m2 m1) /\
      (env_glb glb m1 m2 = env_glb glb m2 m1).

    intros; apply env_lub_glb_commute'.
    intros; apply lub_glb_commute.
   Qed.

  Lemma env_lub_glb_sound' : forall p (m1 m2:map typ),
     (forall i t, m1#i = Some t -> forall t2,
       (X.subtype (interp_ty p t) (interp_ty p (lub t t2))) /\
       (X.subtype (interp_ty p (glb t t2)) (interp_ty p t))) ->

       (X.env_sub (interp_env p m1) (interp_env p (env_lub lub m1 m2))) /\
       (X.env_sub (interp_env p (env_glb glb m1 m2)) (interp_env p m1)).

     induction m1; intro m2; destruct m2; simpl; intros; split;
       try (red; intros; destruct v; discriminate).
     red; intros; auto; exists t'; split; auto with typing.
     
     red; intros; destruct v; simpl in *.
     destruct (IHm1_2 m2_2); auto.
     intros; apply H with (xI i); auto.
     destruct (IHm1_1 m2_1); auto.
     intros; apply H with (xO i); auto.
     destruct o; destruct o0; try discriminate; simpl in *.
     exists (interp_ty p t); split; auto.
     destruct (H xH t (refl_equal _) t0).
     replace t' with (interp_ty p (lub t t0)); [ auto | congruence ].
     
     red; intros; destruct v; simpl in *.
     destruct (IHm1_2 m2_2); auto.
     intros; apply H with (xI i); auto.
     destruct (IHm1_1 m2_1); auto.
     intros; apply H with (xO i); auto.
     destruct o; destruct o0; try discriminate; simpl in *.
     exists (interp_ty p (glb t t0)); split; auto.
     destruct (H xH t (refl_equal _) t0).
     replace t' with (interp_ty p t); [ auto | congruence ].
     exists (interp_ty p t); split; auto.
     replace t' with (interp_ty p t); [ auto with typing | congruence ].
    Qed.

  Lemma lub_glb_sound : forall p t1 t2,
    (X.subtype (interp_ty p t1) (interp_ty p (lub t1 t2))) /\
    (X.subtype (interp_ty p (glb t1 t2)) (interp_ty p t1)).
    
    induction t1 using ty_full_rect; intro t2; destruct t2; split;
        intros; subst; simpl; auto with typing.
    apply X.sub_list; destruct (IHt1 t2); auto.
    apply X.sub_list; destruct (IHt1 t2); auto.
    apply X.sub_list; destruct (IHt1 t2); auto.
    apply X.sub_listmixed; destruct (IHt1 t2); auto.
    apply X.sub_listmixed; destruct (IHt1 t2); auto.
    apply X.sub_listcons; destruct (IHt1 t2); auto.
    apply X.sub_listcons; destruct (IHt1 t2); auto.
    apply X.sub_listcons; destruct (IHt1 t2); auto.
    apply X.sub_cont; destruct (env_lub_glb_sound' p _ m0 H); auto.
    apply X.sub_cont; destruct (env_lub_glb_sound' p _ m0 H); auto.
   Qed.

  Lemma env_lub_glb_sound : forall p (m1 m2:map typ),
    (X.env_sub (interp_env p m1) (interp_env p (env_lub lub m1 m2))) /\
    (X.env_sub (interp_env p (env_glb glb m1 m2)) (interp_env p m1)).

    intros; apply env_lub_glb_sound'.
    intros; apply lub_glb_sound.
   Qed.

  Lemma lub_list_invert : forall p t1 t2 t3,
    lub (ty_list t1) t2 = ty_list t3 ->
    X.subtype (interp_ty p t1) (interp_ty p t3).

    intros.
    destruct t2; simpl in *; try discriminate.
    replace t3 with t1; [ | congruence ].
    apply X.sub_refl.
    replace t3 with t1; [ | congruence ].
    apply X.sub_refl.
    simplify_eq H; intros.
    subst t3.
    destruct (lub_glb_sound p t1 t2); auto.
    simplify_eq H; intros.
    subst t3.
    destruct (lub_glb_sound p t1 t2); auto.
   Qed.

  Definition check_branch (p:program) (pt:program_typing) (vt:env) (l:label) : Prop :=
    match pt#l with
      | None => False
      | Some vt' => X.env_sub (interp_env p vt) (interp_env p vt')
    end.

  Lemma typecheck_branch_sound : forall p pt vt l,
    typecheck_branch pt vt l = true ->
    check_branch p pt vt l.

    unfold typecheck_branch, check_branch; intros.
    case_eq (pt#l); intros; rewrite H0 in H; try discriminate.
    apply check_env_sub_sound; auto.
   Qed.

  Lemma typecheck_instr_sound : forall p
    (pt: program_typing) (e e':env) (i:instr),

    typecheck_instr pt e i = Some e' ->
      X.check_instr p (interp_pt p pt) (interp_env p e) i (interp_env p e').

    intros p pt e e' i; revert e e'; induction i;
      intros; try discriminate.

    cut (l = 1%positive \/ l <> 1%positive).
    intro Hl; destruct Hl.
    subst l; simpl in *.
    replace e' with (e#v <- ty_nil); [ | congruence ].
    unfold interp_env.    
    rewrite fmap_update; simpl.
    apply X.check_instr_getnil.
    cut (Some e' =
      match pt#l with
        | Some cont_x => Some (e#v <- (ty_cont cont_x))
        | _ => None
      end).
    intro H1; rewrite H1 in H.
    case_eq (pt#l); intros.
    rewrite H2 in H.
    rewrite H2 in H1.
    replace e' with (e#v <- (ty_cont e0)); [ | congruence ].
    unfold interp_env.
    rewrite fmap_update; simpl.
    apply X.check_instr_getlabel; auto.
    unfold interp_pt.
    refine (fmap_eqn _ _ _ _ _ _ _); auto.
    rewrite H2 in H1; discriminate.
    destruct l; try discriminate; simpl in *; auto.
    contradiction H0; auto.
    destruct l; auto; right; discriminate.

    simpl in H.  
    case_eq (e#v); intros; rewrite H0 in H.
    destruct t; try discriminate.
    case_eq (typecheck_branch pt e l); intros;
      rewrite H1 in H; try discriminate.
    replace e' with e; [ | congruence ].
    generalize (typecheck_branch_sound p _ _ _ H1).
    unfold check_branch.
    case_eq (pt#l); intuition.
    eapply X.check_instr_branch_nil with (interp_env p e0).
    refine (fmap_eqn _ _ _ _ _ _ H0); auto.
    refine (fmap_eqn _ _ _ _ _ _ H2); auto.
    auto.

    case_eq (typecheck_branch pt (e#v <- ty_nil) l); intros;
      rewrite H1 in H; try discriminate.
    replace e' with (e#v <- (ty_listcons t)); [ | congruence ].
    unfold interp_env.
    rewrite fmap_update; simpl.
    generalize (typecheck_branch_sound p _ _ _ H1).
    unfold check_branch.
    case_eq (pt#l); intuition.
    apply X.check_instr_branch_list with (interp_env p e0); auto.
    refine (fmap_eqn _ _ _ _ _ _ H0).
    refine (fmap_eqn _ _ _ _ _ _ H2).
    change (X.t_nil) with (interp_ty p ty_nil).
    rewrite <- fmap_update.
    apply H3.

    replace e' with e; [ | congruence ].
    apply X.check_instr_branch_listcons with (interp_ty p t); auto.
    refine (fmap_eqn _ _ _ _ _ _ H0).
    discriminate.
    
    simpl in H.
    destruct n.
    case_eq (e#v); intros; rewrite H0 in H; try discriminate.
    destruct t; try discriminate.
    replace e' with (e#v0 <- t); [ | congruence ].
    unfold interp_env.
    rewrite fmap_update.
    apply X.check_instr_fetch_0; auto.
    refine (fmap_eqn _ _ _ _ _ _ H0).
    destruct n; try discriminate.
    case_eq (e#v); intros; rewrite H0 in H; try discriminate.
    destruct t; try discriminate.
    replace e' with (e#v0 <- (ty_list t)); [ | congruence ].
    unfold interp_env.
    rewrite fmap_update; simpl.
    apply X.check_instr_fetch_1; auto.
    refine (fmap_eqn _ _ _ _ _ _ H0).
    
    unfold typecheck_instr in H.
    case_eq (e#v); intros; rewrite H0 in H; try discriminate.
    case_eq (e#v0); intros; rewrite H1 in H; try discriminate.
    case_eq (lub (ty_list t) t0); intros; rewrite H2 in H; try discriminate.
    replace e' with (e#v1 <- (ty_listcons t1)); [ | congruence ].
    unfold interp_env.
    rewrite fmap_update; simpl.
    apply X.check_instr_cons with (interp_ty p t) (interp_ty p t0); auto.
    refine (fmap_eqn _ _ _ _ _ _ H0).
    refine (fmap_eqn _ _ _ _ _ _ H1).
    apply lub_list_invert with t0; auto.
    replace (X.t_list (interp_ty p t1)) with (interp_ty p (lub t0 (ty_list t))).
    destruct (lub_glb_sound p t0 (ty_list t)); auto.
    destruct (lub_glb_commute t0 (ty_list t)).
    rewrite H3.
    rewrite H2.
    auto.

    simpl in H.
    case_eq (typecheck_instr pt e i1); intros; rewrite H0 in H; try discriminate.
    apply X.check_instr_seq with (interp_env p e0).
    apply IHi1; auto.
    apply IHi2; auto.
  Qed.

  Lemma typecheck_block_sound : forall p
    (pt: program_typing) (e:env) (i:instr),

    typecheck_block pt e i = true ->
    X.check_block p (interp_pt p pt) (interp_env p e) i.

    intros p pt e i; revert e; induction i; simpl; intros;
      try discriminate.
    case_eq (e#v); intros.
    rewrite H0 in H.
    destruct t; intros; simpl in *; try discriminate.

    apply X.check_block_jump with (interp_env p m); auto.
    refine (fmap_eqn _ _ _ _ _ _ H0).
    apply check_env_sub_sound; auto.
    rewrite H0 in H; discriminate.

    apply X.check_block_halt.

    case_eq (typecheck_instr pt e i1); intros;
      rewrite H0 in H; try discriminate.
    apply X.check_block_seq with (interp_env p e0).
    apply typecheck_instr_sound; auto.
    apply IHi2; auto.
  Qed.

  Lemma typecheck_blocks_sound : forall p pt,
    typecheck_blocks p pt = true ->
    X.check_blocks p (interp_pt p pt).
   
   unfold typecheck_blocks; intros.
   unfold X.check_blocks; intros l i Hl.
   generalize (map_forall2_correct _ _  _ _ _ l i H Hl).
   case_eq (pt#l); intros.
   exists (interp_env p e); split.
   refine (fmap_eqn _ _ _ _ _ _ H0).
   refine (typecheck_block_sound _ _ _ _ _); auto.
   discriminate.
  Qed.
  
  Lemma check_pt_domain_sound : forall pt p,
    check_pt_domain p pt = true ->
    X.typing_dom_match p (interp_pt p pt).

    unfold check_pt_domain; red; intros.
    destruct (fmap_eqn2 _ _ _ _ _ _ H0) as [e0 ?]; intuition.
    generalize (map_forall2_correct _ _ _ _ _ l e0 H H2).
    unfold check_one_domain; intros.
    case_eq (p#l); intros.
    exists i; auto.
    rewrite H4 in H1.
    discriminate.
   Qed.    

  Theorem typecheck_safeprog : forall p pt,
    typecheck p pt = true -> safe_prog p.

    unfold typecheck; intros.
    destruct (andb_prop _ _ H) as [H1 H2]; clear H.
    destruct (andb_prop _ _ H1) as [H3 H4]; clear H1.
    apply X.safety with (interp_pt p pt).
    apply typecheck_blocks_sound; auto.
    apply check_pt_domain_sound; auto.
    case_eq (pt#entry_label); intros; rewrite H in H2; try discriminate.
    destruct e; simpl in H2; try discriminate.
    refine (fmap_eqn _ _ _ _ _ _ H).
   Qed.

End AdvancedTCSafetyTheorem.
