(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import ModelSpec.
Require Import Worlds.
Require Import Logic.
Require Import Modal.
Require Import GoedelLoeb.

Module FieldLogicTheory (W':WORLDS) :
  FIELD_LOGIC_THEORY with Module W:=W'.

  Module W:=W'.
  Module GL:=GoedelLoebTheory W'.

  Export GL.

  Open Local Scope pred_scope.

  Definition with_subst_rel (z:subst) :=
    fun w w' => world_update z w = w'.

  Definition with_subst (z:subst) : modality pol_neutral.
    intro z.
    refine (Build_modality _ (appPosMode (with_subst_rel z)) (with_subst_rel z) _).
    simpl; intuition.
    red; intros.
    destruct H; intuition.
    replace w' with x; auto.
    unfold with_subst_rel in *; congruence.

    red in H; unfold with_subst_rel in *.
    exists (world_update z w); intuition.
   Defined.
  Lemma with_subst_spec : forall z,
     modal_op (with_subst z) = appPosMode (with_subst_rel z).
    intros; auto.
   Qed.
  Lemma with_subst_rel_spec : forall z,
     modal_rel (with_subst z) = (with_subst_rel z).
    intros; auto.
   Qed.
  Opaque with_subst.

  Lemma with_subst_transitive : forall z,
     transitive _ (modal_rel (with_subst z)).

     red; intros.
     rewrite with_subst_rel_spec in H.
     rewrite with_subst_rel_spec in H0.
     rewrite with_subst_rel_spec.
     unfold with_subst_rel in *.
     subst.
     apply world_update_dup.
    Qed.

  Hint Resolve with_subst_transitive : modal.

  Definition respecting_rel (l:list field) :=
    fun w w' => 
       (forall x, In x l -> w_proj x w = w_proj x w')
      /\ world_ttl w = world_ttl w'.

  Definition respecting (l:list field) :=
     mkPosModality (respecting_rel l).

  Lemma respecting_spec : forall l,
     modal_op (respecting l) = appPosMode (respecting_rel l).
     intros; auto.
    Qed.

  Lemma respecting_rel_spec : forall l,
     modal_rel (respecting l) = respecting_rel l.
     intros; auto.
    Qed.
  Opaque respecting.


  Lemma respecting_refl : forall l, reflexive _ (modal_rel (respecting l)).
     intros; rewrite respecting_rel_spec.
     red; intros.
     red; intros.
     intuition; auto.
    Qed.

  Lemma respecting_trans : forall l, transitive _ (modal_rel (respecting l)).
    intros; rewrite respecting_rel_spec.
    red; intros.
    unfold respecting_rel in *; intuition.
    rewrite H1; auto.
   Qed.

  Hint Resolve respecting_refl respecting_trans.

  Lemma w_upd_noninterference_n : forall n w w' x,
    tick_world_n n w = Some w' ->
    w_proj x w = w_proj x w'.

    induction n; simpl; intros.
    congruence.
    case_eq (tick_world w); intros; rewrite H0 in H; try discriminate.
    transitivity (w_proj x w0).
    apply w_upd_noninterference; auto.
    apply IHn; auto.
   Qed.

  Lemma respecting_later_commute1 : forall l p,
    respecting l (|> p) |-- |> (respecting l p).

    intros.
    rewrite respecting_spec.
    rewrite laterM_spec.
    unfold later, derives, appPosMode, later_rel, respecting_rel.
    simpl; intuition.
    destruct H as [n H].
    case_eq (tick_world w); intros; rewrite H2 in H; try discriminate.
    destruct (untick_n (S n) w'0) as [w1 ?].
    apply H0 with w1; intros.
    split; intros.
    rewrite (w_upd_noninterference_n (S n) w w' x); auto.
    rewrite H3; auto.
    rewrite (w_upd_noninterference_n (S n) w1 w'0 x); auto.
    simpl; rewrite H2; auto.
    rewrite tick_ttl with w w0; auto.
    rewrite tick_n_ttl with n w0 w'; auto.
    rewrite tick_n_ttl with (S n) w1 w'0; auto.
    omega.
    exists n; auto.
   Qed.

  Lemma respecting_later_commute2 : forall l p,
    |> (respecting l p) |-- respecting l (|> p).

    intros.
    rewrite respecting_spec.
    rewrite laterM_spec.
    unfold later, derives, appPosMode, later_rel, respecting_rel.
    simpl; intuition.
    destruct H as [n H].
    case_eq (tick_world_n (S n) w); intros.
    apply H0 with w0.
    exists n; auto.
    split; intros.
    rewrite <- (w_upd_noninterference_n (S n) w w0 x); auto.
    rewrite <- (w_upd_noninterference_n (S n) w' w'0 x); auto.
    revert H3.
    rewrite (tick_n_ttl (S n) w w0); auto.
    rewrite (tick_n_ttl (S n) w' w'0); auto.
    intros; omega.
    destruct (tick_n_succeeds (S n) w).
    rewrite H3.
    rewrite (tick_n_ttl (S n) w' w'0).
    omega.
    auto.
    rewrite H4 in H5; discriminate.
   Qed.

  Lemma respecting_later_commute : forall l p,
    respecting l (|> p) = |> (respecting l p).
    
    intros; apply derives_equal.
    apply respecting_later_commute1; auto.
    apply respecting_later_commute2; auto.
   Qed.

  Lemma respecting_nec_commute : forall l p,
    respecting l ([] p) = [] (respecting l p).

    intros.
    rewrite nec_now_later.
    rewrite nec_now_later.
    rewrite pmodal_conj; auto.
    rewrite respecting_later_commute; auto.
   Qed.

  Lemma everywhere_respecting : forall l p,
    !p |-- respecting l p.

    intros.
    rewrite everywhere_spec.
    rewrite respecting_spec.
    unfold everywhere, everywhere_rel, respecting_rel, appPosMode, derives.
    simpl; intuition.
   Qed.

  Lemma respecting_incl : forall l1 l2 p,
    incl l1 l2 ->
    respecting l1 p |-- respecting l2 p.

    intros.
    repeat rewrite respecting_spec.
    unfold appPosMode, respecting_rel, derives; simpl; intuition.
   Qed.

  Lemma everywhere_respecting_nil : forall p,
    !p = respecting nil p.

    intros.
    rewrite everywhere_spec.
    rewrite respecting_spec.
    unfold everywhere, appPosMode.
    refine (extensionality _ _ _); intro w.
    apply prop_ext; intuition.
    apply H.
    destruct H0; auto.
    apply H.
    split; auto.
    simpl; intuition.
   Qed.

  Lemma respecting_compose1 : forall l l1 l2 p,
    incl l l1 -> incl l l2 ->
    respecting l p |-- respecting l1 (respecting l2 p).

    intros.
    repeat rewrite respecting_spec.
    unfold appPosMode, respecting_rel, derives.
    simpl; intuition.
    apply H2; clear H2; intuition.

    revert H H1 H0 H4 H6; clear.
    revert x l1 l2 w w' w'0.
    induction l; simpl; intuition; subst.
    rewrite H4.
    rewrite H6.
    auto.
    apply H0; simpl; auto.
    apply H; simpl; auto.
    apply IHl with l1 l2 w'; auto.
    red; intros; apply H; simpl; auto.
    red; intros; apply H0; simpl; auto.
   Qed.

  Lemma respecting_compose2 : forall l l1 l2 p,
    (forall x, In x l1 /\ In x l2 -> In x l) ->
    respecting l1 (respecting l2 p) |-- respecting l p.

    intros.
    repeat rewrite respecting_spec.
    unfold appPosMode, respecting_rel, derives.
    simpl; intuition.
    set (w0 := world_update (map (fun x => U x (w_proj x w')) l2) w).
    apply H1 with w0; clear H1; intuition; subst w0.
    cut (In x l2 -> w_proj x w = w_proj x w').
    clear; revert x.
    induction l2; simpl; intuition; subst.
    destruct (field_eq_dec x a).
    subst a.
    rewrite proj_upd_same; auto.
    rewrite proj_upd_diff; auto.
    intros; apply H3; auto.
    clear; induction l2; simpl; intros; auto.
    rewrite <- world_ttl_noninterference.
    apply IHl2; auto.

    revert H0; clear; induction l2; simpl; intuition; subst.
    rewrite proj_upd_same; auto.
    destruct (field_eq_dec x a); subst.
    rewrite proj_upd_same; auto.
    rewrite proj_upd_diff; auto.

    rewrite <- H4.
    clear; induction l2; simpl; intros; auto.
    rewrite <- world_ttl_noninterference.
    auto.
   Qed.

  Lemma respecting_compose : forall l l1 l2 p,
    incl l l1 -> incl l l2 ->
    (forall x, In x l1 /\ In x l2 -> In x l) ->
    respecting l1 (respecting l2 p) = respecting l p.

    intros; apply derives_equal.
    apply respecting_compose2; auto.
    apply respecting_compose1; auto.
   Qed.

  Lemma respecting_intersect : forall l1 l2 p,
    respecting l1 (respecting l2 p) =
    respecting (set_inter field_eq_dec l1 l2) p.

    intros; apply respecting_compose.
    red; intros x H; refine (set_inter_elim1 _ _ l1 l2 H).
    red; intros x H; refine (set_inter_elim2 _ _ l1 l2 H).
    intros; apply set_inter_intro; intuition.
   Qed.

  Lemma equ_respecting : forall l p q,
    p <=> q |-- respecting l p <=> respecting l q.

    intros; unfold equ; intuition_d; unfold sub.
    clear_at 1.
    cut_with ([]!!(p --> q)).
    clear_at 1.
    do 2 apply derives_modal.
    intuition_d.
    cut_with (respecting l (p --> q)).
    coalesce_modal.
    elim_at 1; auto.
    apply_d everywhere_respecting; auto.
    rewrite modal_idem; auto.
    clear_at 0.
    cut_with ([]!!(q --> p)).
    clear_at 1.
    do 2 apply derives_modal.
    intuition_d.
    cut_with (respecting l (q --> p)).
    coalesce_modal.
    elim_at 1; auto.
    apply_d everywhere_respecting; auto.
    rewrite modal_idem; auto.
   Qed.

  Lemma every_with_subst : forall z p,
    !p |-- with_subst z p.

    intros z p.
    rewrite everywhere_spec.
    rewrite with_subst_spec.
    unfold derives, everywhere, appPosMode, everywhere_rel, with_subst_rel.
    simpl; intuition.
    subst w'.
    apply H0.
    clear; induction z; simpl; auto.
    destruct a; simpl.
    rewrite <- (world_ttl_noninterference (world_update z w) x f).
    auto.
   Qed.

  Lemma with_subst_same : forall z1 z2 t,
    (forall w, world_update z1 w = world_update z2 w) ->
    with_subst z1 t = with_subst z2 t.

    intros.
    repeat rewrite with_subst_spec.
    unfold appPosMode.
    refine (extensionality _ _ _); intro w.
    unfold with_subst_rel.
    rewrite H; auto.
   Qed.

  Lemma with_subst_nil : forall t,
    with_subst nil t = t.

    intros.
    rewrite with_subst_spec.
    unfold appPosMode.
    refine (extensionality _ _ _); intro w.
    unfold with_subst_rel.
    apply prop_ext.
    split; intros.
    apply H; simpl; auto.
    simpl in *; subst; auto.
   Qed.

  Lemma with_subst_float : forall x z t,
    with_subst z t = with_subst (float_sel x z) t.

    intros.
    apply with_subst_same.
    intros; apply float_sel_eq.
   Qed.

  Lemma derives_with_subst_commute : forall z1 z2 t,
    (forall x, In x (upd_set z1) -> In x (upd_set z2) -> False) ->
    with_subst z1 (with_subst z2 t) |-- with_subst z2 (with_subst z1 t).

    unfold derives; simpl; intuition.
    rewrite with_subst_spec.
    rewrite with_subst_spec.
    rewrite with_subst_spec in H1.
    rewrite with_subst_spec in H1.
    unfold appPosMode in *; intros.
    unfold with_subst_rel in *.
    subst w'.
    rewrite <- world_update_app in H3.
    rewrite commute_subst in H3; auto.
    rewrite world_update_app in H3.
    apply H1 with (world_update z1 w); auto.
   Qed.

  Lemma with_subst_commute : forall z1 z2 t,
    (forall x, In x (upd_set z1) -> In x (upd_set z2) -> False) ->
    with_subst z1 (with_subst z2 t) = with_subst z2 (with_subst z1 t).

    intros; apply derives_equal; apply derives_with_subst_commute; auto.
    intros; apply H with x; auto.
   Qed.

  Lemma with_subst_app : forall z1 z2 t,
    with_subst (z2++z1) t = with_subst z1 (with_subst z2 t).

    intros; repeat rewrite with_subst_spec.
    unfold appPosMode.
    refine (extensionality _ _ _); intro w.
    apply prop_ext.
    split; intros.
    apply H.  
    unfold with_subst_rel in *.
    rewrite world_update_app.
    congruence.
    unfold with_subst_rel in *.
    rewrite world_update_app in H0.
    apply H with (world_update z1 w); auto.
   Qed.

  Lemma respecting_with_subst : forall l z p,
    (forall x, In x (upd_set z) -> ~In x l) ->
    respecting l p |-- with_subst z p.

    intros.
    rewrite respecting_spec.
    rewrite with_subst_spec.
    unfold derives, appPosMode, respecting_rel, with_subst_rel.
    simpl; intuition.
    subst w'.
    apply H1.
    split.
    intros.
    symmetry; apply proj_subst_notin with (upd_set z).
    red; intros; auto.
    red; intro.
    apply (H x); auto.
    clear; induction z; simpl; auto.
    destruct a.
    rewrite <- world_ttl_noninterference.
    auto.
   Qed.

  Lemma with_subst_normalize : forall z p,
    with_subst z p = with_subst (normalize_update z) p.

    intros; apply with_subst_same.
    intros; apply world_update_normalize.
   Qed.

  Lemma with_subst_split_one : forall x a z p,
    with_subst (U x a :: z) p = with_subst z (with_subst (U x a :: nil) p).

    intros; rewrite <- with_subst_app; auto.
   Qed.

  Lemma with_subst_later_commute : forall z p,
    with_subst z (|> p) = |> (with_subst z p).

    intros; apply derives_equal.

    (* forward direction *)
    rewrite with_subst_spec.
    rewrite laterM_spec.
    unfold derives, later, appPosMode, with_subst_rel.
    simpl; intuition.
    subst w'0.
    apply H0 with (world_update z w); auto.
    unfold later_rel in *.
    destruct H as [n H].
    exists n.
    revert H.
    generalize (S n).
    clear; intro n; revert w w' z.
    induction n; simpl; intros; auto.
    replace w' with w; [ auto | congruence ].
    case_eq (tick_world w); intros; rewrite H0 in H; try discriminate.
    destruct (tick_succeeds (world_update z w)).
    replace (world_ttl (world_update z w))
       with (world_ttl w).
    rewrite (tick_ttl w w0); auto with arith.
    clear; induction z; simpl; auto.
    destruct a.
    rewrite <- world_ttl_noninterference; auto.
    rewrite H1.
    destruct (tick_n_succeeds n x).    
    cut (S (world_ttl x) >= S n); [ omega | ].
    rewrite <- (tick_ttl (world_update z w) x); auto.
    replace (world_ttl (world_update z w)) with (world_ttl w).
    rewrite (tick_ttl w w0); auto.
    rewrite (tick_n_ttl n w0 w'); auto.
    omega.
    clear; induction z; simpl; intros; auto.
    destruct a.
    rewrite <- world_ttl_noninterference; auto.
    generalize (IHn w0 w' z H); intros.
    rewrite <- H3.
    cut (x = world_update z w0).    
    intros; subst x; auto.
    revert H1 H0; clear.
    revert w w0 x; induction z; simpl; intros; auto.
    congruence.
    destruct a.
    destruct (tick_upd_commute (world_update z w) x x0 f).
    destruct (H2 H1).
    clear H H2.
    destruct H3.
    subst x.
    rewrite (IHz w w0 x1); auto.
   
    (* backward direction *)
    rewrite with_subst_spec.
    rewrite laterM_spec.
    unfold derives, later, appPosMode, with_subst_rel.
    simpl; intuition.
    subst w'.
    cut (exists a, later_rel w a /\ world_update z a = w'0).
    intros [a [Ha1 Ha2]].
    apply H0 with a; auto.
    clear H0 H1.
    destruct H2 as [n H2].
    revert w w'0 z H2.
    induction n; simpl; intros.
    case_eq (tick_world (world_update z w)); intros;
      rewrite H in H2; try discriminate.
    replace w'0 with w0 in *; [ | congruence ].
    clear H2 w'0.
    revert w w0 H.
    induction z; simpl; intros.
    exists w0; auto; split; auto.
    exists 0; simpl; auto.
    rewrite H; auto.
    destruct a.
    destruct (tick_upd_commute (world_update z w) w0 x f).
    destruct (H1 H); clear H0 H1.
    destruct H2.
    destruct (IHz _ _ H0).    
    destruct H2.
    exists x1; split; auto.
    subst; auto.
    case_eq (tick_world (world_update z w)); intros; rewrite H in H2; try discriminate.
    change (tick_world_n (S n) w0 = Some w'0) in H2.    
    cut (exists a, tick_world w = Some a /\ world_update z a = w0).
    intros [a [Ha1 Ha2]].
    subst w0.
    destruct (IHn _ _ _ H2).
    destruct H0.
    exists x; split; auto.
    destruct H0.
    exists (S x0).
    simpl.
    rewrite Ha1; auto.
    revert H; clear; revert w w0.
    induction z; simpl; intros.
    exists w0; auto.
    destruct a.
    destruct (tick_upd_commute (world_update z w) w0 x f).
    destruct (H1 H); clear H0 H1.
    destruct H2.
    destruct (IHz _ _ H0).    
    destruct H2.
    subst.
    exists x1; split; auto.
   Qed.

  Lemma with_subst_nec_commute : forall z p,
    with_subst z ([] p) = [] (with_subst z p).

    intros; repeat rewrite nec_now_later.
    rewrite pmodal_conj; auto.
    rewrite with_subst_later_commute; auto.
   Qed.


  Lemma sub_with_subst : forall z p q,
    p >=> q |-- with_subst z p >=> with_subst z q.

    intros.
    unfold sub.
    rewrite <- (modal_idem _ everywhereM); auto.
    do 2 apply derives_modal.
    intro_d.
    cut_with (with_subst z (p --> q)).
    coalesce_modal.
    apply_at 1; auto.
    clear_at 0.
    rewrite everywhere_respecting_nil.
    apply respecting_with_subst.
    simpl; intuition.
   Qed.

  Lemma equ_with_subst : forall z p q,
    p <=> q |-- with_subst z p <=> with_subst z q.

    intros; unfold equ; intuition_d.
    apply_d sub_with_subst; auto.
    apply_d sub_with_subst; auto.
   Qed.

  Lemma every_eats_with_subst : forall z p,
    with_subst z (!p) = !p.

    intros; apply derives_equal.
    rewrite everywhere_spec.    
    rewrite with_subst_spec.
    unfold everywhere, appPosMode, with_subst_rel, everywhere_rel.
    unfold derives; simpl; intuition.
    apply H0 with (world_update z w); auto.
    rewrite <- H.
    clear; induction z; simpl; auto.
    destruct a.
    rewrite <- world_ttl_noninterference; auto.

    rewrite everywhere_spec.    
    rewrite with_subst_spec.
    unfold everywhere, appPosMode, with_subst_rel, everywhere_rel.
    unfold derives; simpl; intuition.
    apply H0.
    rewrite <- H2.
    revert H; clear; revert w w'; induction z; simpl; intros.
    subst; auto.
    destruct a.
    subst w'.
    rewrite <- world_ttl_noninterference; auto.
   Qed.

  Lemma sub_eats_with_subst : forall z p q,
    with_subst z (p >=> q) = p >=> q.

    intros; unfold sub.
    rewrite <- every_nec_commute.
    rewrite every_eats_with_subst.
    auto.
   Qed.

  Lemma with_subst_field_pred : forall x (a:field_type x) z F,
    with_subst (U x a :: z) (field_pred x F) = 
    with_subst (U x a :: z) (F a).

    intros.
    rewrite with_subst_spec.
    unfold appPosMode, with_subst_rel, field_pred.
    refine (extensionality _ _ _); intro w.
    apply prop_ext; split; intros.
    subst w'.
    simpl.
    pattern a at 1.
    rewrite <- (proj_upd_same x a (world_update z w)).
    apply H.
    simpl; auto.
    subst w'.
    simpl.
    rewrite proj_upd_same.
    apply H; simpl; auto.
   Qed.

  Lemma with_subst_projecting : forall x (a b:field_type x) z,
    with_subst (U x a :: z) (projecting x b) = (const (a = b)).

    intros.
    rewrite with_subst_spec.
    unfold appPosMode, with_subst_rel, projecting, const.
    refine (extensionality _ _ _); intro w.
    apply prop_ext.
    split; intros.
    generalize (H (world_update (U x a::z) w) (refl_equal _)).
    simpl.
    rewrite proj_upd_same.
    auto.
    subst w'.
    simpl.
    rewrite proj_upd_same.
    auto.
   Qed.

End FieldLogicTheory.
