(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import Max.

Require Import CompleteInduction.
Require Import ModelSpec.
Require Import Worlds.
Require Import Logic.
Require Import Modal.
Require Import Tactics.


Module GoedelLoebTheory (W':WORLDS)
  : GOEDEL_LOEB_THEORY with Module W:=W'.

  Module W:=W'.
  Module Tac := LogicTactics W'.
  
  Export Tac.

  Open Local Scope pred_scope.

  Definition later_rel (w1 w2:world) :=
    exists n, tick_world_n (S n) w1 = Some w2.
  Definition later := appPosMode later_rel.
  
  Definition laterM : modality pol_pos := mkPosModality later_rel.
  Lemma laterM_spec : modal_op laterM = later.
     auto.
   Qed.
  Lemma laterM_rel_spec : modal_rel laterM = later_rel.
     auto.
   Qed.
  Notation "'|>' e" := (laterM e) (at level 30, right associativity): pred_scope.
  Opaque laterM.

  Definition nec_rel (w1 w2:world) :=
    exists n, tick_world_n n w1 = Some w2.
  Definition nec := appPosMode nec_rel.

  Definition necM : modality pol_pos :=
    mkPosModality nec_rel.
  Lemma necM_spec : modal_op necM = nec.
    auto.
   Qed.    
  Lemma necM_rel_spec : modal_rel necM = nec_rel.
    auto.
   Qed.
  Notation "'[]' e" := (necM e) (at level 30, right associativity): pred_scope.
  Opaque necM.

  Definition eventually := appNegMode nec_rel.

  Definition eventuallyM : modality pol_neg :=
    mkNegModality nec_rel.
  Lemma eventually_spec : modal_op eventuallyM = eventually.
     auto.
   Qed.
  Lemma eventually_rel_spec : modal_rel eventuallyM = nec_rel.
     auto.
   Qed.
  Opaque eventuallyM.
  Notation "'<>' e" := (eventuallyM e) (at level 30, right associativity): pred_scope.

  Lemma nec_eventually_dual : modal_dual necM eventuallyM.
    red; intuition.
   Qed.

  Definition everywhere_rel (w1 w2:world) := world_ttl w1 = world_ttl w2.
  Definition everywhere := appPosMode everywhere_rel.

  Definition everywhereM : modality pol_pos :=
       mkPosModality everywhere_rel.
  Lemma everywhere_spec : modal_op everywhereM = everywhere.
    auto.
   Qed.
  Lemma everywhere_rel_spec : modal_rel everywhereM = everywhere_rel.
    auto.
   Qed.
  Notation "'!' e" := (everywhereM e) (at level 30, right associativity) : pred_scope. 
  Opaque everywhereM.

  Definition sub (t1 t2:pred) := []!(t1 --> t2).
  Notation "P '>=>' Q" := (sub P Q) (at level 60, no associativity) : pred_scope.
  Definition equ (t1 t2:pred) := (t1 >=> t2) && (t2 >=> t1).
  Notation "P '<=>' Q" := (equ P Q) (at level 60, no associativity) : pred_scope.

  Fixpoint composeN (A: Type) (n: nat) (F: A -> A) (x: A) : A :=
   match n with O => x | S n' => F (composeN A n' F x) end.
  Implicit Arguments composeN.

  Definition rec (F: pred -> pred) : pred :=
    fun w => composeN (S (world_ttl w)) F bot w.

  Definition contractive  (F : pred -> pred) : Prop := forall t1 t2,
     |> (t1 <=> t2) |-- (F t1) <=> (F t2).

  Definition nonexpansive (F : pred -> pred) : Prop := forall t1 t2,
     (t1 <=> t2) |-- (F t1) <=> (F t2).

  Lemma laterM_trans : transitive _ (modal_rel laterM).
    rewrite laterM_rel_spec.
    unfold transitive, later_rel; intros.
    destruct H; destruct H0.
    exists (x0 + (S x1)).
    revert x y z x1 H H0.
    induction x0; simpl in *; intros.

    destruct (tick_world x); try discriminate.
    replace w with y; [ auto | congruence ].
    
    destruct (tick_world x); try discriminate.
    eapply IHx0.
    apply H.
    apply H0.    
   Qed.

  Hint Resolve laterM_trans.

  Lemma loeb_rule: forall t,
      (|>t |-- t) -> (|-- t).

    rewrite laterM_spec; unfold later, derives; simpl; intros.
    cut (forall w, appPosMode later_rel t w -> t w);
       [ | intros; apply H; intuition ].
    clear H H0.
    unfold appPosMode, later_rel.
    intro H.
    cut (forall n w, n = world_ttl w -> t w).
    intro H1; apply H1 with (world_ttl w); auto.
    intro n; pattern n; apply CompleteInduction; clear n.
    intros.
    apply H.
    intros.
    apply H0 with (world_ttl w'); subst; auto.
    destruct H2.
    apply tick_world_n_lt with x; auto.
   Qed.

  Lemma generalized_loeb_rule: forall s t,  
      s |-- |> s ->
      [ s :: (|> t) :: nil ] |-- t  ->
      s |-- t.

    intros.
    cut_with (s --> t); [  apply implies_elim; assumed | ].
    apply derives_axiom.
    apply loeb_rule.
    intro_d.
    cut_with (|> s --> |> t);
      [ | weaken_with modal_K ].
    cut_with (|> s); [ | weaken_with H ].
    cut_with (|> t); [ | move_at 1; apply implies_elim; assumed ].
    weaken_with H0.
   Qed.

  Lemma nec_rel_refl : reflexive _ nec_rel.
    unfold nec_rel; red; intros.
    exists 0; simpl; auto.
   Qed.

  Lemma nec_rel_trans : transitive _ nec_rel.
    unfold transitive, nec_rel; intros.
    destruct H; destruct H0.
    exists (x0 + x1).
    revert x y z x1 H H0.
    induction x0; simpl in *; intros; auto.
    replace x with y; [ auto | congruence ].
    
    destruct (tick_world x); try discriminate.
    eapply IHx0.
    apply H.
    apply H0.    
   Qed.

  Lemma nec_rel_confluent : confluent nec_rel.
    red; intros.
    destruct H as [n1 ?].
    destruct H0 as [n2 ?].
    set (n:=max n1 n2).
    case_eq (tick_world_n n w); intros.
    exists w0; intuition.
    destruct (tick_n_split n n1 (n - n1) w w0); intuition.
    cut(n1 <= n).
    omega.
    unfold n; apply le_max_l.
    replace x with w1 in H4; [ | congruence ].
    exists (n-n1); auto.
    destruct (tick_n_split n n2 (n - n2) w w0); intuition.
    cut(n2 <= n).
    omega.
    unfold n; apply le_max_r.
    replace x with w2 in H4; [ | congruence ].
    exists (n-n2); auto.

    subst n.
    destruct (max_dec n1 n2);
      rewrite e in H1.
    rewrite H in H1; discriminate.
    rewrite H0 in H1; discriminate.
   Qed.

  Lemma necM_refl : reflexive _ (modal_rel necM).
    rewrite necM_rel_spec.
    apply nec_rel_refl.
   Qed.

  Lemma necM_trans : transitive _ (modal_rel necM).
    rewrite necM_rel_spec.
    apply nec_rel_trans.
   Qed.

  Lemma necM_confluent : confluent (modal_rel necM).
    rewrite necM_rel_spec.
    apply nec_rel_confluent.
   Qed.

  Lemma eventuallyM_refl : reflexive _ (modal_rel eventuallyM).
    rewrite eventually_rel_spec.
    apply nec_rel_refl.
   Qed.    

  Lemma eventuallyM_trans : transitive _ (modal_rel eventuallyM).
    rewrite eventually_rel_spec.
    apply nec_rel_trans.
   Qed.

  Lemma eventuallyM_confluent : confluent (modal_rel eventuallyM).
    rewrite eventually_rel_spec.
    apply nec_rel_confluent.
   Qed.


  Hint Resolve necM_refl necM_trans necM_confluent.

  Lemma nec_now : forall t,
    []t |-- t.

    simpl; intros; rewrite necM_spec; unfold nec.
    unfold derives, appPosMode, nec_rel; simpl; intuition.
    apply H0.
    exists 0; simpl; auto.
   Qed.
  
  Lemma nec_later : forall t,
    []t |-- |> t.

    simpl; intros; rewrite necM_spec; unfold nec.
    unfold derives, appPosMode, nec_rel; simpl; intuition.
    rewrite laterM_spec; unfold later, appPosMode, later_rel.
    intros; apply H0; auto with arith.
    destruct H.
    exists (S x); auto.
   Qed.

  Lemma nec_split : forall G t,
    [ G ] |-- t ->
    [ G ] |-- (|> t) ->
    [ G ] |-- ([] t).
  
    simpl; intros; rewrite necM_spec; unfold nec.
    unfold derives, appPosMode, nec_rel; simpl; intuition.
    destruct H2.
    destruct x.
    simpl in H2.
    replace w' with w; [ | congruence ].
    apply H; auto.
    rewrite laterM_spec in H0.
    unfold derives, later, appPosMode, later_rel in H0.
    apply H0 with w; auto.
    exists x; auto.
   Qed.

  Hint Resolve nec_now nec_later nec_split.

  Lemma nec_now_later : forall t,
    []t = t && |>t.

    intuition.
   Qed.

  Lemma later_nec_later1 : forall t,
    |>[] t |-- |> t.

    intuition.
   Qed.  

  Lemma later_nec_later2 : forall t,
    |> t |-- |>[] t.

    intros.
    cut_with (|>|>t); auto.
    coalesce_modal; auto.
   Qed.

  Lemma later_nec_later : forall t,
    |>[]t = |>t.

    intros; apply derives_equal.
    apply later_nec_later1.
    apply later_nec_later2.
   Qed.

  Lemma nec_later_later : forall t,
    []|>t = |>t.

    intuition.
   Qed.

  Lemma nec_later_commute : forall t,
    []|> t = |>[] t.

    intros; rewrite nec_later_later; rewrite later_nec_later; auto.
   Qed.

  Lemma everywhereM_refl :
     reflexive _ (modal_rel everywhereM).

    rewrite everywhere_rel_spec.
    unfold everywhere_rel, reflexive.
    auto.
   Qed.

  Lemma everywhereM_trans : 
     transitive _ (modal_rel everywhereM).

    rewrite everywhere_rel_spec.
    unfold everywhere_rel, transitive.
    intros; rewrite H; auto.
   Qed.

  Lemma everywhereM_sym :
     symmetric _ (modal_rel everywhereM).

    rewrite everywhere_rel_spec.
    unfold everywhere_rel, symmetric.
    intros; symmetry; auto.
   Qed.

  Hint Resolve everywhereM_refl everywhereM_trans everywhereM_sym.

  Lemma every_later_commute : forall p,
    !|>p = |>!p.
  
    intros.
    repeat rewrite everywhere_spec.
    repeat rewrite laterM_spec.
    unfold everywhere, later.
    unfold appPosMode, everywhere_rel.
    apply derives_equal.
    unfold derives; simpl; intuition.
    destruct H as [n Hn].
    destruct (untick_n (S n) w'0) as [z Hz].
    apply H0 with z; auto.
    eapply tick_congruence2.
    apply Hn.
    apply Hz.
    auto.
    exists n; auto.

    unfold derives; simpl; intuition.
    destruct H2 as [n Hn].
    destruct (tick_n_succeeds (S n) w) as [z Hz].
    rewrite H.
    rewrite (tick_n_ttl (S n) w' w'0); auto with arith.
    eapply H0.
    exists n.
    apply Hz.
    eapply tick_congruence1.
    apply Hz.
    apply Hn.
    auto.
   Qed.

  Lemma every_nec_commute : forall p,
    ![]p = []!p.

    intros; repeat rewrite nec_now_later.
    rewrite pmodal_conj; auto.
    rewrite every_later_commute.
    auto.
   Qed.

  Lemma sub_refl : forall t:pred, |-- t >=> t.
    unfold sub; intros; auto.
   Qed.

  Lemma sub_trans : forall t1 t2 t3,
    [ t1 >=> t2 :: t2 >=> t3 :: nil ] |-- t1 >=> t3.

    unfold sub; intros.
    coalesce_modal.
    coalesce_modal.
    intro_d.
    apply_at 2.
    apply_at 1.
    assumed.
   Qed.

  Lemma equ_refl : forall t, |-- t <=> t.
    unfold equ; intros.
    split_d; apply sub_refl.
   Qed.

  Lemma equ_trans : forall t1 t2 t3,
    [ t1 <=> t2 :: t2 <=> t3 :: nil ] |-- t1 <=> t3.

    unfold equ; intros.
    elim_at 0.
    elim_at 2.
    split_d.
    apply_d sub_trans.
    exact_at 2.
    assumed.
    apply_d sub_trans.
    exact_at 1.
    assumed.
   Qed.

  Lemma equ_sym : forall t1 t2, t1 <=> t2 |-- t2 <=> t1.
    unfold equ; intros.
    elim_at 0; split_d; assumed.
   Qed.

  Lemma sub_antisym : forall t1 t2,
    [ t1 >=> t2 :: t2 >=> t1 :: nil ] |-- t1 <=> t2.

    unfold equ; intros; split_d; assumed.
   Qed.

  Lemma sub_project_herenow : forall p q,
    [ p >=> q :: p :: nil ] |-- q.

    intros.
    unfold sub.
    cut_with (p --> q).
    elim_at 0; assumed.
    apply_d (modal_T _ everywhereM).
    apply_d (modal_T _ necM).
    assumed.
   Qed.

  Lemma sub_implies : forall t1 t2 w,
     (t1 >=> t2) w ->
     (t1 w -> t2 w).

    intros.
    refine (sub_project_herenow t1 t2 _ _).
    simpl; intuition.
   Qed.

  Lemma equ_implies : forall t1 t2 w,
     (t1 <=> t2) w ->
     (t1 w -> t2 w).

    intros t1 t2 w H.
    apply sub_implies.
    red in H; destruct H; auto.
   Qed.   

  Lemma equ_equal : forall t1 t2,
     |-- t1 <=> t2 ->
     t1 = t2.

    intros; apply derives_equal.
    cut_with (t1 <=> t2); [ | apply_d H ].
    unfold equ; elim_at 0.
    apply_d sub_project_herenow.
    exact_at 0.
    assumed.    
    cut_with (t1 <=> t2); [ | apply_d H ].
    unfold equ; elim_at 0.
    apply_d sub_project_herenow.
    exact_at 1.
    assumed. 
   Qed.

  Hint Resolve sub_refl equ_refl.

  Lemma equ_elim : forall p q G,
    [ G ] |-- q ->
    [ p <=> q :: G ] |-- p.

    intros.
    cut_with q; auto.
    unfold equ.
    elim_at 1.
    clear_at 0.
    apply_d sub_project_herenow.
    exact_at 0.
    auto.
   Qed.

  Lemma flip_equ : forall p q G t,
    [ p <=> q :: G ] |-- t ->
    [ q <=> p :: G ] |-- t.

    intros.
    cut_with (p <=> q).
    weaken_with H.
    apply_d equ_sym; auto.
   Qed.

  Lemma later_lift_sub : forall p q,
    |> (p >=> q) |-- |>p >=> |> q.

    intros.
    unfold sub.
    rewrite <- nec_later_commute.
    rewrite <- every_later_commute.
    do 2 apply derives_modal.
    apply modal_K; auto.
   Qed.

  Lemma equ_sub : forall p q,
    p <=> q |-- p >=> q.

    intros; unfold equ; intuition_d.
   Qed.

  Lemma sub_inter_right : forall x p q G,
    [ G ] |-- x >=> p ->
    [ G ] |-- x >=> q ->
    [ G ] |-- x >=> p && q.

    intros.
    cut_with (x >=> p); auto.
    cut_with (x >=> q); auto.
    weaken_to (x >=> p :: x >=> q :: nil).
    unfold sub; do 2 coalesce_modal.
    intuition_d.
    apply_at 1; auto.
    apply_at 2; auto.
   Qed.

  Lemma sub_inter_left1 : forall x p q G,
    [ G ] |-- p >=> x ->
    [ G ] |-- p && q >=> x.

    intros.
    apply_d sub_trans; [ | apply H ].
    apply derives_axiom.
    unfold sub; repeat (apply modal_N; auto).
   Qed.

  Lemma sub_modal : forall z (M:modality z) p q,
     |-- p >=> q ->
     |-- M p >=> M q.

    intros.
    unfold sub.
    do 2 (apply modal_N; auto).
    intro_d; apply derives_modal.
    apply_d sub_project_herenow.
    apply_d H.
    auto.
   Qed.

  Lemma sub_inter_left2 : forall x p q G,
    [ G ] |-- q >=> x ->
    [ G ] |-- p && q >=> x.

    intros.
    apply_d sub_trans; [ | apply H ].
    apply derives_axiom.
    unfold sub; repeat (apply modal_N; auto).
   Qed.

  Lemma contractive_nonexpansive: forall F,
     contractive F -> nonexpansive F.

     unfold contractive, nonexpansive; intros.
     apply_d H.
     unfold equ.
     rewrite pmodal_conj; auto.
     elim_at 0; split_d; unfold sub;
       rewrite later_nec_later;
       apply_d nec_later; assumed.
    Qed.

   Lemma contractive_o_nonexpansive: forall F G,
     contractive F ->
     nonexpansive G ->
     contractive (fun x => F (G x)).

     unfold contractive, nonexpansive; intros.
     apply_d H.
     apply derives_modal.
     apply H0.
    Qed.

   Lemma nonexpansive_o_contractive: forall F G, 
     nonexpansive F ->
     contractive G ->
     contractive (fun x => F (G x)).

     unfold contractive, nonexpansive; intros.
     apply_d H.
     apply_d H0.
     assumed.
    Qed.

   Lemma nonexpansive_o_nonexpansive: forall F G, 
     nonexpansive F ->
     nonexpansive G ->
     nonexpansive (fun x => F (G x)).

     unfold nonexpansive; intros.
     apply_d H.
     apply_d H0.  
     assumed.
   Qed.

   Lemma contractive_const: forall t, contractive (fun _ => t).
     red; intros.
     apply derives_axiom.
     apply equ_refl.
    Qed.


   Lemma contractive_later: contractive laterM.
      red; intros.
      unfold equ; rewrite pmodal_conj; auto.
      elim_at 0; split_d.
      clear_at 1.
      unfold sub.
      rewrite <- nec_later_commute.
      apply derives_modal.
      rewrite <- every_later_commute.
      apply derives_modal.
      apply modal_K; auto.
      clear_at 0.
      unfold sub.
      rewrite <- nec_later_commute.
      apply derives_modal.
      rewrite <- every_later_commute.
      apply derives_modal.
      apply modal_K; auto.
     Qed.
     
    Lemma nonexpansive_id : nonexpansive (fun x => x).
       red; intros; assumed.
     Qed.

    Lemma sub_later : forall t1 t2,
      t1 >=> t2 |-- |>t1 >=> |>t2.

      intros.
      unfold sub.
      cut_with ([]!|>(t1 --> t2)).
      apply derives_weaken_head.
      apply derives_modal.
      apply derives_modal.
      apply modal_K; auto.
      rewrite every_later_commute.
      rewrite nec_later_later.
      apply nec_later.
     Qed.

    Lemma sub_nec : forall t1 t2,
      t1 >=> t2 |-- []t1 >=> []t2.

      intros.
      unfold sub.
      cut_with ([]![](t1 --> t2)).
      apply derives_weaken_head.
      do 2 apply derives_modal.
      apply modal_K.
      auto.
      rewrite every_nec_commute; auto.
     Qed.

    Lemma sub_every : forall t1 t2,
      t1 >=> t2 |-- !t1 >=> !t2.

      intros.
      unfold sub.
      cut_with ([]!!(t1 --> t2)).
      apply derives_weaken_head.
      do 2 apply derives_modal.
      apply modal_K; auto.
      rewrite modal_idem; auto.
     Qed.

    Lemma sub_top : forall t, |-- t >=> top. 
      intros; unfold sub; auto.
     Qed.

   Lemma sub_bot : forall t, |-- bot >=> t.
      intros; unfold sub; auto.
     Qed.

   Lemma sub_conj : forall a a' b b',
     [ a >=> a' :: b >=> b' :: nil ] |-- a && b >=> a' && b'.

     intros; unfold sub.
     coalesce_modal.
     coalesce_modal.
     intro_d.
     elim_at 0; split_d.
     apply_at 2; assumed.
     apply_at 3; assumed.
    Qed.

   Lemma sub_disj : forall a a' b b',
     [ a >=> a' :: b >=> b' :: nil ] |-- a || b >=> a' || b'.

     intros; unfold sub.
     do 2 coalesce_modal.
     intro_d.
     elim_at 0.
     left_d;  apply_at 1; assumed.
     right_d; apply_at 2; assumed.
    Qed.

   Lemma sub_forall : forall A f f',
     All_ a:A, f a >=> f' a |-- all f >=> all f'.
     
     unfold sub; intros.
     rewrite <- pmodal_forall; auto.
     rewrite <- pmodal_forall; auto.
     apply derives_modal.
     apply derives_modal.
     intros_d.
     apply_at 1.
     apply_at 0.
    Qed.

   Lemma sub_exist : forall A f f',
     All_ a:A, f a >=> f' a |-- ex f >=> ex f'.

     intros.
     unfold sub.
     rewrite <- pmodal_forall; auto.
     rewrite <- pmodal_forall; auto.
     apply derives_modal.
     apply derives_modal.
     intro_d.
     elim_at 0; intro x.
     exist_d x.
     apply_at 1; assumed.
    Qed.

   Lemma equ_later : forall t1 t2,
      t1 <=> t2 |-- |>t1 <=> |>t2.

      unfold equ; intros; intuition_d.
      apply_d sub_later; auto.
      apply_d sub_later; auto.
     Qed.

   Lemma equ_nec : forall t1 t2,
     t1 <=> t2 |-- []t1 <=> []t2.

     unfold equ; intros; intuition_d.
     apply_d sub_nec; auto.
     apply_d sub_nec; auto.
    Qed.

   Lemma equ_every : forall t1 t2,
     t1 <=> t2 |-- !t1 <=> !t2.

     unfold equ; intros; intuition_d.
     apply_d sub_every; auto.
     apply_d sub_every; auto.
    Qed.

   Lemma equ_conj : forall a a' b b',
     [ a <=> a' :: b <=> b' :: nil ] |-- a && b <=> a' && b'.

     intros; unfold equ.
     elim_at 0.
     elim_at 2.
     split_d; apply_d sub_conj; assumed.
    Qed.

   Lemma equ_disj : forall a a' b b',
     [ a <=> a' :: b <=> b' :: nil ] |-- a || b <=> a' || b'.

     intros; unfold equ.
     elim_at 0.
     elim_at 2.
     split_d; apply_d sub_disj; assumed.
    Qed.

   Lemma equ_forall : forall A f f',
     All_ a:A, f a <=> f' a |-- all f <=> all f'.

     intros.
     unfold equ.
     split_d.
     apply_d sub_forall; intros_d.
     apply (forall_elim _ x).
     elim_at 0; assumed.
     apply_d sub_forall.
     intros_d.
     apply (forall_elim _ x).
     elim_at 0; assumed.
    Qed.

   Lemma equ_exist : forall A f f',
     All_ a:A, f a <=> f' a |-- ex f <=> ex f'.

     intros; unfold equ.
     apply conj_intro.
     apply_d sub_exist.
     intros_d.
     apply (forall_elim _ x).
     elim_at 0; assumed.

     apply_d sub_exist.
     intros_d.
     apply (forall_elim _ x).
     elim_at 0; assumed.
    Qed.

   Lemma nonexpansive_conj : forall f g,
      nonexpansive f ->
      nonexpansive g ->
      nonexpansive (fun x => f x && g x).

     unfold nonexpansive; intros.
     apply_d equ_conj; auto.
    Qed.

   Lemma contractive_conj : forall f g,
      contractive f ->
      contractive g ->
      contractive (fun x => f x && g x).

     unfold contractive; intros.
     apply_d equ_conj; auto.
    Qed.

   Lemma nonexpansive_disj : forall f g,
      nonexpansive f ->
      nonexpansive g ->
      nonexpansive (fun x => f x || g x).

     unfold nonexpansive; intros.
     apply_d equ_disj; auto.
    Qed.

   Lemma contractive_disj : forall f g,
      contractive f ->
      contractive g ->
      contractive (fun x => f x || g x).

     unfold contractive; intros.
     apply_d equ_disj; auto.
    Qed.

   Lemma nonexpansive_forall : forall A (F:pred -> A -> pred),
      (forall a:A, nonexpansive (fun x => F x a)) ->
      nonexpansive (fun x => (all (F x))).

      unfold nonexpansive; intros.
      apply_d equ_forall; auto.
     Qed.

   Lemma contractive_forall : forall A (F:pred -> A -> pred),
      (forall a:A, contractive (fun x => F x a)) ->
      contractive (fun x => (all (F x))).

      unfold contractive; intros.
      apply_d equ_forall; auto.
     Qed.

   Lemma nonexpansive_exist : forall A (F:pred -> A -> pred),
      (forall a:A, nonexpansive (fun x => F x a)) ->
      nonexpansive (fun x => (ex (F x))).

      unfold nonexpansive; intros.
      apply_d equ_exist; auto.
     Qed.

   Lemma contractive_exist : forall A (F:pred -> A -> pred),
      (forall a:A, contractive (fun x => F x a)) ->
      contractive (fun x => (ex (F x))).

      unfold contractive; intros.
      apply_d equ_exist; auto.
     Qed.

   Lemma composeN_outer: forall n F t,
     |-- (F (composeN n F t)) <=> composeN (S n) F t.

     intros; simpl; apply equ_refl.
   Qed.

   Lemma composeN_inner: forall n F t, 
     nonexpansive F ->
     |-- (composeN n F (F t)) <=> composeN (S n) F t.

     intros; red in H.
     induction n; simpl.
     apply equ_refl.
     apply_d H.
     auto.
    Qed.

   Lemma nonexpansive_composeN_nonexpansive: forall n F,
     nonexpansive F ->
     nonexpansive (composeN (S n) F).

     intros n F nonexF.
     induction n; auto.
     simpl in *; apply nonexpansive_o_nonexpansive; auto.
    Qed.

   Lemma composeN_split: forall n n1 n2 F t,
     nonexpansive F ->
     n = n1 + n2 ->
     |-- (composeN n F t) <=> (composeN n1 F (composeN n2 F t)).

     intros n n1; revert n; induction n1.
     simpl; intros; subst n; apply equ_refl.
     simpl; intros; subst n; simpl.
     apply_d H.
     apply IHn1; auto.
    Qed.


(* Lemmas from based on those of the same number in
    "An Indexed Model of Recursive Types..." (Appel, McAllester)
    used to for fold_unfold - Matthew Meola *)

   Lemma lemma15a : forall F w t1 t2,
     contractive F ->
     (|> ((composeN (world_ttl w) F t1) <=> (composeN (world_ttl w) F t2))) w.

     rewrite laterM_spec; unfold later; intros.
     unfold later_rel; red.
     set (n:=world_ttl w) in *.
     cut (n = world_ttl w); [ | auto ].
     clearbody n.
     revert w.
     induction n; intros.

     simpl.
     destruct H1 as [n' Hn'].
     generalize (tick_n_ttl (S n') w w' Hn'); intros.
     rewrite H1 in H0.
     elimtype False; omega.

     simpl.
     apply H; simpl; split_right.
     rewrite laterM_spec; unfold later.
     unfold later_rel; red; intros.
     case_eq (tick_world w); intros.
     apply IHn with w0.
     cut (S n = S (world_ttl w0)); [ congruence | ].
     rewrite H0.
     apply tick_ttl; auto.
     destruct H1 as [n1 Hn1].
     destruct H2 as [n2 Hn2].
     simpl in Hn1.
     rewrite H3 in Hn1.
     exists (n1 + n2).
     replace (S (n1 + n2)) with (n1 + S n2); [ | omega ].
     clear H3.
     revert w0 Hn1.
     induction n1; intros; simpl in Hn1.
     replace w0 with w'; [ auto | congruence ].
     case_eq (tick_world w0); intros.
     rewrite H1 in Hn1.
     simpl.
     rewrite H1.
     apply IHn1; auto.
     rewrite H1 in Hn1; discriminate.
     destruct (tick_succeeds w).
     rewrite <- H0; auto with arith.
     rewrite H3 in H4; discriminate.
    Qed.

   Lemma compose_more : forall F w t1 t2 n1 n2,
     contractive F ->
     n1 > world_ttl w ->
     n2 > world_ttl w ->
     (composeN n1 F t1 <=> composeN n2 F t2) w.

     intros.
     replace n1 with ((1 + world_ttl w) + (n1 - world_ttl w - 1)); [ | omega ].
     replace n2 with ((1 + world_ttl w) + (n2 - world_ttl w - 1)); [ | omega ].
     refine (equ_trans _ (composeN (1+world_ttl w) F (composeN (n1 - world_ttl w - 1) F t1)) _ _ _).
     simpl; intuition.
     apply H; simpl; intuition.
     apply (modal_N _ laterM); simpl; intros; intuition.
     apply composeN_split; auto.
     apply contractive_nonexpansive; auto.
     refine (equ_trans _ (composeN (1+world_ttl w) F (composeN (n2 - world_ttl w - 1) F t2)) _ _ _).
     simpl; intuition.
     apply H; simpl; intuition.
     apply lemma15a; auto.
     apply H; simpl; intuition.
     apply (modal_N _ laterM); simpl; intuition.
     apply_d equ_sym.
     apply composeN_split; auto.
     apply contractive_nonexpansive; auto.
   Qed.  

   Lemma rec_unfold1 : forall w F,
     contractive F ->
     (rec F <=> F (composeN (world_ttl w) F bot)) w.

     unfold equ, sub, conj, rec; intros.
     split; rewrite necM_spec; unfold nec, nec_rel; red; intros.
     rewrite everywhere_spec; unfold everywhere, everywhere_rel, implies; red; intros.
     simpl in H2.
     rewrite <- H1 in H2.
     revert H2.
     destruct H0 as [n Hn].
     destruct n.
     simpl in Hn.
     replace w' with w; [ auto | congruence ].
     apply equ_implies.
     change (((composeN (S (world_ttl w')) F bot) <=> (composeN (S (world_ttl w)) F bot)) w'0).
     apply compose_more; auto.
     rewrite H1; auto.
     rewrite <- H1.
     rewrite (tick_n_ttl (S n) w w'); auto.
     omega.
     
     rewrite everywhere_spec; unfold everywhere, everywhere_rel, implies; red; intros.
     destruct H0 as [n Hn].
     rewrite <- H1.
     destruct n.
     simpl in Hn.
     replace w' with w; [ auto | congruence ].
     simpl.
     revert H2.
     apply equ_implies.
     change (((composeN (S (world_ttl w)) F bot) <=> (composeN (S (world_ttl w')) F bot)) w'0).
     apply compose_more; auto.
     rewrite <- H1; auto.
     rewrite (tick_n_ttl (S n) w w'); auto.
     omega.
     rewrite H1; auto.
    Qed.


   Lemma lemma18a: forall w F,
     contractive F ->
     (|> ((composeN (world_ttl w) F bot) <=> (rec F))) w.

     intros.
     rewrite laterM_spec; unfold later; intros.
     unfold later_rel; red; intros.
     destruct H0 as [n Hn].
     revert w w' Hn.
     induction n; intros.
     simpl in Hn.
     case_eq (tick_world w); intros.
     rewrite H0 in Hn.
     rewrite (tick_ttl w w0); auto.
     replace w' with w0; [ | congruence ].
     refine (equ_sym _ _ _ _); simpl; split_right.     
     apply rec_unfold1; auto.
     rewrite H0 in Hn; discriminate.
     case_eq (tick_world w); intros.
     simpl in Hn.
     rewrite H0 in Hn.
     refine (equ_trans _ (composeN (world_ttl w0) F bot) _ _ _);
      simpl; split_right.
     cut (world_ttl w0 > world_ttl w'); intros.
     apply compose_more; auto.
     rewrite (tick_ttl w w0); auto.
     rewrite (tick_n_ttl (S n) w0 w').
     omega.
     simpl; auto.
     apply IHn; auto.
     simpl in Hn; rewrite H0 in Hn; discriminate.
    Qed.

   Lemma rec_equ : forall F,
     contractive F ->
     |-- (rec F) <=> (F (rec F)).

     unfold derives; intros; clear H0.
     apply (equ_trans (rec F) (F (composeN (world_ttl w) F bot)) (F (rec F))).
     simpl; intuition.
     apply rec_unfold1; auto.
     apply H.
     simpl; intuition.
     apply lemma18a; auto.
    Qed.

   Lemma fold_unfold: forall F,
     contractive F ->
     rec F = F (rec F).

     intros; apply equ_equal; apply rec_equ; auto.
    Qed.

(*
   Lemma adjective_rec : forall F z (M:modality z),
    contractive F ->
    (forall P, |> (P <=> M P) |-- F (M P) <=> M (F P)) ->
    rec F = M (rec F).

    intros.
    apply equ_equal.
    apply loeb_rule.
    in_goal rewrite fold_unfold; auto.
    apply_d equ_trans; [ | apply H0 ].
    apply_d H; assumed.
   Qed.
*)

End GoedelLoebTheory.
