(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import ModelSpec.
Require Import Worlds.
Require Import Logic.

Module ModalTheory (W':WORLDS) :
   MODAL_THEORY with Module W := W'.
  Module W := W'.
  Module LT := LogicTheory W'.

  Export LT.

  Open Local Scope pred_scope.

  Inductive polarity : Set :=
    | pol_pos
    | pol_neg
    | pol_neutral
    .

  Definition polarity_spec
     (z:polarity) (op:pred->pred) (rel:world->world->Prop) :=

     match z with
     | pol_pos => forall (p:pred) w, (forall w', rel w w' -> p w') <-> op p w
     | pol_neg => forall (p:pred) w, (exists w', rel w w' /\ p w') <-> op p w
     | pol_neutral =>
       (forall (p:pred) w, (forall w', rel w w' -> p w') <-> op p w) /\
       (forall (p:pred) w, (exists w', rel w w' /\ p w') <-> op p w)
     end.


  Definition is_pos (z:polarity) :=
      match z with
      | pol_pos => True
      | pol_neg => False
      | pol_neutral => True
      end.

  Definition is_neg (z:polarity) :=
      match z with
      | pol_pos => False
      | pol_neg => True
      | pol_neutral => True
      end.

  Hint Unfold is_pos is_neg.

  Record modality (z:polarity) : Type :=
  { modal_op    :> pred -> pred
  ; modal_rel   :  world -> world -> Prop
  ; modal_spec  :  polarity_spec z modal_op modal_rel
  }.

  Implicit Arguments modal_op.
  Implicit Arguments modal_rel.
  Implicit Arguments modal_spec.

  Definition appPosMode (rel:relation world) (p:pred) : pred :=
    fun w => forall w', rel w w' -> p w'.
  Definition appNegMode (rel:relation world) (p:pred) : pred :=
    fun w => exists w', rel w w' /\ p w'.

  Definition mkPosModality (rel:relation world) : modality pol_pos :=
     Build_modality pol_pos (appPosMode rel) rel
      (fun (p:pred) (w:world) => Logic.conj (fun H => H) (fun H => H)).
  Definition mkNegModality (rel:relation world) : modality pol_neg :=
     Build_modality pol_neg (appNegMode rel) rel
      (fun (p:pred) (w:world) => Logic.conj (fun H => H) (fun H => H)).

  Definition modal_dual zm zn (M:modality zm) (N:modality zn) :=
     is_pos zm /\ is_neg zn /\ modal_rel M = modal_rel N.

  Implicit Arguments modal_dual.

  Lemma polarity_neutral_pos : forall op rel,
      polarity_spec pol_neutral op rel -> polarity_spec pol_pos op rel.

     unfold polarity_spec; simpl; intuition.
    Qed.

  Lemma polarity_neutral_neg : forall op rel,
      polarity_spec pol_neutral op rel -> polarity_spec pol_neg op rel.

     unfold polarity_spec; simpl; intuition.
    Qed.

  Definition neutral_pos_modal (B:modality pol_neutral) : modality pol_pos :=
    Build_modality pol_pos (modal_op B) (modal_rel B) (polarity_neutral_pos _ _ (modal_spec B)).
  Definition neutral_neg_modal (B:modality pol_neutral) : modality pol_neg :=
    Build_modality pol_neg (modal_op B) (modal_rel B) (polarity_neutral_neg _ _ (modal_spec B)).

  Definition serial (A:Type) (R:relation A) :=
    forall x:A, exists y:A, R x y.

  Lemma reflexive_serial : forall A (R:relation A),
    reflexive A R -> serial A R.

    red; intros; exists x; auto.
   Qed.

  Lemma bimodal_dual : forall (B:modality pol_neutral),
      modal_dual B B.

    unfold modal_dual; simpl; auto.
   Qed.

  Definition dual_p (M:modality pol_pos) :=
    mkNegModality (modal_rel M).
  Definition dual_n (N:modality pol_neg) :=
    mkPosModality (modal_rel N).

  Lemma dual_p_is_dual : forall (M:modality pol_pos),
    modal_dual M (dual_p M).

    unfold modal_dual; intros; destruct M; simpl; auto.
   Qed.

  Lemma dual_n_is_dual : forall (N:modality pol_neg),
    modal_dual (dual_n N) N.

    unfold modal_dual; intros; destruct N; simpl; auto.
   Qed.

  Lemma pos_dual_unique : forall zm zn (M1 M2:modality zm) (N:modality zn),
    modal_dual M1 N ->
    modal_dual M2 N ->
    M1 = M2.

    unfold modal_dual; intros.
    destruct M1; destruct M2; simpl in *; intuition.
    subst.
    cut (modal_op1 = modal_op0).
    intro Hop.
    revert modal_spec1.
    rewrite Hop.
    intro modal_spec1.
    replace modal_spec1 with modal_spec0; [ | apply proof_irr ].
    auto.
    refine (extensionality _ _ _); intro p.
    refine (extensionality _ _ _); intro w.
    apply prop_ext.
    destruct zm; simpl in *; try contradiction.
    rewrite <- modal_spec1.
    rewrite <- modal_spec0.
    split; auto.
    destruct modal_spec0.
    destruct modal_spec1.
    rewrite <- H5.
    rewrite <- H3.
    split; auto.
   Qed.

  Lemma neg_dual_unique : forall zm zn (M:modality zm) (N1 N2:modality zn),
    modal_dual M N1 ->
    modal_dual M N2 ->
    N1 = N2.

    unfold modal_dual; intros.
    destruct N1; destruct N2; simpl in *.
    intuition.
    subst.
    cut (modal_op1 = modal_op0).
    intro Hop.
    revert modal_spec1.
    rewrite Hop.
    intro modal_spec1.
    replace modal_spec1 with modal_spec0; [ | apply proof_irr ].
    auto.
    refine (extensionality _ _ _); intro p.
    refine (extensionality _ _ _); intro w.
    apply prop_ext.
    destruct zn; simpl in *; try contradiction.
    rewrite <- modal_spec1.
    rewrite <- modal_spec0.
    split; auto.
    destruct modal_spec1.
    destruct modal_spec0.
    rewrite <- H3.
    rewrite <- H5.
    split; auto.
   Qed.

  Lemma pos_modal_spec : forall z (M:modality z), is_pos z ->
     polarity_spec pol_pos (modal_op M) (modal_rel M).

     intros; destruct z; generalize (modal_spec M); simpl in *.
     auto.
     contradiction.
     intro H1; destruct H1; auto.
    Qed.

  Lemma neg_modal_spec : forall z (M:modality z), is_neg z ->
     polarity_spec pol_neg (modal_op M) (modal_rel M).

     intros; destruct z; generalize (modal_spec M); simpl in *.
     contradiction.
     auto.
     intro H1; destruct H1; auto.
    Qed.


  Section modal_duals.
    Variable nnp : forall P:Prop, ~~P -> P.

    Variables zm zn:polarity.
    Variable M : modality zm.
    Variable N : modality zn.
    Variable Hdual : modal_dual M N.

    Theorem neg_dual : forall p, N p = neg (M (neg p)).

     destruct Hdual as [Hpos [Hneg Hrel]].
     intros; apply derives_equal.
     unfold derives; simpl; intuition.
     do 2 red; intros.
     destruct (neg_modal_spec _ N Hneg p w) as [_ H2].
     destruct H2 as [w' [? ?]]; auto.
     destruct (pos_modal_spec _ M Hpos (neg p) w) as [_ H4].
     cut (bot w'); auto.
     apply H4; auto.
     rewrite Hrel; auto.

     unfold derives; simpl; intuition.
     destruct (neg_modal_spec _ N Hneg p w) as [H2 _].
     apply H2; clear H2.
     apply nnp.
     intro.
     apply H0.
     destruct (pos_modal_spec _ M Hpos (neg p) w) as [H2 _].
     apply H2; clear H2.
     intros.
     repeat red; intros.
     apply H.
     exists w'; intuition.
     rewrite <- Hrel; auto.
    Qed.

   Theorem pos_dual : forall p, M p = neg (N (neg p)).

     destruct Hdual as [Hpos [Hneg Hrel]].
     intros; apply derives_equal.
     unfold derives; simpl; intuition.
     repeat red; intros.
     destruct (neg_modal_spec _ N Hneg (neg p) w) as [_ H2].
     destruct H2 as [w' [? ?]]; auto.
     apply H3.
     destruct (pos_modal_spec _ M Hpos p w) as [_ H4].
     apply H4; auto.
     rewrite Hrel; auto.

     unfold derives; simpl; intuition.
     destruct (pos_modal_spec _ M Hpos p w) as [H3 _].
     apply H3; clear H3; intros.
     apply nnp; intro.
     apply H0.
     destruct (neg_modal_spec _ N Hneg (neg p) w) as [H3 _].
     apply H3; clear H3.
     exists w'; intuition.
     rewrite <- Hrel; auto.
    Qed.
  End modal_duals.


  Lemma modal_N : forall z (M:modality z) p, is_pos z ->
    |-- p ->
    |-- M p.

    unfold derives; simpl; intros.
    destruct (pos_modal_spec z M H p w) as [H2 H3].
    apply H2.
    intros; apply H0; auto.
   Qed.

  Lemma modal_K : forall z (M:modality z) p q, is_pos z ->
    M (p --> q) |-- M p --> M q.

    unfold derives; simpl; intuition.
    red; intro.
    destruct (pos_modal_spec z M H q w) as [H3 H4].
    apply H3.
    intros.
    destruct (pos_modal_spec z M H (p --> q) w) as [H6 H7].
    apply H7; auto.
    destruct (pos_modal_spec z M H p w) as [H8 H9].
    apply H9; auto.
   Qed.

  Lemma nmodal_K : forall zm zn (M:modality zm) (N:modality zn) p q,
     modal_dual M N ->
     M (p --> q) |-- N p --> N q.

     unfold derives; simpl; intuition.
     red; intros.
     destruct H as [Hpos [Hneg Hrel]].
     destruct (neg_modal_spec _ N Hneg p w) as [_ H3].
     destruct H3 as [w' [? ?]]; auto.
     destruct (neg_modal_spec _ N Hneg q w) as [H4 _].
     apply H4; clear H4.
     exists w'.
     intuition.
     destruct (pos_modal_spec _ M Hpos (p --> q) w) as [_ H4].
     apply H4; auto.
     rewrite Hrel; auto.
    Qed.

  Lemma modal_T : forall z (M:modality z) p, is_pos z ->
    reflexive _ (modal_rel M) ->
    M p |-- p.

    unfold derives; simpl; intuition.
    destruct (pos_modal_spec z M H p w) as [H4 H5].
    apply H5; auto.
   Qed.

  Lemma nmodal_T : forall z (N:modality z) p, is_neg z ->
    reflexive _ (modal_rel N) ->
    p |-- N p.

    unfold derives; simpl; intuition.
    destruct (neg_modal_spec z N H p w) as [H4 H5].
    apply H4.
    exists w; auto.
   Qed.

  Lemma modal_4 : forall z (M:modality z) p, is_pos z ->
    transitive _ (modal_rel M) ->
    M p |-- M (M p).

    unfold derives; simpl; intuition.
    destruct (pos_modal_spec z M H (M p) w) as [H4 _].
    apply H4; clear H4; intros.
    destruct (pos_modal_spec z M H p w') as [H4 _].
    apply H4; clear H4; intros.
    destruct (pos_modal_spec z M H p w) as [_ H5].
    apply H5; auto.
    apply H0 with w'; auto.
   Qed.

  Lemma nmodal_4 : forall z (N:modality z) p, is_neg z ->
    transitive _ (modal_rel N) ->
    N (N p) |-- N p.

    unfold derives; simpl; intuition.
    destruct (neg_modal_spec z N H (N p) w) as [_ H4].
    destruct H4 as [w' [? ?]]; auto.
    destruct (neg_modal_spec z N H p w') as [_ H5].
    destruct H5 as [w'' [? ?]]; auto.
    destruct (neg_modal_spec z N H p w) as [H7 _].
    apply H7.
    exists w''; intuition.
    apply H0 with w'; auto.
   Qed.

  Lemma modal_D : forall zm zn (M:modality zm) (N:modality zn) p,
     modal_dual M N ->
     serial _ (modal_rel M) ->

     M p |-- N p.

     intros.
     destruct H as [Hpos [Hneg Hrel]].
     unfold derives; simpl; intuition.
     destruct (neg_modal_spec _ N Hneg p w) as [H3 _].
     apply H3; clear H3.
     destruct (H0 w) as [w' Hw'].
     exists w'; intuition.
     rewrite <- Hrel; auto.
     destruct (pos_modal_spec _ M Hpos p w) as [_ H3].
     apply H3; auto.
    Qed.

  Lemma modal_B : forall zm zn (M:modality zm) (N:modality zn) p,
     modal_dual M N ->
     symmetric _ (modal_rel M) ->

     p |-- M (N p).

     unfold derives; simpl; intuition.
     clear H3.
     destruct H as [Hpos [Hneg H]].
     destruct (pos_modal_spec _ M Hpos (N p) w) as [H3 _].
     apply H3; clear H3.
     intros w' H3.
     destruct (neg_modal_spec _ N Hneg p w') as [H4 _].
     apply H4; clear H4.
     exists w; intuition.
     rewrite <- H.
     apply H0; auto.
    Qed.

  Lemma pmodal_idem : forall z (M:modality z) p, is_pos z ->
     reflexive _ (modal_rel M) ->
     transitive _ (modal_rel M) ->
     M (M p) = M p.

     intros; apply derives_equal.
     apply modal_T; auto.
     apply modal_4; auto.
    Qed.

  Lemma nmodal_idem : forall z (N:modality z) p, is_neg z ->
     reflexive _ (modal_rel N) ->
     transitive _ (modal_rel N) ->
     N (N p) = N p.

     intros; apply derives_equal.
     apply nmodal_4; auto.
     apply nmodal_T; auto.
    Qed.

  Lemma bimodal_idem : forall (B:modality pol_neutral) p,
     transitive _ (modal_rel B) ->
     B (B p) = B p.

     intros; apply derives_equal.
     apply nmodal_4; auto.
     apply modal_4; auto.
    Qed.

  Lemma modal_idem : forall z (M:modality z) p,
     reflexive _ (modal_rel M) ->
     transitive _ (modal_rel M) ->
     M (M p) = M p.

     intro z; destruct z; intros.
     apply pmodal_idem; auto.
     apply nmodal_idem; auto.
     apply pmodal_idem; auto.
    Qed.

  Lemma derives_pmodal : forall z (M:modality z) t1 t2, is_pos z ->
    t1 |-- t2 ->
    M t1 |-- M t2.

    intros.
    apply derives_cut with (M t1 --> M t2).
    apply implies_elim; assumed.
    apply derives_cut with (M (t1 --> t2)).
    apply derives_weaken with (M (t1 --> t2) :: nil).
    red; simpl; intuition.
    apply modal_K; auto.
    apply derives_weaken with (nil : list pred).
    red; simpl; intuition.
    apply modal_N; auto.
   Qed.

  Lemma derives_nmodal : forall z (N:modality z) t1 t2, is_neg z ->
    t1 |-- t2 ->
    N t1 |-- N t2.

    unfold derives; simpl; intuition.
    destruct (neg_modal_spec _ N H t1 w).
    destruct H4 as [w' [? ?]]; auto.
    destruct (neg_modal_spec _ N H t2 w).
    apply H6.
    exists w'; intuition.
   Qed.

  Lemma derives_modal : forall z (M:modality z) t1 t2,
    t1 |-- t2 ->
    M t1 |-- M t2.

    intro z; destruct z; intros.
    apply derives_pmodal; auto.
    apply derives_nmodal; auto.
    apply derives_pmodal; auto.
   Qed.

  Lemma pmodal_distrib_const : forall z (M:modality z) P, is_pos z ->
    const P |-- M (const P).

    intros.
    apply const_elim; intros.
    apply modal_N; auto.
   Qed.

  Lemma pmodal_distrib_const2 : forall z (M:modality z) P, is_pos z ->
    serial _ (modal_rel M) ->
    M (const P) |-- const P.

    intros.
    unfold derives; simpl; intuition.
    destruct (pos_modal_spec _ M H (const P) w).
    destruct (H0 w) as [w' Hw'].
    cut (const P w').
    unfold const; auto.
    apply H4; auto.
   Qed.

  Lemma pmodal_const : forall z (M:modality z) P, is_pos z ->
    serial _ (modal_rel M) ->
    M (const P) = const P.

    intros; apply derives_equal.
    apply pmodal_distrib_const2; auto.
    apply pmodal_distrib_const; auto.
   Qed.

  Lemma nmodal_distrib_const : forall z (N:modality z) P, is_neg z ->
    N (const P) |-- const P.

    intros.
    unfold derives; simpl; intuition.
    destruct (neg_modal_spec _ N H (const P) w).
    destruct H3 as [w' [? ?]]; auto.
   Qed.

  Lemma nmodal_distrib_const2 : forall z (N:modality z) P, is_neg z ->
    serial _ (modal_rel N) ->
    const P |-- N (const P).

    intros.
    unfold derives; simpl; intuition.
    destruct (neg_modal_spec _ N H (const P) w).
    apply H1.
    destruct (H0 w) as [w' Hw'].
    exists w'; auto.
   Qed.

  Lemma nmodal_const : forall z (N:modality z) P, is_neg z ->
    serial _ (modal_rel N) ->
    N (const P) = (const P).

    intros; apply derives_equal.
    apply nmodal_distrib_const; auto.
    apply nmodal_distrib_const2; auto.
   Qed.

  Lemma bimodal_const : forall (B:modality pol_neutral) P,
    B (const P) = const P.

    intros.
    apply derives_equal.
    apply nmodal_distrib_const; auto.
    apply pmodal_distrib_const; auto.
   Qed.

  Lemma modal_const : forall z (M:modality z) P,
    serial _ (modal_rel M) ->
    M (const P) = const P.

    intros; destruct z.
    apply pmodal_const; auto.
    apply nmodal_const; auto.
    apply pmodal_const; auto.
   Qed.

  Lemma pmodal_const_impl : forall z (M:modality z) P q, is_pos z ->
    M (const P --> q) = const P --> M q.

    intros; apply derives_equal;
      unfold derives; simpl; intuition.
    red; intros.
    destruct (pos_modal_spec _ M H q w).
    apply H3; clear H3 H4; intros.
    destruct (pos_modal_spec _ M H (const P --> q) w).
    apply H5; auto.
    red in H1.
    destruct (pos_modal_spec _ M H (const P --> q) w).
    apply H0; clear H0 H3; red; intros.
    destruct (pos_modal_spec _ M H q w).
    apply H5; auto.
   Qed.

  Lemma modal_distrib_conj : forall z (N:modality z) p q,
    N (p && q) |-- N p && N q.

    intros.
    apply conj_intro.
    apply derives_modal.
    apply conj_elim.
    assumed.
    apply derives_modal.
    apply conj_elim.
    assumed.
   Qed.

  Lemma modal_distrib_disj : forall z (N:modality z) p q,
     N p || N q |-- N (p || q).

     intros.
     apply disj_elim.
     apply derives_modal.
     apply disj_intro1.
     assumed.
     apply derives_modal.
     apply disj_intro2.
     assumed.
    Qed.

  Lemma modal_distrib_forall : forall z (M:modality z) A (f:A->pred),
    M (all f) |-- All_ a:A, M (f a).

    intros.
    apply forall_intro; intro x.
    apply derives_modal.
    apply (forall_elim _ x).
    assumed.
   Qed.

  Lemma modal_distrib_exist : forall z (M:modality z) A (f:A->pred),
    Ex_ a:A, M (f a) |-- M (ex f).

    intros.
    apply exist_elim; intro x.
    apply derives_modal.
    apply exist_intro with x.
    assumed.
   Qed.

  Lemma pmodal_distrib_conj2 : forall z (M:modality z) p q, is_pos z ->
    M p && M q |-- M (p && q).

    intros.
    apply conj_elim.
    apply derives_cut with (M (p --> q --> p && q)).
    apply derives_cut with (M p --> M (q --> p && q)).
    apply derives_cut with (M (q --> (p && q))).
    apply derives_cut with (M q --> M (p && q)).
    apply implies_elim; assumed.
    apply derives_weaken with (M (q --> p && q) :: nil).
    red; simpl; intuition.
    apply modal_K; auto.
    apply implies_elim; assumed.
    apply derives_weaken with (M (p --> q --> p && q) :: nil).
    red; simpl; intuition.
    apply modal_K; auto.
    apply derives_weaken with (nil : list pred).
    red; simpl; intuition.
    apply modal_N; auto.
   Qed.

  Lemma pmodal_conj : forall z (M:modality z) p q,
    is_pos z ->
    M (p && q) = M p && M q.

    intros; apply derives_equal.
    apply modal_distrib_conj.
    apply pmodal_distrib_conj2; auto.
   Qed.

  Lemma nmodal_distrib_disj2 :  forall z (N:modality z) p q,
     is_neg z ->
     N (p || q) |-- N p || N q.

     unfold derives; simpl; intuition.
     destruct (neg_modal_spec _ N H (p || q) w).
     destruct H3 as [w' [? ?]]; auto.
     destruct H4.
     left.
     destruct (neg_modal_spec _ N H p w).
     apply H5.
     exists w'; auto.
     right.
     destruct (neg_modal_spec _ N H q w).
     apply H5.
     exists w'; auto.
    Qed.

  Lemma nmodal_disj : forall z (N:modality z) p q,
     is_neg z ->
     N (p || q) = N p || N q.

     intros; apply derives_equal.
     apply nmodal_distrib_disj2; auto.
     apply modal_distrib_disj.
    Qed.

  Lemma pmodal_distrib_forall2 : forall z (M:modality z) A (f:A->pred),
    is_pos z ->
    All_ a:A, M (f a) |-- M (all f).

    unfold derives; simpl; intuition.
    unfold all; intros.
    destruct (pos_modal_spec _ M H (fun v => forall a:A, f a v) w) as [H3 _].
    apply H3; clear H3; intros.
    destruct (pos_modal_spec _ M H (f a) w) as [_ H3].
    apply H3; simpl in *; intuition.
   Qed.

  Lemma pmodal_forall : forall z (M:modality z) A (f:A->pred),
    is_pos z ->
    M (all f) = All_ a:A, M (f a).

    intros; apply derives_equal.
    apply modal_distrib_forall.
    apply pmodal_distrib_forall2; auto.
   Qed.

  Lemma nmodal_distrib_exist2 : forall z (N:modality z) A (f:A->pred),
    is_neg z ->
    N (ex f) |-- Ex_ a:A, N (f a).

    unfold derives; simpl; intuition.
    destruct (neg_modal_spec _ N H (ex f) w).
    destruct H3 as [w' [? ?]]; auto.
    destruct H4.
    exists x.
    destruct (neg_modal_spec _ N H (f x) w).
    apply H5.
    exists w'; auto.
   Qed.

  Lemma nmodal_exist : forall z (N:modality z) A (f:A->pred),
    is_neg z ->
    Ex_ a:A, N (f a) = N (ex f).

    intros; apply derives_equal.
    apply modal_distrib_exist.
    apply nmodal_distrib_exist2; auto.
   Qed.

  Lemma pdrop_modal : forall z (M:modality z) t G, is_pos z ->
     (forall t', In t' G -> t' |-- M t') ->
     [ G ] |-- t ->
     [ G ] |-- M t.

    intros z M t G H; revert t; induction G; intros.
    apply modal_N; auto.
    cut_with (M a --> M t).
    apply implies_elim.
    apply_d (H0 a); simpl; auto.
    apply_d modal_K.
    weaken_to G.
   Qed.

  Definition confluent (R:relation world) := forall w w1 w2,
    R w w1 -> R w w2 -> exists w', R w1 w' /\ R w2 w'.

  Lemma modal_G : forall zm zn (M:modality zm) (N:modality zn) p,
    modal_dual M N ->
    confluent (modal_rel M) ->

    N (M p) |-- M (N p).

    intros; red in H.
    unfold derives; simpl; intuition.
    clear H5.
    destruct (neg_modal_spec _ N H (M p) w).
    destruct H5 as [w1 ?]; clear H2; intuition.
    destruct (pos_modal_spec _ M H1 (N p) w).
    apply H5; clear H5 H7; intros w2 ?.
    destruct (H0 w w1 w2) as [w' ?]; intuition.
    rewrite H3; auto.
    destruct (neg_modal_spec _ N H p w2).
    apply H7; clear H7 H10.
    exists w'; intuition.
    rewrite <- H3; auto.
    destruct (pos_modal_spec _ M H1 p w1).
    apply H10; intuition.
   Qed.

  Lemma neg_pos_conj : forall zm zn (M:modality zm) (N:modality zn) p q,
    modal_dual M N ->
    transitive _ (modal_rel M) ->
    confluent (modal_rel M) ->

    (N (M p)) && (N (M q)) = N (M p && M q).

    intros; red in H; intuition.
    apply derives_equal; [ | apply modal_distrib_conj ].

    apply conj_elim.
    in_goal rewrite <- pmodal_conj; auto.
    unfold derives; simpl; intuition.
    clear H7.
    destruct (neg_modal_spec _ N H (M p) w).
    destruct H7 as [w1 ?]; clear H6; intuition.
    destruct (neg_modal_spec _ N H (M q) w).
    destruct H9 as [w2 ?]; clear H7; intuition.
    destruct (H1 w w1 w2) as [w' ?]; intuition.
    rewrite H4; auto.
    rewrite H4; auto.
    destruct (neg_modal_spec _ N H (M (p && q)) w).
    apply H9; clear H9 H13.
    exists w'; split.
    rewrite <- H4.
    apply H0 with w1; auto.
    rewrite H4; auto.
    destruct (pos_modal_spec _ M H2 (p && q) w').
    apply H9; clear H9 H13; intros.
    split.
    destruct (pos_modal_spec _ M H2 p w1).
    apply H14; clear H13 H14; auto.
    apply H0 with w'; auto.
    destruct (pos_modal_spec _ M H2 q w2).
    apply H14; clear H13 H14; auto.
    apply H0 with w'; auto.
   Qed.

   Lemma elim_pos_modal : forall z (M:modality z) p G t,
    is_pos z ->
    reflexive world (modal_rel M) ->
    [ p :: G ] |-- t ->
    [ M p :: G ] |-- t.

    intros.
    cut_with p.
    weaken_with H1.
    apply derives_weaken_head.
    apply modal_T; auto.
   Qed.


  Lemma dual_implies : forall zm zn (M:modality zm) (N:modality zn) p q,
    modal_dual M N ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->

    M (N p --> M q) = N p --> M q.

    intros; red in H; intuition.
    apply derives_equal.
    apply modal_T; auto.

    unfold derives; simpl; intuition.
    clear H6.
    destruct (pos_modal_spec _ M H2 (N p --> M q) w).
    apply H3; clear H3 H6.
    red; intros.
    destruct (pos_modal_spec _ M H2 q w').
    apply H7; clear H7 H8; intros.
    assert (M q w).
    apply H5.
    destruct (neg_modal_spec _ N H p w').
    generalize (H9 H6); clear H8 H9; intros.
    destruct H8 as [w1 ?]; intuition.
    destruct (neg_modal_spec _ N H p w).
    apply H8; clear H8 H11.
    exists w1; intuition.
    rewrite <- H4.
    apply H1 with w'; auto.
    rewrite H4; auto.
    destruct (pos_modal_spec _ M H2 q w).
    apply H10; clear H9 H10; auto.
    apply H1 with w'; auto.
   Qed.

  Lemma dual_implies2 : forall zm zn (M:modality zm) (N:modality zn) p q,
    modal_dual M N ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->

    N (M p --> N q) = M p --> N q.

    intros.
    destruct H; intuition.
    apply derives_equal;
      [ | apply nmodal_T; auto; rewrite <- H4; auto ].
    unfold derives; simpl; intuition.
    clear H6.
    red; intros.
    destruct (neg_modal_spec _ N H3 (M p --> N q) w).
    generalize (H7 H5); clear H6 H7; intros.
    destruct H6 as [w' ?]; intuition.
    assert (N q w').
    apply H8.
    destruct (pos_modal_spec _ M H p w').
    apply H6; clear H6 H9; intros.
    destruct (pos_modal_spec _ M H p w).
    apply H10; auto.
    apply H1 with w'; auto.
    rewrite H4; auto.

    destruct (neg_modal_spec _ N H3 q w').
    generalize (H10 H6); clear H9 H10; intros.
    destruct H9 as [w1 ?]; intuition.
    destruct (neg_modal_spec _ N H3 q w).
    apply H9; clear H9 H12; intros.
    exists w1; intuition.
    rewrite <- H4.
    apply H1 with w'; auto.
    rewrite H4; auto.
    rewrite H4; auto.
   Qed.

End ModalTheory.
