(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Export Arith.
Require Export List.
Require Export ListSet.
Require Export Omega.
Require Export Relation_Definitions.

Require Export Base.

Module Type WORLD_CARRIER.
  Parameter world : Type.
  Parameter sample_world : world.

  Parameter field : Set.
  Parameter field_enum : field -> nat.
  Axiom field_enum_injective : forall x y,
    field_enum x = field_enum y -> x = y.
  
  (* This parameter needs to be a _transparent_ definition.  It is
     used by some tactics for simplification and it needs to 
     unfold so that computations using it can progress. *)
  Parameter field_eq_dec : forall x y:field, {x=y}+{x<>y}.

  Parameter field_type : field -> Type.
  Parameter w_upd  : forall x:field, field_type x -> world -> world.
  Parameter w_proj : forall x:field, world -> field_type x.

  Axiom proj_upd_same : forall x a w,
     w_proj x (w_upd x a w) = a.

  Axiom upd_proj_same : forall x w,
     w_upd x (w_proj x w) w = w.

  Axiom proj_upd_diff : forall x y a w,
     x <> y ->
     w_proj x (w_upd y a w) = w_proj x w.

  Axiom upd_collapse : forall x a b w,
     w_upd x a (w_upd x b w) =
     w_upd x a w.

  Axiom upd_injective : forall x a a' w w',
     w_upd x a w = w_upd x a' w' -> a = a'.

  Axiom upd_eq : forall x a b w w',
     w_upd x a w = w_upd x a w' ->
     w_upd x b w = w_upd x b w'.

  Axiom upd_commute : forall x y a b w,
     x <> y ->
     w_upd x a (w_upd y b w) =
     w_upd y b (w_upd x a w).
End WORLD_CARRIER.

Module Type WORLD_CARRIER_WITH_HOLE.
  Parameter world_functor : Type -> Type.
  Parameter sample_world : forall A, A -> world_functor A.

  Parameter field : Set.
  Parameter field_enum : field -> nat.
  Axiom field_enum_injective : forall x y,
    field_enum x = field_enum y -> x = y.
  
  Parameter field_eq_dec : forall x y:field, {x=y}+{x<>y}.

  Parameter field_type : field -> Type.
  Parameter w_upd  : forall A (x:field), field_type x -> world_functor A -> world_functor A.
  Parameter w_proj : forall A (x:field), world_functor A -> field_type x.

  Axiom proj_upd_same : forall A x a w,
     w_proj A x (w_upd A x a w) = a.

  Axiom upd_proj_same : forall A x w,
     w_upd A x (w_proj A x w) w = w.

  Axiom proj_upd_diff : forall A x y a w,
     x <> y ->
     w_proj A x (w_upd A y a w) = w_proj A x w.

  Axiom upd_collapse : forall A x a b w,
     w_upd A x a (w_upd A x b w) =
     w_upd A x a w.

  Axiom upd_injective : forall A x a a' w w',
     w_upd A x a w = w_upd A x a' w' -> a = a'.

  Axiom upd_eq : forall A x a b w w',
     w_upd A x a w = w_upd A x a w' ->
     w_upd A x b w = w_upd A x b w'.

  Axiom upd_commute : forall A x y a b w,
     x <> y ->
     w_upd A x a (w_upd A y b w) =
     w_upd A y b (w_upd A x a w).
  
  Parameter get_hole : forall A, world_functor A -> A.
  Parameter set_hole : forall A B, B -> world_functor A -> world_functor B.

  Axiom get_set_hole : forall A B x w,
     get_hole B (set_hole A B x w) = x.

  Axiom set_get_hole_same : forall A w,
     set_hole A A (get_hole A w) w = w.

  Axiom set_hole_collapse : forall A B C x y w,
     set_hole B C y (set_hole A B x w) =
     set_hole A C y w.

  Axiom set_hole_injective : forall A B x x' w w',
     set_hole A B x w = set_hole A B x' w' -> x = x'.

  Axiom set_hole_eq : forall A B x y w w',
     set_hole A B x w = set_hole A B x w' ->
     set_hole A B y w = set_hole A B y w'.

  Axiom set_hole_commute : forall A B x a b w,
     w_upd B x a (set_hole A B b w) =
     set_hole A B b (w_upd A x a w).

  Axiom get_hole_upd_collapse : forall A x a w,
    get_hole A (w_upd A x a w) = get_hole A w.

End WORLD_CARRIER_WITH_HOLE.

Module Type WORLD_TICK.
  Declare Module WC:WORLD_CARRIER.
  
  Import WC.

  Parameter world_ttl : world -> nat.
  Parameter tick_world : world -> option world.

  Axiom tick_ttl : forall w w',
     tick_world w = Some w' -> 
     world_ttl w = S (world_ttl w').

  Axiom tick_succeeds : forall w,
     world_ttl w > 0 ->
     exists w', tick_world w = Some w'.

  Axiom untick : forall w', exists w, tick_world w = Some w'.

  Axiom tick_inj : forall w1 w2 w,
     tick_world w1 = Some w ->
     tick_world w2 = Some w ->
     w1 = w2.

  Axiom tick_upd_commute : forall w1 w2 x a,
    (exists w', tick_world w1 = Some w' /\ w_upd x a w' = w2) <->
    (tick_world (w_upd x a w1) = Some w2).

End WORLD_TICK.

Module Type WORLDS.
  Declare Module WC:WORLD_CARRIER.
  Declare Module WT:WORLD_TICK
    with Module WC:=WC.

  Export WC.
  Export WT.
End WORLDS.

Module Type WORLD_STEP.
  Declare Module W:WORLDS.

  Import W.

  Parameter prog_desc : Type.
  Parameter prog_cont : Type.

  Parameter step : prog_desc -> prog_cont -> world -> prog_cont -> world -> Prop.
  Parameter halted : prog_desc -> prog_cont -> world -> Prop.

  Axiom step_deterministic : forall p k k1 k2 w w1 w2,
     step p k w k1 w1 -> step p k w k2 w2 -> k1 = k2 /\ w1 = w2.

  Axiom step_preserves_ttl : forall p k k' w w',
     step p k w k' w' -> world_ttl w = world_ttl w'.

  Axiom halted_no_step : forall p  k k' w w',
     halted p k w -> step p k w k' w' -> False.

  Axiom step_tick_commute : forall p k k' w w',
     (exists w1, step p k w k' w1 /\ tick_world w1 = Some w') <->
     (exists w2, tick_world w = Some w2 /\ step p k w2 k' w').

  Axiom halt_tick : forall p k w w',
     tick_world w = Some w' ->
     (halted p k w <-> halted p k w').
End WORLD_STEP.

Module Type WORLD_THEORY.
  Declare Module W:WORLDS.

  Export W.

  Fixpoint tick_world_n (n:nat) (w:world) { struct n } :=
    match n with
    | 0 => Some w
    | S n' =>
        match tick_world w with
        | Some w' => tick_world_n n' w'
        | None    => None
        end
    end.

  Inductive upd : Type :=
  | U : forall x:field, field_type x -> upd.

  Definition subst := list upd.

  Fixpoint upd_set (z:subst) : list field :=
   match z with
   | nil => nil
   | U x _ :: z' => x :: upd_set z'
   end.

  Fixpoint world_update (z:subst) : world -> world :=
   match z with
   | nil         => fun w => w
   | U x a :: z' => fun w => w_upd x a (world_update z' w)
   end.

  Axiom world_ttl_noninterference : forall w x a,
    world_ttl w = world_ttl (w_upd x a w).

  Axiom w_upd_noninterference : forall w w' x,
    tick_world w = Some w' ->
    w_proj x w = w_proj x w'.

  Axiom tick_world_n_lt : forall n w w',
     tick_world_n (S n) w = Some w' ->
     world_ttl w' < world_ttl w.

  Axiom tick_congruence1 : forall n w1 w1' w2 w2',
     tick_world_n n w1 = Some w1' ->
     tick_world_n n w2 = Some w2' ->
     world_ttl w1 = world_ttl w2 ->
     world_ttl w1' = world_ttl w2'.

  Axiom tick_congruence2 : forall n w1 w1' w2 w2',
     tick_world_n n w1 = Some w1' ->
     tick_world_n n w2 = Some w2' ->
     world_ttl w1' = world_ttl w2' ->
     world_ttl w1 = world_ttl w2.

  Axiom tick_n_ttl : forall n w w',
     tick_world_n n w = Some w' ->
     world_ttl w = world_ttl w' + n.

  Axiom untick_n : forall n w', exists w, tick_world_n n w = Some w'.

  Axiom tick_n_succeeds : forall n w,
     world_ttl w >= n ->
     exists w', tick_world_n n w = Some w'.

  Axiom tick_n_inj : forall n w1 w2 w,
     tick_world_n n w1 = Some w -> tick_world_n n w2 = Some w -> w1 = w2.

  Axiom tick_n_split : forall n n1 n2 w1 w2,
    n = n1 + n2 ->
    tick_world_n n w1 = Some w2 ->
    exists w',
      tick_world_n n1 w1 = Some w' /\
      tick_world_n n2 w' = Some w2.

  Axiom upd_set_app : forall z1 z2,
    upd_set (z1++z2) = upd_set z1 ++ upd_set z2.

  Axiom world_update_app : forall w z1 z2,
    world_update (z1++z2) w = world_update z1 (world_update z2 w).

  Fixpoint insert_update (x:field) (a:field_type x) (z:subst) { struct z } : subst :=
    match z with
    | nil         => U x a :: nil 
    | U y b :: l' => if (le_lt_dec (field_enum y) (field_enum x))
                          then if W.WC.field_eq_dec x y
                                  then U x a :: l'
                                  else U y b :: (insert_update x a l')
                          else U x a :: U y b :: l'

  end.

  Fixpoint normalize_update (z:subst) : subst :=
    match z with
    | nil         => nil
    | U x a :: z' => insert_update x a (normalize_update z')
    end.

  Axiom change_world_insert : forall w l x a,
       world_update (U x a :: l) w =
       world_update (insert_update x a l) w.

  Axiom world_update_normalize : forall w l,
     world_update l w = world_update (normalize_update l) w.

  Definition w_upd_norm := fun l w => world_update (normalize_update l) w.

  Axiom world_update_normalize_rewrite : forall w l,
     world_update l w = w_upd_norm l w.

  Axiom proj_subst_notin : forall x z l w,
    incl (upd_set z) l ->
    ~In x l ->
    w_proj x (world_update z w) =
    w_proj x w.

  Axiom commute_single : forall z1 z2 w
    (x:field) (a:field_type x),
    (~In x (upd_set z1)) ->
    world_update (U x a :: z1++z2) w = 
    world_update (z1 ++ (U x a):: z2) w.

  Axiom commute_subst : forall z1 z2 w,
    (forall x, In x (upd_set z1) -> In x (upd_set z2) -> False) ->
    world_update (z1++z2) w = world_update (z2++z1) w.

  Fixpoint drop_sel (x:field) (z:subst) {struct z}: subst :=
    match z with
    | nil => nil
    | U y a :: z' =>
       if W.WC.field_eq_dec x y
         then drop_sel x z'
         else U y a :: drop_sel x z'
    end.

  Axiom drop_sel_notin : forall x z,
    ~In x (upd_set (drop_sel x z)).

  Axiom subst_drop_app : forall x z1 z2,
    drop_sel x (z1++z2) = drop_sel x z1 ++ drop_sel x z2.

  Axiom subst_drop_head : forall x a z w,
    world_update (U x a :: z) w =
    world_update (U x a :: drop_sel x z) w.

  Fixpoint float_sel' (x:field) (head:subst) (z:subst) {struct z} : subst :=
    match z with
    | nil => head
    | U y a :: z' =>
       if W.WC.field_eq_dec x y
          then U y a :: head ++ (drop_sel x z')
          else float_sel' x (head ++ (U y a::nil)) z'
    end.

  Definition float_sel (x:field) := float_sel' x nil.

  Axiom float_head_out : forall x y a head z w,
     x <> y -> 
     world_update (float_sel' x (U y a :: head) z) w =
     world_update (U y a :: float_sel' x head z) w.

  Axiom float_sel_eq : forall w z x,
    world_update z w = world_update (float_sel x z) w.

  Axiom float_sel'_notin : forall z2 z1 z' x a,
    ~In x (upd_set z1) -> 
    float_sel' x z1 z2 = U x a :: z' ->
    ~In x (upd_set z').
  
  Axiom float_sel_notin : forall z z' x a,
    float_sel x z = U x a :: z' ->
    ~In x (upd_set z').

  Fixpoint filter_subst (l:list field) (z:subst) { struct z } : subst :=
   match z with
   | nil => nil
   | U x a :: z' =>
     if In_dec field_eq_dec x l
        then filter_subst l z'
        else U x a :: filter_subst l z'
   end.

  Axiom filter_subst_notin : forall x l z,
    In x l -> ~In x (upd_set (filter_subst l z)).

  Axiom filter_subst_empty : forall l z,
    incl (upd_set z) l ->
    filter_subst l z = nil.

  Axiom proj_update_filter : forall x z l w,
     ~ In x l ->
     w_proj x (world_update z w) =
     w_proj x (world_update (filter_subst l z) w).

  Axiom world_update_drop : forall z1 z2 w,
    incl (upd_set z2) (upd_set z1) ->
    world_update z1 (world_update z2 w) = world_update z1 w.

  Axiom world_update_dup : forall z w,
    world_update z w = world_update z (world_update z w).

End WORLD_THEORY.


Module Type LOGIC_THEORY.
  Declare Module W:WORLDS.
  Declare Module WT:WORLD_THEORY
    with Module W:=W.

  Export WT.

  Definition pred := world -> Prop.

  Fixpoint all_true (l:list pred) (w:world) { struct l } : Prop :=
    match l with
    | nil => True
    | x :: l' => x w /\ all_true l' w
    end.
    
  Definition derives (l:list pred) (p:pred) := 
      forall w, all_true l w -> p w.

  Notation "|--  t" := (derives nil t) (at level 80, no associativity).
  Notation " t1  |--  t2" := (derives (t1::nil) t2) (at level 80, no associativity).
  Notation " [  G  ]  |--  t" := (derives G t)
        (at level 80, no associativity).

  (* Primitives *)
  Definition bot: pred := fun _ => False.
  Definition top: pred := fun _ => True.
  Definition disj (t1 t2: pred) : pred := fun w => t1 w \/ t2 w.
  Definition conj (t1 t2: pred) : pred := fun w => t1 w /\ t2 w.
  Definition implies (t1 t2: pred) : pred := fun w => t1 w -> t2 w.
  Definition ex (A: Type) (F: A -> pred) := fun w => exists x:A, F x w.
  Implicit Arguments ex.
  Definition all (A: Type) (F: A -> pred) := fun w => forall x:A, F x w.
  Implicit Arguments all.
  Definition neg (p:pred) := implies p bot.
  Definition const (P:Prop) : pred := fun _ => P.
  Definition projecting (x:field) (a:field_type x) : pred :=
     fun w => w_proj x w = a.
  Definition field_pred (x:field) (P:field_type x -> pred) : pred :=
     fun w => P (w_proj x w) w.

  Delimit Scope pred_scope with pred.

  Notation "'Ex_'  x ':' T ',' P " := (ex (fun x:T => P%pred)) (at level 65, x at level 99) : pred_scope.
  Notation "'All_'  x ':' T  ',' P " := (all (fun x:T => P%pred)) (at level 65, x at level 99) : pred_scope.
  Infix "||" := disj (at level 50, left associativity) : pred_scope.
  Infix "&&" := conj (at level 40, left associativity) : pred_scope.
  Notation "P '-->' Q" := (implies P Q) (at level 60, right associativity) : pred_scope.

  Open Local Scope pred_scope.

  Tactic Notation "in_goal" tactic(tac) :=
    let Z := fresh "Z" in
      match goal with [ |- derives ?G ?t ] =>
        set (Z:=G); tac; subst Z
      end.

  Axiom derives_equal : forall t1 t2,
    t1 |-- t2 ->
    t2 |-- t1 ->
    t1 = t2.

  Axiom equal_derives : forall t1 t2,
    t1 = t2 -> t1 |-- t2.

  Axiom eta_equal : forall t:pred,
    t = (fun w:world => t w).

  (* Structural rules for derives judgements *)

  Axiom derives_weaken : forall G1 G2 t,
    (forall P, In P G1 -> In P G2) ->
    [ G1 ] |-- t ->
    [ G2 ] |-- t.

  Axiom derives_weaken_head : forall P G t,
    [ P :: nil ] |-- t ->
    [ P :: G   ] |-- t.

  Ltac weaken_to l
    := refine (derives_weaken l _ _ _ _);
      try (simpl; intuition; fail).

  Ltac weaken_with Z
    := refine (derives_weaken _ _ _ _ _);
      [ | apply Z ]; try (simpl; intuition; fail).

  Axiom derives_axiom : forall G t,
    |-- t ->
    [ G ] |-- t.

  Axiom derives_permute : forall G1 G2 t,
    Permutation G1 G2 ->
    [ G1 ] |-- t ->
    [ G2 ] |-- t.

  Axiom derives_move_left : forall P l1 l2 t,
    [ P :: l1 ++ l2 ] |-- t ->
    [ l1 ++ P :: l2 ] |-- t.

  Axiom derives_dup_left : forall P l1 l2 t,
    [ P :: l1 ++ P :: l2 ] |-- t ->
    [ l1 ++ P :: l2 ] |-- t.

  Axiom derives_assumption : forall G t,
    In t G ->
    [ G ] |-- t.

  Axiom assumed1 : forall t G,
    [ t :: G ] |-- t.
  Axiom assumed2 : forall t1 t G,
    [ t1 :: t :: G ] |-- t.
  Axiom assumed3 : forall t1 t2 t G,
    [ t1 :: t2 :: t :: G ] |-- t.
  Axiom assumed4 : forall t1 t2 t3 t G,
    [ t1 :: t2 :: t3 :: t :: G ] |-- t.
  Axiom assumed5 : forall t1 t2 t3 t4 t G,
    [ t1 :: t2 :: t3 :: t4 :: t :: G ] |-- t.
  Axiom assumed6 : forall t1 t2 t3 t4 t5 t G,
    [ t1 :: t2 :: t3 :: t4 :: t5 :: t :: G ] |-- t.
  Axiom assumed7 : forall t1 t2 t3 t4 t5 t6 t G,
    [ t1 :: t2 :: t3 :: t4 :: t5 :: t6 :: t :: G ] |-- t.
  Axiom assumed8 : forall t1 t2 t3 t4 t5 t6 t7 t G,
    [ t1 :: t2 :: t3 :: t4 :: t5 :: t6 :: t7 :: t :: G ] |-- t.

  Hint Resolve assumed1 assumed2 assumed3 assumed4
     assumed5 assumed6 assumed7 assumed8.

  Ltac assumed := auto; apply derives_assumption; simpl; intuition; fail.

  Axiom derives_left : forall l t,
    [ t :: l ] |-- t.

  Axiom derives_cut : forall c t G,
    [ c :: G ] |-- t ->
    [ G ]      |-- c ->
    [ G ]      |-- t.

  Ltac cut_with t := apply derives_cut with t.

  Axiom derives_apply : forall G G' t,
    (fold_right (fun p Z => [ G ] |-- p /\ Z) True G') ->
    [ G' ] |-- t ->
    [ G ]  |-- t.

  Ltac split_right := try trivial; try (split; [ | trivial; split_right ]).

  Ltac apply_d H := refine (derives_apply _ _ _ _ _);
    [ | apply H; auto ]; simpl; split_right.

  Ltac cut_using X :=
   match type of X with
     derives ?G ?t  =>
     cut_with t; [ | apply_d X ]
   end.

  Axiom in_all_true : forall G1 G2,
     (forall x, In x G2 -> In x G1) ->
     (forall w, all_true G1 w -> all_true G2 w).

  Fixpoint split_list (n:nat) (l1 l2:list pred) {struct n}
    : (option pred * list pred * list pred) :=

    match n, l2 with
      | 0   , x :: l2' => (Some x, rev l1, l2')
      | S n', x :: l2' => split_list n' (x::l1) l2'
      | _   , nil      => (None, rev l1, nil)
    end.
  
  Axiom split_list_together : forall n l l' z l1 l2,
    split_list n l' l = (z, l1, l2) ->
    rev l' ++ l = l1 ++ (match z with Some x => x::nil | None => nil end) ++ l2.

  Definition calc_move_at (n:nat) (l:list pred) :=
    match split_list n nil l with
      | ( None, _, _ )    => l
      | ( Some x, l1, l2) => x :: l1 ++ l2
    end.

  Definition calc_dup_at (n:nat) (l:list pred) :=
    match split_list n nil l with
      | ( None, _, _ )    => l
      | ( Some x, l1, l2) => x :: l1 ++ x :: l2
    end.
  
  Axiom calc_move_at_in : forall l n x,
    In x (calc_move_at n l) -> In x l.

  Axiom calc_dup_at_in : forall l n x,
    In x (calc_dup_at n l) -> In x l.

  Ltac move_at n :=
    match goal with
      | [ |- derives ?G ?t ] =>
        cut (derives (calc_move_at n G) t);
          [ apply derives_weaken; apply calc_move_at_in
          | unfold calc_move_at; simpl
          ]
    end.

  Ltac dup_at n :=
    match goal with
      | [ |- derives ?G ?t ] =>
        cut (derives (calc_dup_at n G) t);
          [ apply derives_weaken; apply calc_dup_at_in
          | unfold calc_dup_at; simpl
          ]
      end.

  Ltac exact_at n := move_at n; apply derives_left.

  Axiom derives_drop_head : forall t' t G,
    [ G ] |-- t ->
    [ t' :: G ] |-- t.

  Ltac clear_at x := move_at x; apply derives_drop_head.

  (* introduction and elimination rules for
     the basic logical constructs *)

  Axiom bot_elim : forall t G,
    [ bot :: G ] |-- t.

  Axiom top_intro : forall G,
    [ G ] |-- top.

  Axiom disj_intro1 : forall t1 t2 G,
    [ G ] |-- t1 ->
    [ G ] |-- (t1 || t2).

  Axiom disj_intro2 : forall t1 t2 G,
    [ G ] |-- t2 ->
    [ G ] |-- (t1 || t2).

  Axiom disj_elim : forall t1 t2 t G,
    [ t1 :: G ] |-- t ->
    [ t2 :: G ] |-- t ->
    [ t1 || t2 :: G ] |-- t.

  Axiom derives_disj_commute : forall t1 t2,
    t1 || t2 |-- t2 || t1.

  Axiom disj_commute : forall t1 t2,
    t1 || t2 = t2 || t1.

  Axiom disj_assoc : forall t1 t2 t3,
    t1 || (t2 || t3) = (t1 || t2) || t3.

  Axiom disj_bot_unit : forall t,
    bot || t = t.

  Axiom conj_intro : forall t1 t2 G,
    [ G ] |-- t1 ->
    [ G ] |-- t2 ->
    [ G ] |-- (t1 && t2).

  Axiom conj_elim : forall t1 t2 t G,
    [ t1 :: t2 :: G ] |-- t ->
    [ t1 && t2 :: G ] |-- t.

  Axiom conj_elim1 : forall t1 t2 t G,
    [ t1 :: G ]       |-- t ->
    [ t1 && t2 :: G ] |-- t.

  Axiom conj_elim2 : forall t1 t2 t G,
    [ t2 :: G ]       |-- t ->
    [ t1 && t2 :: G ] |-- t.

  Axiom derives_conj_commute : forall t1 t2,
    t1 && t2 |-- t2 && t1.

  Axiom conj_commute : forall t1 t2,
    t1 && t2 = t2 && t1.
  
  Axiom conj_assoc : forall t1 t2 t3,
    t1 && (t2 && t3) = (t1 && t2) && t3.

  Axiom intersection_top_unit : forall t,
    top && t = t.

  Axiom conj_bot : forall t,
    bot && t = bot.

  Axiom disj_top : forall t,
    top || t = top.

  Axiom distrib_disj_conj : forall n m p,
    (n || m) && p = (n&&p) || (m&&p).

  Axiom distrib_conj_disj : forall n m p,
    (n&&m) || p = (n||p) && (m||p).

  Axiom implies_intro : forall t1 t2 G,
    [ t1 :: G ] |-- t2 ->
    [ G ]       |-- (t1 --> t2).

  Axiom implies_elim : forall t1 t2 G,
    [ G ] |-- t1 ->
    [ (t1 --> t2) :: G ] |-- t2.

  Axiom modus_ponens : forall p q G t,
    [ G ] |-- p ->
    [ q :: G ] |-- t ->
    [ (p --> q) :: G ] |-- t.

  Axiom exist_intro : forall A (f:A -> pred) (x:A) G,
    [ G ] |-- (f x) ->
    [ G ] |-- (ex f).

  Axiom exist_elim : forall A (f:A -> pred) t G,
    (forall x:A, [ f x :: G ] |-- t) ->
    [ ex f :: G ] |-- t.

  Axiom forall_intro : forall A (f:A -> pred) G,
    (forall x:A, [  G ] |-- f x) ->
    [ G ] |-- all f.

  Axiom forall_elim : forall A (x:A) (f:A -> pred) t G,
    [ f x :: G ] |-- t ->
    [ all f :: G ] |-- t.

  Axiom const_intro : forall G (P:Prop),
    P -> 
    [ G ] |-- const P.

  Axiom const_elim : forall t G (P:Prop),
    (P -> [ G ] |-- t) ->
    [ const P :: G ] |-- t.

  Axiom with_projecting : forall x t G,
    (forall a:field_type x, [ projecting x a :: G ] |-- t) ->
    [ G ] |-- t.

  Axiom field_pred_elim : forall x a f t G,
    [ G ] |-- projecting x a ->
    [ f a :: G ] |-- t ->
    [ field_pred x f :: G ] |-- t.

  Axiom field_pred_intro : forall x a f G,
    [ G ] |-- projecting x a ->
    [ G ] |-- f a ->
    [ G ] |-- field_pred x f.

  Axiom projecting_eq : forall x a b G t,
    [ G ] |-- projecting x a ->
    [ G ] |-- projecting x b ->
    [ const (a = b) :: G ] |-- t ->
    [ G ] |-- t.

  Hint Resolve top_intro conj_intro disj_intro1 disj_intro2 implies_intro forall_intro const_intro
         bot_elim conj_elim disj_elim forall_elim exist_elim const_elim modus_ponens
         derives_drop_head derives_equal.

End LOGIC_THEORY.

Module Type MODAL_THEORY.
  Declare Module W:WORLDS.
  Declare Module LT:LOGIC_THEORY with Module W:=W.

  Export LT.

  Open Local Scope pred_scope.

  Inductive polarity : Set :=
    | pol_pos
    | pol_neg
    | pol_neutral
    .

  Definition polarity_spec
     (z:polarity) (op:pred->pred) (rel:world->world->Prop) :=

     match z with
     | pol_pos => forall (p:pred) w, (forall w', rel w w' -> p w') <-> op p w
     | pol_neg => forall (p:pred) w, (exists w', rel w w' /\ p w') <-> op p w
     | pol_neutral =>
       (forall (p:pred) w, (forall w', rel w w' -> p w') <-> op p w) /\
       (forall (p:pred) w, (exists w', rel w w' /\ p w') <-> op p w)
     end.

  Definition is_pos (z:polarity) :=
      match z with
      | pol_pos => True
      | pol_neg => False
      | pol_neutral => True
      end.

  Definition is_neg (z:polarity) :=
      match z with
      | pol_pos => False
      | pol_neg => True
      | pol_neutral => True
      end.

  Hint Unfold is_pos is_neg.

  Record modality (z:polarity) : Type :=
  { modal_op    :> pred -> pred
  ; modal_rel   :  world -> world -> Prop
  ; modal_spec  :  polarity_spec z modal_op modal_rel
  }.

  Implicit Arguments modal_op.
  Implicit Arguments modal_rel.
  Implicit Arguments modal_spec.

  Definition appPosMode (rel:relation world) (p:pred) : pred :=
    fun w => forall w', rel w w' -> p w'.
  Definition appNegMode (rel:relation world) (p:pred) : pred :=
    fun w => exists w', rel w w' /\ p w'.

  Definition mkPosModality (rel:relation world) : modality pol_pos :=
     Build_modality pol_pos (appPosMode rel) rel
      (fun (p:pred) (w:world) => Logic.conj (fun H => H) (fun H => H)).

  Definition mkNegModality (rel:relation world) : modality pol_neg :=
     Build_modality pol_neg (appNegMode rel) rel
      (fun (p:pred) (w:world) => Logic.conj (fun H => H) (fun H => H)).

  Definition modal_dual zm zn (M:modality zm) (N:modality zn) :=
     is_pos zm /\ is_neg zn /\ modal_rel M = modal_rel N.

  Implicit Arguments modal_dual.

  Axiom polarity_neutral_pos : forall op rel,
      polarity_spec pol_neutral op rel -> polarity_spec pol_pos op rel.
  Axiom polarity_neutral_neg : forall op rel,
      polarity_spec pol_neutral op rel -> polarity_spec pol_neg op rel.

  Definition neutral_pos_modal (B:modality pol_neutral) : modality pol_pos :=
    Build_modality pol_pos (modal_op B) (modal_rel B) (polarity_neutral_pos _ _ (modal_spec B)).

  Definition neutral_neg_modal (B:modality pol_neutral) : modality pol_neg :=
    Build_modality pol_neg (modal_op B) (modal_rel B) (polarity_neutral_neg _ _ (modal_spec B)).

  Definition serial (A:Type) (R:relation A) :=
    forall x:A, exists y:A, R x y.

  Definition confluent (R:relation world) := forall w w1 w2,
    R w w1 -> R w w2 -> exists w', R w1 w' /\ R w2 w'.

  Axiom reflexive_serial : forall A (R:relation A),
    reflexive A R -> serial A R.

  Axiom bimodal_dual : forall (B:modality pol_neutral),
      modal_dual B B.

  Definition dual_p (M:modality pol_pos) :=
    mkNegModality (modal_rel M).
  Definition dual_n (N:modality pol_neg) :=
    mkPosModality (modal_rel N).

  Axiom dual_p_is_dual : forall (M:modality pol_pos),
    modal_dual M (dual_p M).

  Axiom dual_n_is_dual : forall (N:modality pol_neg),
    modal_dual (dual_n N) N.

  Axiom pos_dual_unique : forall zm zn (M1 M2:modality zm) (N:modality zn),
    modal_dual M1 N ->
    modal_dual M2 N ->
    M1 = M2.

  Axiom neg_dual_unique : forall zm zn (M:modality zm) (N1 N2:modality zn),
    modal_dual M N1 ->
    modal_dual M N2 ->
    N1 = N2.

  Axiom pos_modal_spec : forall z (M:modality z), is_pos z ->
     polarity_spec pol_pos (modal_op M) (modal_rel M).

  Axiom neg_modal_spec : forall z (M:modality z), is_neg z ->
     polarity_spec pol_neg (modal_op M) (modal_rel M).

  Axiom neg_dual :
       (forall P:Prop, ~~P -> P) ->
       forall (zm zn:polarity)
              (M:modality zm)
              (N:modality zn),
         modal_dual M N ->
         forall p, N p = neg (M (neg p)).

  Axiom pos_dual :
       (forall P:Prop, ~~P -> P) ->
       forall (zm zn:polarity)
              (M:modality zm)
              (N:modality zn),
         modal_dual M N ->
         forall p, M p = neg (N (neg p)).

  Axiom modal_N : forall z (M:modality z) p, is_pos z ->
    |-- p ->
    |-- M p.

  Axiom modal_K : forall z (M:modality z) p q, is_pos z ->
    M (p --> q) |-- M p --> M q.

  Axiom nmodal_K : forall zm zn (M:modality zm) (N:modality zn) p q,
    modal_dual M N ->
    M (p --> q) |-- N p --> N q.

  Axiom modal_T : forall z (M:modality z) p, is_pos z ->
    reflexive _ (modal_rel M) ->
    M p |-- p.

  Axiom nmodal_T : forall z (N:modality z) p, is_neg z ->
    reflexive _ (modal_rel N) ->
    p |-- N p.

  Axiom modal_4 : forall z (M:modality z) p, is_pos z ->
    transitive _ (modal_rel M) ->
    M p |-- M (M p).

  Axiom nmodal_4 : forall z (N:modality z) p, is_neg z ->
    transitive _ (modal_rel N) ->
    N (N p) |-- N p.

  Axiom modal_D : forall zm zn (M:modality zm) (N:modality zn) p,
     modal_dual M N ->
     serial _ (modal_rel M) ->

     M p |-- N p.

  Axiom modal_B : forall zm zn (M:modality zm) (N:modality zn) p,
     modal_dual M N ->
     symmetric _ (modal_rel M) ->

     p |-- M (N p).

  Axiom modal_idem : forall z (M:modality z) p,
     reflexive _ (modal_rel M) ->
     transitive _ (modal_rel M) ->
     M (M p) = M p.

  Axiom bimodal_idem : forall (B:modality pol_neutral) p,
     transitive _ (modal_rel B) ->
     B (B p) = B p.

  Axiom derives_modal : forall z (M:modality z) t1 t2,
    t1 |-- t2 ->
    M t1 |-- M t2.

  Axiom pmodal_distrib_const : forall z (M:modality z) P, is_pos z ->
    const P |-- M (const P).

  Axiom nmodal_distrib_const : forall z (N:modality z) P, is_neg z ->
    N (const P) |-- const P.

  Axiom bimodal_const : forall (B:modality pol_neutral) P,
    B (const P) = const P.

  Axiom modal_const : forall z (M:modality z) P,
    serial _ (modal_rel M) ->
    M (const P) = const P.

  Axiom pmodal_const_impl : forall z (M:modality z) P q, is_pos z ->
    M (const P --> q) = const P --> M q.

  Axiom modal_distrib_conj : forall z (M:modality z) p q,
    M (p && q) |-- M p && M q.

  Axiom modal_distrib_disj : forall z (M:modality z) p q,
    M p || M q |-- M (p || q).

  Axiom modal_distrib_forall : forall z (M:modality z) A (f:A->pred),
    M (all f) |-- All_ a:A, M (f a).

  Axiom modal_distrib_exist : forall z (M:modality z) A (f:A->pred),
    Ex_ a:A, M (f a) |-- M (ex f).

  Axiom pmodal_conj : forall z (M:modality z) p q,
    is_pos z ->
    M (p && q) = M p && M q.

  Axiom nmodal_disj : forall z (N:modality z) p q,
    is_neg z ->
    N (p || q) = N p || N q.

  Axiom pmodal_forall : forall z (M:modality z) A (f:A->pred),
    is_pos z ->
    M (all f) = All_ a:A, M (f a).

  Axiom nmodal_exist : forall z (N:modality z) A (f:A->pred),
    is_neg z ->
    Ex_ a:A, N (f a) = N (ex f).

  Axiom modal_G : forall zm zn (M:modality zm) (N:modality zn) p,
    modal_dual M N ->
    confluent (modal_rel M) ->

    N (M p) |-- M (N p).

  Axiom neg_pos_conj : forall zm zn (M:modality zm) (N:modality zn) p q,
    modal_dual M N ->
    transitive _ (modal_rel M) ->
    confluent (modal_rel M) ->

    (N (M p)) && (N (M q)) = N (M p && M q).

  Axiom elim_pos_modal : forall z (M:modality z) p G t,
    is_pos z ->
    reflexive world (modal_rel M) ->
    [ p :: G ] |-- t ->
    [ M p :: G ] |-- t.

  Hint Resolve modal_N modal_K nmodal_K modal_T nmodal_T modal_4 nmodal_4
         modal_idem derives_modal modal_distrib_conj modal_distrib_disj
         pmodal_distrib_const nmodal_distrib_const
         modal_distrib_forall modal_distrib_exist pmodal_conj nmodal_disj
         pmodal_forall nmodal_exist reflexive_serial
         modal_G neg_pos_conj elim_pos_modal.

  Axiom dual_implies : forall zm zn (M:modality zm) (N:modality zn) p q,
    modal_dual M N ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->

    M (N p --> M q) = N p --> M q.

  Axiom dual_implies2 : forall zm zn (M:modality zm) (N:modality zn) p q,
    modal_dual M N ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->

    N (M p --> N q) = M p --> N q.

  Axiom pdrop_modal : forall z (M:modality z) t G, is_pos z ->
     (forall t', In t' G -> t' |-- M t') ->
     [ G ] |-- t ->
     [ G ] |-- M t.

End MODAL_THEORY.

Module Type LOGIC_TACTICS.
  Declare Module W:WORLDS.
  Declare Module MT:MODAL_THEORY
     with Module W:=W.

  Export MT.

  Open Local Scope pred_scope.

  Inductive ModalImpl : Type :=
  | ModalImpl_lift : pred -> ModalImpl
  | ModalImpl_modal_pos : modality pol_pos -> ModalImpl -> ModalImpl
  | ModalImpl_modal_neutral : modality pol_neutral -> ModalImpl -> ModalImpl
  | ModalImpl_implies : pred -> pred -> ModalImpl
  .

  Fixpoint interpModalImpl (x:ModalImpl) : pred :=
    match x with
      | ModalImpl_modal_pos m x' => modal_op m (interpModalImpl x')
      | ModalImpl_modal_neutral m x' => modal_op m (interpModalImpl x')
      | ModalImpl_lift p => p
      | ModalImpl_implies p q => implies p q
    end.

  Fixpoint apply_modals (l:list { z:polarity & { m:modality z | is_pos z }}) (p:pred) { struct l } : pred :=
    match l with
      | nil => p
      | (existT _ (exist m _)) :: xs => m (apply_modals xs p)
    end.

  Fixpoint do_pushdown_modal (x:ModalImpl) (l:list { z:polarity & { m:modality z | is_pos z }}) { struct x } : pred :=
       match x with
         | ModalImpl_lift z => apply_modals l z
         | ModalImpl_modal_pos m z => do_pushdown_modal z (l ++ (@existT _ _ pol_pos (@exist _ _ m I))::nil)
         | ModalImpl_modal_neutral m z => do_pushdown_modal z (l ++ (@existT _ _ pol_neutral (@exist _ _ m I))::nil)
         | ModalImpl_implies p q => implies (apply_modals l p) (apply_modals l q)
       end.

  Fixpoint split_positive_modal_head (x:ModalImpl) : (list { z:polarity & { m:modality z | is_pos z }} * ModalImpl) :=
    match x with
      | ModalImpl_modal_pos m x' =>
        let (l, z) := split_positive_modal_head x' in ((@existT _ _ pol_pos (@exist _ _ m I)) :: l, z)
      | ModalImpl_modal_neutral m x' =>
        let (l, z) := split_positive_modal_head x' in ((@existT _ _ pol_neutral (@exist _ _ m I)) :: l, z)
      | _ => (nil, x)
    end.

  Axiom pullout_modal_aux : forall l p q G,
    [ G ] |-- apply_modals l (p --> q) ->
    [ G ] |-- apply_modals l p --> apply_modals l q.

  Axiom apply_modals_app : forall l1 l2 p,
    apply_modals (l1 ++ l2) p = apply_modals l1 (apply_modals l2 p).

  Axiom apply_modals_derives : forall l t1 t2,
    t1 |-- t2 ->
    apply_modals l t1 |-- apply_modals l t2.

  Axiom apply_modals_all : forall l G t,
    derives G t ->
    derives (List.map (apply_modals l) G) (apply_modals l t).

  Axiom pushdown_modal_sound : forall x,
    interpModalImpl x |-- do_pushdown_modal x nil.

  Ltac uninterpModalImpl X ID :=
    match X with
      | @modal_op pol_pos ?M ?Y =>
        let Y' := fresh in
          uninterpModalImpl Y Y';
          set (ID:= ModalImpl_modal_pos M Y');
            subst Y'
      | @modal_op pol_neutral ?M ?Y =>
        let Y' := fresh in
          uninterpModalImpl Y Y';
          set (ID:= ModalImpl_modal_neutral M Y');
            subst Y'
      | implies ?P ?Q => set (ID := ModalImpl_implies P Q)
      | _ => set (ID := ModalImpl_lift X)
    end.

  Ltac pushdown_modal n := move_at n;
    match goal with
      | [ |- derives (?X :: _) _ ] =>
        let X' := fresh in uninterpModalImpl X X';
          cut_with (do_pushdown_modal X' nil);
          [ clear_at 1; clear X' | apply derives_weaken_head; apply (pushdown_modal_sound X') ]
    end.

  Ltac apply_loop := eapply derives_left || 
    match goal with
      | [ |- derives (all ?f :: _) _ ] =>
        match type of f with
          | ?X -> pred =>
            let e := fresh "e" in
              evar (e:X); apply (forall_elim X e); subst e; apply_loop
        end
      | [ |- derives (implies ?P ?Q :: _) _ ] =>
        apply modus_ponens; [ | apply_loop ]
      | _ => progress pushdown_modal 0; apply_loop
    end.

  Ltac apply_at n := dup_at n; apply_loop.

  Ltac pullout_modal n :=
    match goal with 
      | [ |- derives ?G (?t --> _) ] =>
        let t' := fresh in uninterpModalImpl t t';
          refine (pullout_modal_aux (List.firstn n (fst (split_positive_modal_head t'))) _ _ _ _); subst t'; simpl
    end.

  Ltac drop_non_modal l l' M :=
    match l with
      | cons (@modal_op pol_pos M ?t) ?XS =>
        drop_non_modal XS (cons (modal_op M t) l') M
      | cons (@modal_op pol_neutral M ?t) ?XS =>
        drop_non_modal XS (cons (modal_op M t) l') M
      | cons ?X ?XS =>
        drop_non_modal XS l' M
      | _ => weaken_to l'
    end.

  Ltac coalesce_modal_hyps l M :=
    match l with
      | cons (@modal_op _ M ?t1) (cons (@modal_op _ M ?t2) ?Z) =>
        cut_with (M (t1 && t2)); simpl;
          [ weaken_to (M (t1 && t2) :: Z)
          | cut_with (M t1 && M t2); simpl;
              [ apply derives_weaken_head; (in_goal rewrite pmodal_conj); auto
              | apply conj_intro; assumed
              ]
          ] 
      | _ => fail
    end.

  Ltac coalesce_modal :=
    match goal with [ |- derives ?G (@modal_op _ ?M ?t) ] =>
      drop_non_modal G (@nil pred) M;
      repeat (match goal with [ |- derives ?G ?Z ] => coalesce_modal_hyps G M end);
      apply (derives_modal _ M);
      repeat apply conj_elim
    end.

  Ltac do_elim :=
    match goal with
      | [ |- derives (bot         :: _) _ ] => apply bot_elim
      | [ |- derives (conj _ _    :: _) _ ] => apply conj_elim
      | [ |- derives (disj _ _    :: _) _ ] => apply disj_elim
      | [ |- derives (ex _        :: _) _ ] => apply exist_elim
      | [ |- derives (implies _ _ :: _) _ ] => apply modus_ponens
      | [ |- derives (const _     :: _) _ ] => apply const_elim
      | [ |- derives (@modal_op _ _ _ :: _) _ ] => apply elim_pos_modal; auto
    end.
  
  Ltac elim_at n :=
    match n with
      | 0 => do_elim
      | S _ => move_at n; do_elim
    end.

  Ltac elim_dup_at n := dup_at n; do_elim.

  Ltac spec_at n x :=
    match n with
      | 0 => apply (forall_elim _ x)
      | S _ => move_at n; apply (forall_elim _ x)
      end.

  Ltac spec_dup_at n x := dup_at n; apply (forall_elim _ x).

  Ltac intro_d :=
    match goal with
    | [ |- derives _ top          ] => apply top_intro
    | [ |- derives _ (conj _ _)   ] => apply conj_intro
    | [ |- derives _ (implies _ _)] => apply implies_intro
    | [ |- derives _ (all _)      ] => apply forall_intro
    end.

  Ltac intros_d  := repeat (intro_d; intros).
  Ltac split_d   := apply conj_intro.
  Ltac left_d    := apply disj_intro1.
  Ltac right_d   := apply disj_intro2.
  Ltac exist_d x := apply exist_intro with x.

  Inductive LogicRep : Type :=
    | LR_Lift : pred -> LogicRep
    | LR_Bot : LogicRep
    | LR_Top : LogicRep
    | LR_Disj : LogicRep -> LogicRep -> LogicRep
    | LR_Conj : LogicRep -> LogicRep -> LogicRep
    | LR_Impl : LogicRep -> LogicRep -> LogicRep
    | LR_Ex   : forall A:Type, (A->pred) -> LogicRep
    | LR_All  : forall A:Type, (A->pred) -> LogicRep
    | LR_Neg  : LogicRep -> LogicRep
    | LR_Const : Prop -> LogicRep
    | LR_Proj  : forall x:field, field_type x -> LogicRep
    | LR_FPred : forall x:field, (field_type x -> pred) -> LogicRep
    | LR_PMod  : modality pol_pos -> LogicRep -> LogicRep
    | LR_NMod  : modality pol_neg -> LogicRep -> LogicRep
    | LR_BMod  : modality pol_neutral -> LogicRep -> LogicRep
    .

  Fixpoint interpLogicRep (x:LogicRep) : pred :=
    match x with
    | LR_Lift p         => p
    | LR_Bot            => bot
    | LR_Top            => top
    | LR_Disj p q       => disj (interpLogicRep p) (interpLogicRep q)
    | LR_Conj p q       => conj (interpLogicRep p) (interpLogicRep q)
    | LR_Impl p q       => implies (interpLogicRep p) (interpLogicRep q)
    | LR_Ex A f         => ex f
    | LR_All A f        => all f
    | LR_Neg p          => neg (interpLogicRep p)
    | LR_Const P        => const P
    | LR_Proj x a       => projecting x a
    | LR_FPred x f      => field_pred x f
    | LR_PMod m p       => m (interpLogicRep p)
    | LR_NMod m p       => m (interpLogicRep p)
    | LR_BMod m p       => m (interpLogicRep p)
    end.

  Inductive CxtRep : Type :=
    | CR_Goal  : pred -> CxtRep
    | CR_Hyp   : pred -> CxtRep -> CxtRep
    | CR_Const : Prop -> CxtRep -> CxtRep
    | CR_Ex    : forall A:Type, (A->pred) -> CxtRep -> CxtRep
    .

  Fixpoint interp_cxt (Cxt:CxtRep) (l:list pred) { struct Cxt } : Prop :=
    match Cxt with
    | CR_Goal g     => derives l g
    | CR_Hyp H C'   => interp_cxt C' (l ++ H::nil)
    | CR_Const P C' => P -> interp_cxt C' l
    | CR_Ex A f C'  => forall x:A, interp_cxt C' (l ++ (f x::nil))
    end.

  Definition interp_simpl_cxts (Cxts:list CxtRep) (l:list pred) (P:Prop) :=
    fold_right (fun Cxt X => interp_cxt Cxt l /\ X) P Cxts.

  Fixpoint do_simpl_hyp1 (H:LogicRep) (Cxts:list CxtRep) { struct H } : list CxtRep :=
     match H with
     | LR_Bot       => nil
     | LR_Top       => Cxts
     | LR_Disj p q  => do_simpl_hyp1 p Cxts ++ do_simpl_hyp1 q Cxts
     | LR_Conj p q  => do_simpl_hyp1 p (do_simpl_hyp1 q Cxts)
     | LR_Ex A ex_f => List.map (CR_Ex A ex_f) Cxts
     | LR_Const P   => List.map (CR_Const P) Cxts
     | _            => List.map (CR_Hyp (interpLogicRep H)) Cxts
     end.

  Fixpoint do_simpl_hyps (Hyps:list LogicRep) (Cxts:list CxtRep) { struct Hyps } : list CxtRep :=
    match Hyps with
    | H::Hyps' => do_simpl_hyps Hyps' (do_simpl_hyp1 H Cxts)
    | nil => Cxts
    end.

  Definition simpl_hyps (Hyps:list LogicRep) (t:pred) : Prop :=
    interp_simpl_cxts (do_simpl_hyps (rev Hyps) ((CR_Goal t) :: nil)) nil True.

  Fixpoint simpl_goal_rep (Hyps:list LogicRep) (g:LogicRep) { struct g } : Prop :=
    match g with
    | LR_Top      => True
    | LR_Conj p q => simpl_goal_rep Hyps p /\ simpl_goal_rep Hyps q
    | LR_Impl p q => simpl_goal_rep (p :: Hyps) q
    | LR_All A f  => forall x:A, simpl_hyps Hyps (f x)
    | _ => simpl_hyps Hyps (interpLogicRep g)
    end.

  Axiom simpl_goal_rep_sound : forall Hyps g,
    simpl_goal_rep Hyps g -> derives (List.map interpLogicRep Hyps) (interpLogicRep g).

  Ltac uninterpLogicRep X ID :=
    match X with
      | bot => set (ID:=LR_Bot)
      | top => set (ID:=LR_Top)
      | disj ?p ?q =>
         let p' := fresh "p" in let q' := fresh "q" in
         uninterpLogicRep p p'; uninterpLogicRep q q';
         set (ID:=LR_Disj p' q'); subst p'; subst q'
      | conj ?p ?q =>
         let p' := fresh "p" in let q' := fresh "q" in
         uninterpLogicRep p p'; uninterpLogicRep q q';
         set (ID:=LR_Conj p' q'); subst p'; subst q'
      | implies ?p ?q => 
         let p' := fresh "p" in let q' := fresh "q" in
         uninterpLogicRep p p'; uninterpLogicRep q q';
         set (ID:=LR_Impl p' q'); subst p'; subst q'
      | @ex ?A ?f  => set (ID:=LR_Ex A f)
      | @all ?A ?f => set (ID:=LR_All A f)
      | const ?P  => set (ID:=LR_Const P)
      | _ => set (ID := LR_Lift X)
    end.

  Ltac uninterpHyps Hyps ID :=
     match Hyps with
     | nil => set (ID:=@nil LogicRep)
     | ?H :: ?HS =>
         let H'  := fresh "H" in
         let HS' := fresh "HS" in
         uninterpLogicRep H H';
         uninterpHyps HS HS';
         set (ID:= H' :: HS');
         subst H'; subst HS'
     end.

  Ltac uninterpDerives :=
     match goal with
     | [ |- derives ?Hyps ?G ] =>
         let G' := fresh "G" in
         let Hyps' := fresh "Hyps" in
         uninterpLogicRep G G';
         uninterpHyps Hyps Hyps';
         change (derives (List.map interpLogicRep Hyps') (interpLogicRep G'));
         subst G'; subst Hyps'
     end.

  Ltac simpl_derives :=
     match goal with
     | [ |- derives ?Hyps ?G ] =>
         let G' := fresh "G" in
         let Hyps' := fresh "Hyps" in
         uninterpLogicRep G G';
         uninterpHyps Hyps Hyps';
         change (derives (List.map interpLogicRep Hyps') (interpLogicRep G'));
         apply simpl_goal_rep_sound;
         subst G'; subst Hyps';
         simpl; unfold simpl_hyps; simpl;
          do 2 (split_right; intros); auto
     end.

  Ltac intuition_d := repeat (progress simpl_derives).

End LOGIC_TACTICS.


Module Type GOEDEL_LOEB_THEORY.
  Declare Module W:WORLDS.
  Declare Module Tac : LOGIC_TACTICS
    with Module W:=W.

  Export Tac.

  Open Local Scope pred_scope.

  Definition later_rel (w1 w2:world) :=
    exists n, tick_world_n (S n) w1 = Some w2.
  Definition later := appPosMode later_rel.

  Parameter laterM : modality pol_pos.
  Axiom laterM_spec : modal_op laterM = later.
  Axiom laterM_rel_spec : modal_rel laterM = later_rel.
  Notation "'|>' e" := (laterM e) (at level 30, right associativity): pred_scope.

  Definition nec_rel (w1 w2:world) :=
    exists n, tick_world_n n w1 = Some w2.
  Definition nec := appPosMode nec_rel.
  
  Parameter necM : modality pol_pos.
  Axiom necM_spec : modal_op necM = nec.
  Axiom necM_rel_spec : modal_rel necM = nec_rel.
  Notation "'[]' e" := (necM e) (at level 30, right associativity): pred_scope.

  Definition eventually := appNegMode nec_rel.

  Parameter eventuallyM : modality pol_neg.
  Axiom eventually_spec : modal_op eventuallyM = eventually.
  Axiom eventually_rel_spec : modal_rel eventuallyM = nec_rel.
  Notation "'<>' e" := (eventuallyM e) (at level 30, right associativity): pred_scope.

  Axiom nec_eventually_dual : modal_dual necM eventuallyM.

  Definition everywhere_rel (w1 w2:world) := world_ttl w1 = world_ttl w2.
  Definition everywhere := appPosMode everywhere_rel.

  Parameter everywhereM : modality pol_pos.
  Axiom everywhere_spec : modal_op everywhereM = everywhere.
  Axiom everywhere_rel_spec : modal_rel everywhereM = everywhere_rel.
  Notation "'!' e" := (everywhereM e) (at level 30, right associativity) : pred_scope. 

  Definition sub (t1 t2:pred) := []!(t1 --> t2).
  Notation "P '>=>' Q" := (sub P Q) (at level 60, no associativity) : pred_scope.
  Definition equ (t1 t2:pred) := (t1 >=> t2) && (t2 >=> t1).
  Notation "P '<=>' Q" := (equ P Q) (at level 60, no associativity) : pred_scope.

  Fixpoint composeN (A: Type) (n: nat) (F: A -> A) (x: A) : A :=
   match n with O => x | S n' => F (composeN A n' F x) end.
  Implicit Arguments composeN.

  Definition rec (F: pred -> pred) : pred :=
     fun w => composeN (S (world_ttl w)) F bot w.

  Definition contractive  (F : pred -> pred) : Prop := forall t1 t2,
     |> (t1 <=> t2) |-- (F t1) <=> (F t2).

  Definition nonexpansive (F : pred -> pred) : Prop := forall t1 t2,
     (t1 <=> t2) |-- (F t1) <=> (F t2).

  Axiom laterM_trans : transitive _ (modal_rel laterM).
  Hint Resolve laterM_trans.

  Axiom loeb_rule: forall t, (|>t |-- t) -> (|-- t).

  Axiom generalized_loeb_rule: forall s t,  
      s |-- |> s ->
      [ s :: (|> t) :: nil ] |-- t  ->
      s |-- t.

  Axiom necM_refl : reflexive _ (modal_rel necM).
  Axiom necM_trans : transitive _ (modal_rel necM).
  Axiom necM_confluent : confluent (modal_rel necM).

  Axiom eventuallyM_refl : reflexive _ (modal_rel eventuallyM).
  Axiom eventuallyM_trans : transitive _ (modal_rel eventuallyM).
  Axiom eventuallyM_confluent : confluent (modal_rel eventuallyM).

  Hint Resolve necM_refl necM_trans necM_confluent
     eventuallyM_refl eventuallyM_trans eventuallyM_confluent
     nec_eventually_dual.

  Axiom nec_now   : forall t, []t |-- t.
  Axiom nec_later : forall t, []t |-- |> t.
  Axiom nec_split : forall G t,
    [ G ] |-- t ->
    [ G ] |-- (|> t) ->
    [ G ] |-- ([] t).

  Hint Resolve nec_now nec_later nec_split.

  Axiom nec_now_later : forall t,
    []t = t && |>t.
  Axiom later_nec_later : forall t,
    |>[]t = |>t.
  Axiom nec_later_later : forall t,
    []|>t = |>t.
  Axiom nec_later_commute : forall t,
    []|> t = |>[] t.

  Axiom everywhereM_refl :
     reflexive _ (modal_rel everywhereM).
  Axiom everywhereM_trans : 
     transitive _ (modal_rel everywhereM).
  Axiom everywhereM_sym :
     symmetric _ (modal_rel everywhereM).
  Hint Resolve everywhereM_refl everywhereM_trans everywhereM_sym.
  
  Axiom every_later_commute : forall p,
    !|>p = |>!p.
  Axiom every_nec_commute : forall p,
    ![]p = []!p.

  Axiom sub_refl : forall t:pred, |-- t >=> t.
  Axiom sub_trans : forall t1 t2 t3,
    [ t1 >=> t2 :: t2 >=> t3 :: nil ] |-- t1 >=> t3.
  Axiom sub_antisym : forall t1 t2,
    [ t1 >=> t2 :: t2 >=> t1 :: nil ] |-- t1 <=> t2.

  Axiom equ_refl : forall t, |-- t <=> t.
  Axiom equ_trans : forall t1 t2 t3,
    [ t1 <=> t2 :: t2 <=> t3 :: nil ] |-- t1 <=> t3.
  Axiom equ_sym : forall t1 t2, t1 <=> t2 |-- t2 <=> t1.

  Axiom sub_project_herenow : forall p q,
    [ p >=> q :: p :: nil ] |-- q.

  Axiom sub_implies : forall t1 t2 w,
     (t1 >=> t2) w ->
     (t1 w -> t2 w).

  Axiom equ_implies : forall t1 t2 w,
     (t1 <=> t2) w ->
     (t1 w -> t2 w).

  Axiom equ_equal : forall t1 t2,
     |-- t1 <=> t2 ->
     t1 = t2.
  Hint Resolve sub_refl equ_refl.

  Axiom equ_elim : forall p q G,
    [ G ] |-- q ->
    [ p <=> q :: G ] |-- p.

  Axiom flip_equ : forall p q G t,
    [ p <=> q :: G ] |-- t ->
    [ q <=> p :: G ] |-- t.

  Axiom later_lift_sub : forall p q,
    |> (p >=> q) |-- |>p >=> |> q.

  Axiom equ_sub : forall p q,
    p <=> q |-- p >=> q.

  Axiom contractive_nonexpansive: forall F,
     contractive F -> nonexpansive F.
  Axiom contractive_o_nonexpansive: forall F G,
     contractive F ->
     nonexpansive G ->
     contractive (fun x => F (G x)).
  Axiom nonexpansive_o_contractive: forall F G, 
     nonexpansive F ->
     contractive G ->
     contractive (fun x => F (G x)).
  Axiom nonexpansive_o_nonexpansive: forall F G, 
     nonexpansive F ->
     nonexpansive G ->
     nonexpansive (fun x => F (G x)).
  Axiom contractive_const: forall t, contractive (fun _ => t).
  Axiom contractive_later: contractive laterM.
  Axiom nonexpansive_id : nonexpansive (fun x => x).

  Axiom sub_later : forall t1 t2,
      t1 >=> t2 |-- |>t1 >=> |>t2.
  Axiom sub_nec : forall t1 t2,
      t1 >=> t2 |-- []t1 >=> []t2.
  Axiom sub_every : forall t1 t2,
      t1 >=> t2 |-- !t1 >=> !t2.

  Axiom sub_top : forall t, |-- t >=> top. 
  Axiom sub_bot : forall t, |-- bot >=> t.

  Axiom sub_conj : forall a a' b b',
     [ a >=> a' :: b >=> b' :: nil ] |-- a && b >=> a' && b'.
  Axiom sub_disj : forall a a' b b',
     [ a >=> a' :: b >=> b' :: nil ] |-- a || b >=> a' || b'.
  Axiom sub_forall : forall A f f',
     All_ a:A, f a >=> f' a |-- all f >=> all f'.
  Axiom sub_exist : forall A f f',
     All_ a:A, f a >=> f' a |-- ex f >=> ex f'.

  Axiom equ_later : forall t1 t2,
      t1 <=> t2 |-- |>t1 <=> |>t2.
  Axiom equ_nec : forall t1 t2,
      t1 <=> t2 |-- []t1 <=> []t2.
  Axiom equ_every : forall t1 t2,
      t1 <=> t2 |-- !t1 <=> !t2.

  Axiom equ_conj : forall a a' b b',
     [ a <=> a' :: b <=> b' :: nil ] |-- a && b <=> a' && b'.
  Axiom equ_disj : forall a a' b b',
     [ a <=> a' :: b <=> b' :: nil ] |-- a || b <=> a' || b'.
  Axiom equ_forall : forall A f f',
     All_ a:A, f a <=> f' a |-- all f <=> all f'.
  Axiom equ_exist : forall A f f',
     All_ a:A, f a <=> f' a |-- ex f <=> ex f'.

  Axiom sub_inter_right : forall x p q G,
    [ G ] |-- x >=> p ->
    [ G ] |-- x >=> q ->
    [ G ] |-- x >=> p && q.

  Axiom sub_inter_left1 : forall x p q G,
    [ G ] |-- p >=> x ->
    [ G ] |-- p && q >=> x.

  Axiom sub_inter_left2 : forall x p q G,
    [ G ] |-- q >=> x ->
    [ G ] |-- p && q >=> x.

  Axiom sub_modal : forall z (M:modality z) p q,
     |-- p >=> q ->
     |-- M p >=> M q.

  Axiom nonexpansive_conj : forall f g,
      nonexpansive f ->
      nonexpansive g ->
      nonexpansive (fun x => f x && g x).
  Axiom contractive_conj : forall f g,
      contractive f ->
      contractive g ->
      contractive (fun x => f x && g x).
  Axiom nonexpansive_disj : forall f g,
      nonexpansive f ->
      nonexpansive g ->
      nonexpansive (fun x => f x || g x).
  Axiom contractive_disj : forall f g,
      contractive f ->
      contractive g ->
      contractive (fun x => f x || g x).
  Axiom nonexpansive_forall : forall A (F:pred -> A -> pred),
      (forall a:A, nonexpansive (fun x => F x a)) ->
      nonexpansive (fun x => (all (F x))).
  Axiom contractive_forall : forall A (F:pred -> A -> pred),
      (forall a:A, contractive (fun x => F x a)) ->
      contractive (fun x => (all (F x))).
  Axiom nonexpansive_exist : forall A (F:pred -> A -> pred),
      (forall a:A, nonexpansive (fun x => F x a)) ->
      nonexpansive (fun x => (ex (F x))).
  Axiom contractive_exist : forall A (F:pred -> A -> pred),
      (forall a:A, contractive (fun x => F x a)) ->
      contractive (fun x => (ex (F x))).

  Axiom rec_equ : forall F,
     contractive F ->
     |-- (rec F) <=> (F (rec F)).
  Axiom fold_unfold: forall F,
     contractive F ->
     rec F = F (rec F).
(*
  Axiom adjective_rec : forall F z (M:modality z),
    contractive F ->
    (forall P, |> (P <=> M P) |-- F (M P) <=> M (F P)) ->
    rec F = M (rec F).
*)

End GOEDEL_LOEB_THEORY.

Module Type FIELD_LOGIC_THEORY.
  Declare Module W:WORLDS.
  Declare Module GL:GOEDEL_LOEB_THEORY
    with Module W:=W.

  Export GL.

  Open Local Scope pred_scope.

  Definition with_subst_rel (z:subst) :=
    fun w w' => world_update z w = w'.

  Parameter with_subst : subst -> modality pol_neutral.
  Axiom with_subst_spec : forall z,
     modal_op (with_subst z) = appPosMode (with_subst_rel z).
  Axiom with_subst_rel_spec : forall z,
     modal_rel (with_subst z) = (with_subst_rel z).

  Axiom with_subst_transitive : forall z,
     transitive _ (modal_rel (with_subst z)).

  Hint Resolve with_subst_transitive : modal.

  Definition respecting_rel (l:list field) :=
    fun w w' =>
     (forall x, In x l -> w_proj x w = w_proj x w') /\
     world_ttl w = world_ttl w'.

  Parameter respecting : list field -> modality pol_pos.
  Axiom respecting_spec : forall l,
     modal_op (respecting l) = appPosMode (respecting_rel l).
  Axiom respecting_rel_spec : forall l,
     modal_rel (respecting l) = respecting_rel l.

  Axiom respecting_refl : forall l, reflexive _ (modal_rel (respecting l)).
  Axiom respecting_trans : forall l, transitive _ (modal_rel (respecting l)).

  Hint Resolve respecting_refl respecting_trans.

  Axiom respecting_incl : forall l1 l2 p,
    incl l1 l2 ->
    respecting l1 p |-- respecting l2 p.

  Axiom respecting_later_commute : forall l p,
    respecting l (|> p) = |> (respecting l p).

  Axiom respecting_nec_commute : forall l p,
    respecting l ([] p) = [] (respecting l p).

  Axiom everywhere_respecting : forall l p,
    !p |-- respecting l p.

  Axiom everywhere_respecting_nil : forall p,
    !p = respecting nil p.

  Axiom equ_respecting : forall l p q,
    p <=> q |-- respecting l p <=> respecting l q.

  Axiom every_with_subst : forall z p,
    !p |-- with_subst z p.

  Axiom with_subst_same : forall z1 z2 t,
    (forall w, world_update z1 w = world_update z2 w) ->
    with_subst z1 t = with_subst z2 t.

  Axiom with_subst_nil : forall t,
    with_subst nil t = t.

  Axiom with_subst_float : forall x z t,
    with_subst z t = with_subst (float_sel x z) t.

  Axiom with_subst_commute : forall z1 z2 t,
    (forall x, In x (upd_set z1) -> In x (upd_set z2) -> False) ->
    with_subst z1 (with_subst z2 t) = with_subst z2 (with_subst z1 t).

  Axiom with_subst_app : forall z1 z2 t,
    with_subst (z2++z1) t = with_subst z1 (with_subst z2 t).

  Axiom w_upd_noninterference_n : forall n w w' x,
    tick_world_n n w = Some w' ->
    w_proj x w = w_proj x w'.

  Axiom respecting_with_subst : forall l z p,
    (forall x, In x (upd_set z) -> ~In x l) ->
    respecting l p |-- with_subst z p.

  Axiom with_subst_normalize : forall z p,
    with_subst z p = with_subst (normalize_update z) p.

  Axiom with_subst_split_one : forall x a z p,
    with_subst (U x a :: z) p = with_subst z (with_subst (U x a :: nil) p).

  Axiom with_subst_later_commute : forall z p,
    with_subst z (|> p) = |> (with_subst z p).

  Axiom with_subst_nec_commute : forall z p,
    with_subst z ([] p) = [] (with_subst z p).

  Axiom every_eats_with_subst : forall z p,
    with_subst z (!p) = !p.

  Axiom sub_eats_with_subst : forall z p q,
    with_subst z (p >=> q) = p >=> q.

  Axiom sub_with_subst : forall z p q,
    p >=> q |-- with_subst z p >=> with_subst z q.

  Axiom equ_with_subst : forall z p q,
    p <=> q |-- with_subst z p <=> with_subst z q.

  Axiom with_subst_field_pred : forall x (a:field_type x) z F,
    with_subst (U x a :: z) (field_pred x F) = 
    with_subst (U x a :: z) (F a).

  Axiom with_subst_projecting : forall x (a b:field_type x) z,
    with_subst (U x a :: z) (projecting x b) = (const (a = b)).

  Axiom respecting_compose1 : forall l l1 l2 p,
    incl l l1 -> incl l l2 ->
    respecting l p |-- respecting l1 (respecting l2 p).

  Axiom respecting_compose2 : forall l l1 l2 p,
    (forall x, In x l1 /\ In x l2 -> In x l) ->
    respecting l1 (respecting l2 p) |-- respecting l p.

  Axiom respecting_compose : forall l l1 l2 p,
    incl l l1 -> incl l l2 ->
    (forall x, In x l1 /\ In x l2 -> In x l) ->
    respecting l1 (respecting l2 p) = respecting l p.

  Axiom respecting_intersect : forall l1 l2 p,
    respecting l1 (respecting l2 p) =
    respecting (set_inter field_eq_dec l1 l2) p.

End FIELD_LOGIC_THEORY.

Module Type ADJECTIVE_THEORY.
  Declare Module W:WORLDS.
  Declare Module FT : FIELD_LOGIC_THEORY
    with Module W:=W.

  Export FT.

  Open Local Scope pred_scope.

  Record adjective (z:polarity) (M:modality z) : Type :=
    { adj_pred      :> pred -> pred
    ; adj_pred_spec : forall p:pred, adj_pred p = (p <=> M p)
    ; adj_refl      : reflexive _ (modal_rel M)
    }.

  Parameter mkAdjective : forall z (M:modality z),
    reflexive _ (modal_rel M) ->
    adjective z M.

  Definition necessary  := mkAdjective _ necM necM_refl.
  Definition inevitable := mkAdjective _ eventuallyM eventuallyM_refl.
  Definition ubiquitous := mkAdjective _ everywhereM everywhereM_refl.
  Definition respects l := mkAdjective _ (respecting l) (respecting_refl l).

  Axiom adj_intro : forall z M (J:adjective z M) p G,
    [ G ] |-- p <=> M p ->
    [ G ] |-- J p.

  Axiom adj_elim : forall z M (J:adjective z M) p G t,
    [ p <=> M p :: G ] |-- t ->
    [ J p :: G ] |-- t.

  Axiom add_adj : forall z M (J:adjective z M) p G t,
    [ G ] |-- J p ->
    [ M p :: G ] |-- t ->
    [ p :: G ] |-- t.

  Axiom remove_adj : forall z M (J:adjective z M) p G t,
    [ G ] |-- J p ->
    [ p :: G ] |-- t ->
    [ M p :: G ] |-- t.

  Axiom add_adj_goal : forall z M (J:adjective z M) G t,
    [ G ] |-- J t ->
    [ G ] |-- M t ->
    [ G ] |-- t.

  Axiom remove_adj_goal : forall z M (J:adjective z M) G t,
    [ G ] |-- J t ->
    [ G ] |-- t ->
    [ G ] |-- M t.

  Axiom prove_adj : forall z M (J:adjective z M) p G,
    [ G ] |-- J p ->
    [ p :: G ] |-- M p.

  Axiom adjective_const : forall z M (J:adjective z M) P,
    |-- J (const P).

  Axiom adjective_impl_const : forall z M (J:adjective z M) P q, is_pos z ->
    J q |-- J (const P --> q).

  Axiom adjective_bot : forall z M (J:adjective z M),
    |-- J bot.

  Axiom adjective_top : forall z M (J:adjective z M),
    |-- J top.

  Axiom adjective_conj : forall z M (J:adjective z M) p q,
    [ J p :: J q :: nil ] |-- J (p && q).

  Axiom adjective_disj : forall z M (J:adjective z M) p q,
    [ J p :: J q :: nil ] |-- J (p || q).

  Axiom adjective_forall : forall z M (J:adjective z M) A (f:A->pred),
    All_ a:A, J (f a) |-- J (all f).

  Axiom adjective_exist : forall z M (J:adjective z M) A (f:A->pred),
    All_ a:A, J (f a) |-- J (ex f).

  Axiom adjective_recursive' : forall F M,
    contractive F ->
    (forall P, |> (P <=> M P) |-- F (M P) <=> M (F P)) ->
    |-- rec F <=> M (rec F).

  Axiom adjective_recursive : forall z M (J:adjective z M) F,
    contractive F ->
    (All_ P:pred, []((|> (J P)) --> (F (M P) <=> M (F P)))) |-- J (rec F).

  Axiom adjective_modal : forall z M (J:adjective z M) p,
    transitive _ (modal_rel M) ->
    |-- J (M p).

  Axiom later_necessary : forall p,
    |-- necessary (|> p).

  Axiom nec_necessary : forall p,
    |-- necessary ([] p).

  Axiom projecting_necessary : forall x a,
    |-- necessary (projecting x a).

  Axiom field_pred_necessary : forall x f,
    All_ a:field_type x, necessary (f a) |-- necessary (field_pred x f).

  Axiom respecting_necessary : forall l p,
    necessary p |-- necessary (respecting l p).

  Axiom later_respects : forall l p,
    respects l p |-- respects l (|> p).

  Axiom nec_respects : forall l p,
    respects l p |-- respects l ([] p).

  Axiom projecting_respects : forall x l a,
    In x l ->
    |-- respects l (projecting x a).

  Axiom field_pred_respects : forall x l f,
    In x l ->
    (All_ a:field_type x, respects l (f a)) |--
       respects l (field_pred x f).

  Axiom respecting_respects : forall l1 l2 p,
    incl l2 l1 ->
    |-- respects l1 (respecting l2 p).

  Axiom respects_nil_ubiquitous : forall p,
    respects nil p = ubiquitous p.

  Axiom with_subst_necessary : forall z p,
    necessary p |-- necessary (with_subst z p).

  Axiom with_subst_respects : forall l z p,
    respects l p |-- respects l (with_subst z p).

  Axiom adj_nec : forall z M (J:adjective z M) p,
    J p = [](J p).

  Axiom adj_every : forall z M (J:adjective z M) p,
    J p = !(J p).

  Axiom adj_nec_every : forall z M (J:adjective z M) p,
    J p = []!(J p).

End ADJECTIVE_THEORY.


Module Type PROGRAM_SAFETY_THEORY.
  Declare Module W:WORLDS.
  Declare Module WS:WORLD_STEP
    with Module W:=W.
  Declare Module AT : ADJECTIVE_THEORY
    with Module W:=W.

  Export WS.
  Export AT.

  Open Local Scope pred_scope.

  Definition step_or_halt (p:prog_desc) (k:prog_cont) (w:world) :=
     halted p k w \/ exists k', exists w', step p k w k' w'.

  Inductive stepN (p:prog_desc) : nat -> prog_cont -> world -> prog_cont -> world -> Prop :=
   | stepN_O: forall w k, stepN p 0 k w k w
   | stepN_S: forall n k k' k'' w w' w'',
          step p k w k' w' -> stepN p n k' w' k'' w'' -> stepN p (S n) k w k'' w''
   .

  Definition safeN (p: prog_desc) (n: nat) (k:prog_cont) (w:world) := 
     forall n' k' w', n' <= n -> stepN p n' k w k' w' -> step_or_halt p k' w'.

  Definition safe (p:prog_desc) (k:prog_cont) : pred :=
      fun w => safeN p (world_ttl w) k w.

  Definition guards (p:prog_desc) (k:prog_cont) (pre:pred) :=
     pre >=> safe p k.

  Definition with_step_rel (p:prog_desc) (k k':prog_cont) (w w':world) :=
     step p k w k' w'.

  Parameter with_step : prog_desc -> prog_cont -> prog_cont -> modality pol_neg.
  Axiom with_step_spec : forall p k k', modal_op (with_step p k k') = appNegMode (with_step_rel p k k').
  Axiom with_step_rel_spec : forall p k k', modal_rel (with_step p k k') = with_step_rel p k k'.

  Axiom step_tick_n_commute : forall p n k k' w w',
     (exists w1, step p k w k' w1 /\ tick_world_n n w1 = Some w') <->
     (exists w2, tick_world_n n w = Some w2 /\ step p k w2 k' w').

  Axiom tick_step_or_halt : forall p n k w w',
    tick_world_n n w = Some w' ->
    step_or_halt p k w -> step_or_halt p k w'.

  Axiom tick_step_or_halt2 : forall p n k w w',
    tick_world_n n w = Some w' ->
    step_or_halt p k w' -> step_or_halt p k w.

  Axiom safe_later : forall p k,
    safe p k |-- |> (safe p k).

  Axiom with_step_later_commute : forall p k k' t,
     with_step p k k' (|> t) |-- |> (with_step p k k' t).

  Axiom with_step_everywhere : forall p k k' t,
    with_step p k k' (!t) |-- !t.

  Axiom with_step_everywhere2 : forall p k k' t t',
    [ with_step p k k' t' :: !t :: nil ] |-- with_step p k k' (!t).

  Axiom safe_step : forall p k k',
    with_step p k k' (|> safe p k') |-- safe p k.

  Axiom step_guard : forall p k e G,
     ( forall t, In t G -> t |-- []!t ) ->
     [ e :: G ] |-- Ex_ e':pred, (Ex_ k':prog_cont, |> (guards p k' e') && with_step p k k' (|> e')) ->
     [ G ] |-- guards p k e.

End PROGRAM_SAFETY_THEORY.

Module Type MODAL_MODEL.
  Declare Module W:WORLDS.
  Declare Module WS:WORLD_STEP with Module W:=W.
  Declare Module WThy : WORLD_THEORY with Module W:=W.
  Declare Module LThy : LOGIC_THEORY with Module W:=W.
  Declare Module MThy : MODAL_THEORY with Module W:=W.
  Declare Module Tac  : LOGIC_TACTICS with Module W:=W.
  Declare Module GLThy : GOEDEL_LOEB_THEORY with Module W:=W.
  Declare Module AThy  : ADJECTIVE_THEORY with Module W:=W.
  Declare Module FThy  : FIELD_LOGIC_THEORY with Module W:=W.
  Declare Module PSThy : PROGRAM_SAFETY_THEORY
       with Module W:=W with Module WS:=WS.

  Export PSThy.
End MODAL_MODEL.
