(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import Join.
Require Import CompleteInduction.

Require Import ModelSpec.
Require Import Worlds.
Require Import Logic.
Require Import Modal.
Require Import Tactics.
Require Import GoedelLoeb.
Require Import Fields.
Require Import Safety.

Module Type JOINABLE_WORLDS.
  Declare Module W:WORLDS.

  Import W.

  Parameter worlds_joinable : joinable world.

  Parameter join_tick_compatible : forall w1 w2 w w1' w2' w',
     tick_world w1 = Some w1' ->
     tick_world w2 = Some w2' ->
     tick_world w  = Some w'  ->
     (join worlds_joinable w1 w2 w <-> join worlds_joinable w1' w2' w').

  Parameter tick_joinable_succeeds : forall w1 w2 w,
     join worlds_joinable w1 w2 w ->
      ((exists w1', tick_world w1 = Some w1') <->
       (exists w2', tick_world w2 = Some w2')).

  Parameter tick_joined_succeeds : forall w1 w2 w,
     join worlds_joinable w1 w2 w ->
     ((exists w1', tick_world w1 = Some w1') <->
      (exists w',  tick_world w  = Some w')).

End JOINABLE_WORLDS.


Module Type SEP_LOGIC_THEORY.
  Declare Module W:WORLDS.
  Declare Module WJ : JOINABLE_WORLDS
    with Module W:=W.
  Declare Module ST : PROGRAM_SAFETY_THEORY
    with Module W:=W.

  Export WJ.
  Export ST.

  Open Local Scope pred_scope.

  Definition emp : pred :=
    fun c => is_empty worlds_joinable c.
  Definition sep_con (p q:pred) : pred :=
    fun c => exists c1, exists c2, join worlds_joinable c1 c2 c /\ p c1 /\ q c2.
  Definition sep_impl (p q:pred) : pred :=
    fun c => forall c1, forall c2, join worlds_joinable c c1 c2 -> p c1 -> q c2.

  Notation "P '*' Q" := (sep_con P Q) (at level 40, left associativity) : pred_scope.
  Infix "--*" := sep_impl (at level 60, right associativity) : pred_scope.

  Axiom sep_comm  : forall p q, p * q = q * p.
  Axiom sep_assoc : forall p q r, p * (q * r) = (p * q) * r.
  Axiom sep_unit1 : forall p, emp * p = p.
  Axiom sep_unit2 : forall p, p * emp = p.

  Axiom sep_cut1 : forall p' p q r,
    p |-- p' ->
    p' * q |-- r ->
    p  * q |-- r.

  Axiom sep_cut2 : forall q' p q r,
    q |-- q' ->
    p * q' |-- r ->
    p * q  |-- r.

  Axiom sep_split : forall p q p' q',
    p |-- p' ->
    q |-- q' ->
    p * q |-- p' * q'.

  Axiom wand_intro : forall p q r,
    p * q |-- r ->
    p |-- q --* r.

  Axiom wand_elim : forall p q r s,
    r |-- p ->
    q |-- s ->
    (p --* q) * r |-- s.

  Axiom wand_elim_emp : forall p q s,
    emp |-- p ->
    q |-- s ->
    (p --* q) |-- s.

  Definition lax_monoidal1 (R:relation world)  :=
    forall w1 w2 w w',
       join worlds_joinable w1 w2 w -> R w w' ->
       (exists w1', exists w2', R w1 w1' /\ R w2 w2' /\ join worlds_joinable w1' w2' w').

  Definition lax_monoidal2 (R:relation world)  :=
    forall w1 w2 w w',
       join worlds_joinable w1 w2 w -> R w' w ->
       (exists w1', exists w2', R w1' w1 /\ R w2' w2 /\ join worlds_joinable w1' w2' w').

  Axiom lax_monoidial_distrib_pos : forall p q z (M:modality z), is_pos z ->
    lax_monoidal1 (modal_rel M) ->
    M p * M q |-- M (p * q).

  Axiom lax_monoidial_distrib_neg : forall p q z (M:modality z), is_neg z ->
    lax_monoidal2 (modal_rel M) ->
    M (p * q) |-- M p * M q.

  Axiom modal_distrib_sep_pos : forall z (M:modality z) p q, is_pos z ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->
    lax_monoidal1 (modal_rel M) ->
    M (M p * M q) = M p * M q.

  Axiom modal_distrib_sep_neg : forall z (M:modality z) p q, is_neg z ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->
    lax_monoidal2 (modal_rel M) ->
    M (M p * M q) = M p * M q.

  Definition same_age (c1 c2: world) : Prop :=
    exists e, join worlds_joinable e c1 c1 /\ join worlds_joinable e c2 c2.

  Parameter fashionably : modality pol_pos.
  Axiom fashionably_spec : modal_op fashionably = appPosMode same_age.
  Axiom fashionably_rel_spec : modal_rel fashionably = same_age.

  Notation "'#' e" := (fashionably e) (at level 30, right associativity): pred_scope.

  Axiom fashionably_refl : reflexive _ (modal_rel fashionably).
  Axiom fashionably_trans : transitive _ (modal_rel fashionably).
  Axiom fashionably_sym : symmetric _ (modal_rel fashionably).
  Axiom fashionably_lax1 : lax_monoidal1 (modal_rel fashionably).

  Hint Resolve fashionably_refl fashionably_trans
     fashionably_sym fashionably_lax1.

  Axiom fashion_elim : forall p q r G,
    [ #p :: #q :: G ] |-- r ->
    [ #p * #q :: G ] |-- r.

  Axiom fashion_intro : forall p q G,
    [ G ] |-- #p ->
    [ G ] |-- #q ->
    [ G ] |-- #p * #q.

  Axiom fashion_conj : forall p q,
     #p * #q = #p && #q.

  Definition pure (p:pred) := #p && emp.

  Axiom pure_elim : forall p q r G,
    [ #p :: q ::  G ] |-- r ->
    [ pure p * q :: G ] |-- r.

  Axiom pure_intro : forall p q G,
    [ G ] |-- #p ->
    [ G ] |-- q ->
    [ G ] |-- pure p * q.

  Axiom pure_conj : forall p q,
     pure p * q = #p && q.

  Axiom nec_lax1 : lax_monoidal1 (modal_rel necM).
  Axiom eventually_lax2 : lax_monoidal2 (modal_rel eventuallyM).

  Hint Resolve nec_lax1 eventually_lax2.


  Axiom join_tick_n_compatible : forall n w1 w2 w w1' w2' w',
     tick_world_n n w1 = Some w1' ->
     tick_world_n n w2 = Some w2' ->
     tick_world_n n w  = Some w'  ->
     (join worlds_joinable w1 w2 w <-> join worlds_joinable w1' w2' w').

  Axiom tick_n_joinable_succeeds : forall n w1 w2 w,
     join worlds_joinable w1 w2 w ->
     ((exists w1', tick_world_n n w1 = Some w1') <->
      (exists w2', tick_world_n n w2 = Some w2')).

  Axiom tick_n_joined_succeeds : forall n w1 w2 w,
     join worlds_joinable w1 w2 w ->
     ((exists w1', tick_world_n n w1 = Some w1') <->
      (exists w',  tick_world_n n w = Some w')).

  Axiom eventually_nec_sep : forall p q,
    (<>[]p) * (<>[]q) = <>([]p * []q).

End SEP_LOGIC_THEORY.


Module SepLogicTheory
  (W':WORLDS)
  (WS': WORLD_STEP with Module W:=W')
  (WJ': JOINABLE_WORLDS with Module W:=W')
  : SEP_LOGIC_THEORY.

  Module W:=W'.
  Module WJ:=WJ'.

  Module ST := ProgramSafetyTheory W WS'.

  Export WJ.
  Export ST.

  Open Local Scope pred_scope.

  Definition emp : pred :=
    fun c => is_empty worlds_joinable c.
  Definition sep_con (p q:pred) : pred :=
    fun c => exists c1, exists c2, join worlds_joinable c1 c2 c /\ p c1 /\ q c2.
  Definition sep_impl (p q:pred) : pred :=
    fun c => forall c1, forall c2, join worlds_joinable c c1 c2 -> p c1 -> q c2.

  Notation "P '*' Q" := (sep_con P Q) (at level 40, left associativity) : pred_scope.
  Infix "--*" := sep_impl (at level 60, right associativity) : pred_scope.

  Lemma sep_comm_derives  : forall p q, p * q |-- q * p.
    unfold derives, sep_con; simpl; intuition.
    destruct H0 as [w1 [w2 ?]]; intuition.
    exists w2; exists w1; intuition.
    apply (join_com worlds_joinable); auto.
   Qed.

  Lemma sep_comm : forall p q, p * q = q * p.
    intros; apply derives_equal; apply sep_comm_derives.
   Qed.

  Lemma sep_assoc : forall p q r, p * (q * r) = (p * q) * r.
    intros; apply derives_equal; unfold derives; simpl; intuition.
    destruct H0 as [w1 [w2 ?]]; intuition.
    destruct H3 as [w3 [w4 ?]]; intuition.
    destruct (join_assoc worlds_joinable w4 w3 w1 w2 w) as [w' ?]; intuition.
    apply join_com; auto.
    apply join_com; auto.
    exists w'.
    exists w4.
    intuition.
    apply join_com; auto.
    exists w1.
    exists w3.
    intuition.
    apply join_com; auto.

    destruct H0 as [w1 [w2 ?]]; intuition.
    destruct H as [w3 [w4 ?]]; intuition.
    destruct (join_assoc worlds_joinable w3 w4 w2 w1 w); intuition.
    exists w3; exists x; intuition.
    exists w4; exists w2; intuition.
   Qed.

  Lemma sep_unit1 : forall p, emp * p = p.
    intros; apply derives_equal; unfold derives; simpl; intuition.
    destruct H0 as [w1 [w2 ?]]; intuition.
    replace w2 with w in H3; auto.
    symmetry.
    apply (join_empty worlds_joinable) with w1; auto.
    destruct (exists_empty worlds_joinable w) as [e ?]; intuition.
    exists e; exists w; intuition.
   Qed.

  Lemma sep_unit2 : forall p, p * emp = p.
    intros; rewrite sep_comm.
    apply sep_unit1.
   Qed.

  Lemma sep_cut1 : forall p' p q r,
    p |-- p' ->
    p' * q |-- r ->
    p  * q |-- r.

    unfold derives; simpl; intuition.
    apply H0; intuition.
    destruct H2 as [w1 [w2 ?]]; intuition.
    exists w1; exists w2; intuition.
   Qed.

  Lemma sep_cut2 : forall q' p q r,
    q |-- q' ->
    p * q' |-- r ->
    p * q  |-- r.

    unfold derives; simpl; intuition.
    apply H0; intuition.
    destruct H2 as [w1 [w2 ?]]; intuition.
    exists w1; exists w2; intuition.
   Qed.

  Lemma sep_split : forall p q p' q',
    p |-- p' ->
    q |-- q' ->
    p * q |-- p' * q'.

    unfold derives; simpl; intuition.
    destruct H2 as [w1 [w2 ?]]; intuition.
    exists w1; exists w2; intuition.
   Qed.

  Lemma wand_intro : forall p q r,
    p * q |-- r ->
    p |-- q --* r.

    unfold derives; simpl; intuition.
    red; intros.
    apply H; intuition.
    exists w; exists c1; intuition.
   Qed.

  Lemma wand_elim : forall p q r s,
    r |-- p ->
    q |-- s ->
    (p --* q) * r |-- s.

    unfold derives; simpl; intuition.
    destruct H2 as [w1 [w2 ?]]; intuition.
    apply H0; intuition.
    red in H1; apply H1 with w2; auto.
   Qed.

  Lemma wand_elim_emp : forall p q s,
    emp |-- p ->
    q |-- s ->
    (p --* q) |-- s.

    intros.
    rewrite <- (sep_unit2 (p --* q)).
    apply wand_elim; auto.
   Qed.

  Lemma join_tick_n_compatible : forall n w1 w2 w w1' w2' w',
     tick_world_n n w1 = Some w1' ->
     tick_world_n n w2 = Some w2' ->
     tick_world_n n w  = Some w'  ->
     (join worlds_joinable w1 w2 w <-> join worlds_joinable w1' w2' w').

    induction n; simpl; intros.
    replace w1' with w1; [ | congruence ].
    replace w2' with w2; [ | congruence ].
    replace w' with w; [ | congruence ].
    intuition.

    case_eq (tick_world w1); intros; rewrite H2 in H; try discriminate.
    case_eq (tick_world w2); intros; rewrite H3 in H0; try discriminate.
    case_eq (tick_world w); intros; rewrite H4 in H1; try discriminate.
    rewrite <- (IHn w0 w3 w4 w1' w2' w'); auto.
    apply join_tick_compatible; auto.
   Qed.

  Lemma tick_n_joinable_succeeds : forall n w1 w2 w,
     join worlds_joinable w1 w2 w ->
     ((exists w1', tick_world_n n w1 = Some w1') <->
      (exists w2', tick_world_n n w2 = Some w2')).

    induction n; simpl; intros; split; intros.
    exists w2; auto.
    exists w1; auto.

    destruct H0.
    case_eq (tick_world w1); intros; rewrite H1 in H0; try discriminate.
    destruct (tick_joinable_succeeds w1 w2 w); auto.
    destruct H2; clear H3.
    exists w0; auto.
    change (tick_world w2 = Some x0) in H2.
    rewrite H2.
    case_eq (tick_world w); intros.
    destruct (join_tick_compatible w1 w2 w w0 x0 w3); auto.
    destruct (IHn w0 x0 w3 (H4 H)).
    apply H6.
    exists x; auto.
    destruct (tick_joined_succeeds w1 w2 w); auto.
    destruct H4; clear H5.
    exists w0; auto.
    change (tick_world w = Some x1) in H4.
    rewrite H3 in H4; discriminate.

    destruct H0.
    case_eq (tick_world w2); intros; rewrite H1 in H0; try discriminate.
    destruct (tick_joinable_succeeds w2 w1 w); auto.
    apply join_com; auto.
    destruct H2; clear H3.
    exists w0; auto.
    change (tick_world w1 = Some x0) in H2.
    rewrite H2.
    case_eq (tick_world w); intros.
    destruct (join_tick_compatible w1 w2 w x0 w0 w3); auto.
    destruct (IHn x0 w0 w3 (H4 H)).
    apply H7.
    exists x; auto.
    destruct (tick_joined_succeeds w2 w1 w); auto.
    apply join_com; auto.
    destruct H4; clear H5.
    exists w0; auto.
    change (tick_world w = Some x1) in H4.
    rewrite H3 in H4; discriminate.
   Qed.


  Lemma tick_n_joined_succeeds : forall n w1 w2 w,
     join worlds_joinable w1 w2 w ->
     ((exists w1', tick_world_n n w1 = Some w1') <->
      (exists w',  tick_world_n n w = Some w')).

     induction n; simpl; intros; split; intros.
     exists w; auto.
     exists w1; auto.

     destruct H0.
     case_eq (tick_world w1); intros; rewrite H1 in H0; try discriminate.
     destruct (tick_joined_succeeds w1 w2 w); auto.
     destruct H2; clear H3.
     exists w0; auto.
     change (tick_world w = Some x0) in H2.
     rewrite H2.
     case_eq (tick_world w2); intros.
     destruct (IHn w0 w3 x0).
     destruct (join_tick_compatible w1 w2 w w0 w3 x0); auto.
     apply H4.
     exists x; auto.
     destruct (tick_joinable_succeeds w1 w2 w); auto.
     destruct H4; clear H5.
     exists w0; auto.
     change (tick_world w2 = Some x1) in H4.
     rewrite H3 in H4; discriminate.

     destruct H0.
     case_eq (tick_world w); intros; rewrite H1 in H0; try discriminate.
     destruct (tick_joined_succeeds w1 w2 w); auto.
     destruct H3; clear H2.
     exists w0; auto.
     change (tick_world w1 = Some x0) in H3.
     rewrite H3.
     case_eq (tick_world w2); intros.
     destruct (IHn x0 w3 w0).
     destruct (join_tick_compatible w1 w2 w x0 w3 w0); auto.
     apply H5.
     exists x; auto.
     destruct (tick_joinable_succeeds w1 w2 w); auto.
     destruct H4; clear H5.
     exists x0; auto.
     change (tick_world w2 = Some x1) in H4.
     rewrite H2 in H4; discriminate.
    Qed.


  Definition lax_monoidal1 (R:relation world)  :=
    forall w1 w2 w w',
       join worlds_joinable w1 w2 w -> R w w' ->
       (exists w1', exists w2', R w1 w1' /\ R w2 w2' /\ join worlds_joinable w1' w2' w').

  Definition lax_monoidal2 (R:relation world)  :=
    forall w1 w2 w w',
       join worlds_joinable w1 w2 w -> R w' w ->
       (exists w1', exists w2', R w1' w1 /\ R w2' w2 /\ join worlds_joinable w1' w2' w').

  Lemma lax_monoidial_distrib_pos : forall p q z (M:modality z), is_pos z ->
    lax_monoidal1 (modal_rel M) ->
    M p * M q |-- M (p * q).

    unfold derives; simpl; intuition.
    clear H3.
    destruct H2 as [w1 [w2 ?]]; intuition.
    destruct (pos_modal_spec _ M H (p * q) w).
    apply H3; clear H3 H5; intros.
    destruct (pos_modal_spec _ M H p w1).
    generalize (H6 H1); clear H5 H6; intros.
    destruct (pos_modal_spec _ M H q w2).
    generalize (H7 H4); clear H6 H7; intros.
    destruct (H0 w1 w2 w w' H2 H3) as [w1' [w2' ?]].
    exists w1'; exists w2'; intuition.
   Qed.

  Lemma lax_monoidial_distrib_neg : forall p q z (M:modality z), is_neg z ->
    lax_monoidal2 (modal_rel M) ->
    M (p * q) |-- M p * M q.

    unfold derives; simpl; intuition.
    clear H3.
    destruct (neg_modal_spec _ M H (p * q) w).
    generalize (H3 H2); clear H3 H1; intros.
    destruct H1 as [w' ?]; intuition.
    destruct H4 as [w1' [w2' ?]]; intuition.
    destruct (H0 w1' w2' w' w H4 H3) as [w1 [w2 ?]]; intuition.
    exists w1; exists w2; intuition.
    destruct (neg_modal_spec _ M H p w1).
    apply H8; clear H8 H10.
    exists w1'; auto.
    destruct (neg_modal_spec _ M H q w2).
    apply H8; clear H8 H10.
    exists w2'; auto.
   Qed.

  Lemma modal_distrib_sep_pos : forall z (M:modality z) p q, is_pos z ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->
    lax_monoidal1 (modal_rel M) ->
    M (M p * M q) = M p * M q.

    intros; apply derives_equal.
    apply modal_T; auto.
    apply_d lax_monoidial_distrib_pos.
    apply sep_split; apply modal_4; auto.
   Qed.

  Lemma modal_distrib_sep_neg : forall z (M:modality z) p q, is_neg z ->
    reflexive _ (modal_rel M) ->
    transitive _ (modal_rel M) ->
    lax_monoidal2 (modal_rel M) ->
    M (M p * M q) = M p * M q.

    intros; apply derives_equal.
    in_goal rewrite <- (modal_idem z M p); auto.
    in_goal rewrite <- (modal_idem z M q); auto.
    apply_d lax_monoidial_distrib_neg; auto.
    apply_d nmodal_T; auto.
   Qed.

  Definition same_age (c1 c2: world) : Prop :=
    exists e, join worlds_joinable e c1 c1 /\ join worlds_joinable e c2 c2.

  Lemma same_age_sym : symmetric _ same_age.
    red; unfold same_age; intros.
    destruct H as [e ?]; intuition.
    exists e; intuition.
   Qed.

  Lemma same_age_refl : reflexive _ same_age.
    repeat red; intros.
    destruct (exists_empty worlds_joinable x) as [a [H1 H2]].
    exists a; auto.
   Qed.

  Lemma same_age_trans : transitive _ same_age.
    red; unfold same_age; intros.
    destruct H as [a1 [H1 H2]].
    destruct H0 as [a2 [H3 H4]].
    cut (is_empty worlds_joinable a1)
       ; [ | apply join_empty3 with x; auto ].
    cut (is_empty worlds_joinable a2)
       ; [ | apply join_empty3 with y; auto ].
    intros.
    cut (a1=a2); [ | apply cancellation with worlds_joinable y y; auto ].
    intro; subst a1.
    exists a2; auto.
   Qed.

  Lemma same_age_lax1 : lax_monoidal1 same_age.
    unfold lax_monoidal1, same_age; intros.
    destruct H0 as [e ?]; intuition.
    exists e; exists w'; intuition.
    exists e; split.
    destruct (join_assoc worlds_joinable w2 w1 e w w); intuition.
    apply join_com; auto.
    apply join_com; auto.
    replace x with w1 in H3.
    apply join_com; auto.
    symmetry.
    apply (cancellation worlds_joinable) with w2 w; auto.
    apply join_com; auto.
    destruct (exists_empty worlds_joinable e); intuition.
    replace x with e in H4; auto.
    symmetry.
    apply (join_empty worlds_joinable e).
    apply (join_empty3 worlds_joinable _ _ H1).
    apply join_com; auto.
    exists e; split.
    destruct (join_assoc worlds_joinable w1 w2 e w w); intuition.
    apply join_com; auto.
    replace x with w2 in H3.
    apply join_com; auto.
    symmetry.
    apply (cancellation worlds_joinable) with w1 w; auto.
    apply join_com; auto.
    apply join_com; auto.
    auto.
   Qed.

  Definition fashionably : modality pol_pos := mkPosModality same_age.
  Lemma fashionably_spec : modal_op fashionably = appPosMode same_age.
     auto.
   Qed.
  Lemma fashionably_rel_spec : modal_rel fashionably = same_age.
     auto.
   Qed.
  Opaque fashionably.
  Notation "'#' e" := (fashionably e) (at level 30, right associativity): pred_scope.

  Lemma fashionably_refl : reflexive _ (modal_rel fashionably).
    rewrite fashionably_rel_spec.
    apply same_age_refl.
   Qed.

  Lemma fashionably_trans : transitive _ (modal_rel fashionably).
    rewrite fashionably_rel_spec.
    apply same_age_trans.
   Qed.

  Lemma fashionably_sym : symmetric _ (modal_rel fashionably).
    rewrite fashionably_rel_spec.
    apply same_age_sym.
   Qed.

  Lemma fashionably_lax1 : lax_monoidal1 (modal_rel fashionably).
    rewrite fashionably_rel_spec.
    apply same_age_lax1.
   Qed.

  Hint Resolve fashionably_refl fashionably_trans
     fashionably_sym fashionably_lax1.

  Lemma joined_same_age : forall w1 w2 w,
    join worlds_joinable w1 w2 w ->
    same_age w1 w.

    intros; red.
    destruct (exists_empty worlds_joinable w1) as [e ?].
    exists e; intuition.
    destruct (join_assoc worlds_joinable e w1 w2 w1 w) as [w' ?]; intuition.
    replace w' with w in H4; auto.
    apply (join_eq worlds_joinable w1 w2); auto.
   Qed.

  Lemma joinable_same_age : forall w1 w2 w,
    join worlds_joinable w1 w2 w ->
    same_age w1 w2.

    intros; red.
    destruct (exists_empty worlds_joinable w1) as [e ?].
    exists e; intuition.
    destruct (exists_empty worlds_joinable w2) as [e' ?].
    intuition.
    replace e' with e in H4; auto.
    destruct (join_assoc worlds_joinable e w1 w2 w1 w) as [w' ?]; intuition.
    destruct (join_assoc worlds_joinable e' w2 w1 w2 w) as [w'' ?]; intuition.
    apply (join_com worlds_joinable); auto.
    apply (cancellation worlds_joinable e e' w w).
    replace w' with w in H6; auto.
    apply (join_eq worlds_joinable w1 w2); auto.
    replace w'' with w in H8; auto.
    apply (join_eq worlds_joinable w1 w2); auto.
    apply (join_com worlds_joinable); auto.
   Qed.

  Lemma fashion_elim1 : forall p q r G,
    [ p :: q  :: G ] |-- r ->
    [ #p * #q :: G ] |-- r.

    intros p q r G.
    unfold derives.
    rewrite fashionably_spec.
    simpl; intuition.
    apply H; intuition.
    destruct H1 as [w1 [w2 ?]]; intuition.
    apply H0.
    apply joined_same_age with w2; auto.
    destruct H1 as [w1 [w2 ?]]; intuition.
    apply H4.
    apply joined_same_age with w1; auto.
    apply (join_com worlds_joinable); auto.
   Qed.

  Lemma fashion_elim : forall p q r G,
    [ #p :: #q  :: G ] |-- r ->
    [ #p * #q :: G ] |-- r.

    intros.
    rewrite <- (modal_idem _ fashionably p); auto.
    rewrite <- (modal_idem _ fashionably q); auto.
    apply fashion_elim1.
    auto.
   Qed.

  Lemma fashion_intro : forall p q G,
    [ G ] |-- #p ->
    [ G ] |-- #q ->
    [ G ] |-- #p * #q.

   intros p q G; unfold derives; rewrite fashionably_spec.
   simpl; intuition.
   red.
   destruct (exists_empty worlds_joinable w) as [e ?]; intuition.
   exists e; exists w; intuition.
   red; intros.
   apply H with w; auto.
   revert H2.
   cut (same_age w e).
   apply same_age_trans.
   apply same_age_sym.
   apply joinable_same_age with w; auto.
  Qed.

  Lemma fashion_conj : forall p q,
     #p * #q = #p && #q.

    intros; apply derives_equal.
    apply fashion_elim; simpl_derives.
    apply fashion_intro; simpl_derives.
   Qed.

  Definition pure (p:pred) := #p && emp.

  Lemma pure_elim1 : forall p q r G,
    [ p :: q ::  G ] |-- r ->
    [ pure p * q :: G ] |-- r.

    intros p q r G.
    unfold pure, derives.
    rewrite fashionably_spec.
    simpl; intuition.
    destruct H1 as [w1 [w2 ?]]; intuition.
    destruct H0.
    red in H3.
    red in H0.
    apply H; intuition.
    apply H0.
    apply joined_same_age with w2; auto.
    replace w2 with w in H4; auto.
    symmetry.
    apply (join_empty worlds_joinable) with w1; auto.
   Qed.

  Lemma pure_elim : forall p q r G,
    [ #p :: q ::  G ] |-- r ->
    [ pure p * q :: G ] |-- r.

    intros; unfold pure.
    rewrite <- modal_idem; auto.
    change ([ pure (#p) * q :: G ] |-- r).
    apply pure_elim1; auto.
   Qed.

  Lemma pure_intro : forall p q G,
    [ G ] |-- #p ->
    [ G ] |-- q ->
    [ G ] |-- pure p * q.

    intros p q G; unfold derives, pure.
    rewrite fashionably_spec.
    simpl; intuition.
    destruct (exists_empty worlds_joinable w) as [e ?]; intuition.
    exists e; exists w; intuition.
    split; auto.
    red; intros.
    red in H.
    apply H with w; auto.
    revert H2.
    cut (same_age w e).
    apply same_age_trans.
    apply same_age_sym.
    apply joined_same_age with w; auto.
   Qed.

  Lemma pure_conj : forall p q,
     pure p * q = #p && q.

    intros; apply derives_equal.
    apply pure_elim; intuition_d.
    apply pure_intro; intuition_d.
   Qed.

  Lemma nec_rel_lax1 : lax_monoidal1 nec_rel.

    red; unfold nec_rel; intros.
    destruct H0 as [n Hn].
    destruct (tick_n_joined_succeeds n w1 w2 w); auto.
    destruct H1 as [w1' ?]; clear H0.
    exists w'; auto.
    destruct (tick_n_joined_succeeds n w2 w1 w); auto.
    apply join_com; auto.
    destruct H2 as [w2' ?]; clear H0.
    exists w'; auto.
    exists w1'; exists w2'; intuition.
    exists n; auto.
    exists n; auto.
    rewrite <- (join_tick_n_compatible n w1 w2 w w1' w2' w'); auto.
   Qed.

  Lemma nec_rel_lax2 : lax_monoidal2 nec_rel.

    red; unfold nec_rel; intros.
    destruct H0 as [n Hn].
    destruct (untick_n n w1) as [w1' ?].
    destruct (untick_n n w2) as [w2' ?].
    exists w1'; exists w2'; intuition.
    exists n; auto.
    exists n; auto.
    rewrite (join_tick_n_compatible n w1' w2' w' w1 w2 w); auto.
   Qed.

  Lemma nec_lax1 : lax_monoidal1 (modal_rel necM).
    rewrite necM_rel_spec.
    apply nec_rel_lax1.
   Qed.

  Lemma eventually_lax2 : lax_monoidal2 (modal_rel eventuallyM).
    rewrite eventually_rel_spec.
    apply nec_rel_lax2.
   Qed.

  Hint Resolve nec_lax1 eventually_lax2.

  Lemma eventually_nec_sep : forall p q,
    (<>[]p) * (<>[]q) = <>([]p * []q).

    intros.
    apply derives_equal;
     [ | apply lax_monoidial_distrib_neg; auto ].
    unfold derives; simpl; intuition.
    clear H1.
    destruct H0 as [w1 [w2 ?]]; intuition.
    destruct (neg_modal_spec _ eventuallyM I ([]p) w1).
    generalize (H3 H); clear H1 H3; intros.
    destruct H1 as [w1' ?]; intuition.
    destruct (neg_modal_spec _ eventuallyM I ([]q) w2).
    generalize (H5 H2); clear H1 H5; intros.
    destruct H1 as [w2' ?]; intuition.
    rewrite eventually_rel_spec in H3.
    rewrite eventually_rel_spec in H5.
    destruct (neg_modal_spec _ eventuallyM I ([]p * []q) w).
    apply H1; clear H1 H7.
    destruct H3 as [n1 ?].
    destruct H5 as [n2 ?].
    set (n := max n1 n2).
    case_eq (tick_world_n n w); intros.
    case_eq (tick_world_n n w1); intros.
    case_eq (tick_world_n n w2); intros.
    exists w0; split.
    rewrite eventually_rel_spec.
    exists n; auto.
    exists w3; exists w4; intuition.
    rewrite <- (join_tick_n_compatible n w1 w2 w w3 w4 w0); auto.
    destruct (pos_modal_spec _ necM I p w3).
    apply H9; clear H9 H10; intros.
    destruct (pos_modal_spec _ necM I p w1').
    apply H11; clear H10 H11; auto.
    apply necM_trans with w3; auto.
    rewrite necM_rel_spec.
    destruct (tick_n_split n n1 (n - n1) w1 w3); intuition.
    cut (n1 <= n).
    omega.
    unfold n; apply le_max_l.
    replace x with w1' in H12; [ | congruence ].
    exists (n-n1); auto.
    destruct (pos_modal_spec _ necM I q w4).
    apply H9; clear H9 H10; intros.
    destruct (pos_modal_spec _ necM I q w2').
    apply H11; clear H10 H11; auto.
    apply necM_trans with w4; auto.
    rewrite necM_rel_spec.
    destruct (tick_n_split n n2 (n - n2) w2 w4); intuition.
    cut (n2 <= n).
    omega.
    unfold n; apply le_max_r.
    replace x with w2' in H12; [ | congruence ].
    exists (n-n2); auto.
    destruct (tick_n_joinable_succeeds n w1 w2 w); auto.
    destruct H9; clear H10; auto.
    exists w3; auto.
    rewrite H8 in H9; discriminate.
    destruct (tick_n_joined_succeeds n w1 w2 w); auto.
    destruct H9; clear H8.
    exists w0; auto.
    rewrite H7 in H9; discriminate.
    destruct (max_dec n1 n2).
    destruct (tick_n_joined_succeeds n1 w1 w2 w); auto.
    destruct H7; clear H8.
    exists w1'; auto.
    unfold n in H5; rewrite e in H5.
    rewrite H5 in H7; discriminate.
    destruct (tick_n_joined_succeeds n2 w2 w1 w); auto.
    apply join_com; auto.
    destruct H7; clear H8.
    exists w2'; auto.
    unfold n in H5; rewrite e in H5.
    rewrite H5 in H7; discriminate.
   Qed.

(*
  Inductive SepSimpl_Rep : Type :=
    | SSR_Lift    : pred -> SepSimpl_Rep
    | SSR_Sep     : SepSimpl_Rep -> SepSimpl_Rep -> SepSimpl_Rep
    | SSR_Pure    : pred -> SepSimpl_Rep
    | SSR_Fashion : SepSimpl_Rep -> SepSimpl_Rep
    .

  Fixpoint interpSepSimpl (x:SepSimpl_Rep) : pred :=
    match x with
    | SSR_Lift z    => z
    | SSR_Sep p q   => (interpSepSimpl p) * (interpSepSimpl q)
    | SSR_Pure p    => pure p
    | SSR_Fashion p => #(interpSepSimpl p)
    end.

  Ltac uninterpSepSimpl X ID :=
    match X with
      | pure ?p => set (ID:=SSR_Pure p)
      | @modal_op pol_pos fashionably ?p =>
          let p' := fresh in
          uninterpSepSimpl p p';
          set (ID:=SSR_Fashion p');
          subst p'
      | sep_con ?p ?q =>
          let p' := fresh in
          uninterpSepSimpl p p';
          let q' := fresh in
          uninterpSepSimpl q q';
          set (ID:=SSR_Sep p' q');
          subst p'; subst q'
      | _ => set (ID:=SSR_Lift X)
      end.

Definition fashop x := modal_op fashionably x.

  Fixpoint simplSepSimpl (x:SepSimpl_Rep) (fash:bool) { struct x } : list pred * list pred * list pred :=
    match x with
    | SSR_Pure p    => (p::nil, nil, nil)
    | SSR_Lift p    =>
       if fash
        then (nil, p::nil, nil)
        else (nil, nil, p::nil)
    | SSR_Sep p q   =>
       match simplSepSimpl p false with (x1, y1, z1) =>
       match simplSepSimpl q false with (x2, y2, z2) =>
       if fash
        then (x1++x2, y1++y2, z1++z2)
        else (x1++x2, nil   , map fashop y1 ++ z1 ++
                              map fashop y2 ++ z2 )
       end end

    | SSR_Fashion p => simplSepSimpl p true
    end.
    
  Fixpoint sep_collect' (x:pred) (l:list pred) {struct l} : pred :=
     match l with
     | nil   => x
     | y::ys => x * sep_collect' y ys
     end.

  Definition sep_collect (l:list pred) : list pred :=
     match l with
     | nil => nil
     | x::xs => (sep_collect' x xs)::nil
     end.

  Fixpoint sep_collect2 (l:list pred) : pred :=
     match l with
     | nil => emp
     | x::xs => x * sep_collect2 xs
     end.

  Lemma sep_collect2_app : forall l1 l2,
     sep_collect2 (l1++l2) = sep_collect2 l1 * sep_collect2 l2.

    induction l1; simpl; intros.
    rewrite sep_unit1; auto.
    rewrite IHl1; auto.
    rewrite sep_assoc.
    auto.
   Qed.

  Lemma all_true_app : forall l1 l2,
     all_true (l1++l2) = all_true l1 && all_true l2.

     induction l1; simpl; intros.
     refine (extensionality _ _ _); intros.
     apply prop_ext; unfold conj; intuition.
     refine (extensionality _ _ _); intros.
     rewrite IHl1.
     unfold conj.
     apply prop_ext; intuition.
    Qed.

  Lemma sep_collect'_move : forall l a,
     sep_collect' a l = a * sep_collect' emp l.

     induction l; simpl; intros.
     rewrite sep_unit2; auto.
     rewrite IHl.
     repeat rewrite sep_assoc.
     rewrite sep_unit2.
     auto.
    Qed.

  Lemma all_true_one : forall a,
     all_true (a::nil) = a.

     simpl; intros.
     refine (extensionality _ _ _); intros.
     apply prop_ext; intuition.
    Qed.

  Lemma map_fashop_fashion : forall l,
    all_true (map fashop l) = 
    # (all_true (map fashop l)).

    induction l; intros.
    apply derives_equal.
    apply derives_axiom; apply modal_N; auto.
    apply modal_T; auto.

    change (a::l) with ((a::nil)++l).
    repeat rewrite map_app.
    repeat rewrite all_true_app.

    change (map fashop (a::nil)) with 
        (#a :: nil).
    rewrite all_true_one.
    rewrite pmodal_conj; auto.
    rewrite modal_idem; auto.    
    rewrite <- IHl.
    auto.
   Qed.
   
  Lemma sep_collect_eq : forall l a,
     sep_collect' a l = sep_collect2 (a::l).

     induction l; simpl; intros.
     rewrite sep_unit2; auto.
     rewrite IHl; auto.
    Qed.

(*
  Lemma sep_collect_app : forall l1 l2,
     sep_collect (l1++l2) = sep_collect (sep_collect l1 ++ sep_collect l2 ++ nil).
     
     intros; destruct l1; simpl.
     rewrite <- app_nil_end.
     destruct l2; simpl; auto.
     symmetry.
     rewrite sep_collect'_move.
     rewrite <- app_nil_end.
     rewrite (sep_collect'_app l1 l2 p emp p).
     destruct l2; simpl; auto.
     apply sep_unit2.
    Qed.
*)

  Definition fash_conj (l:list pred) :=
     fold_right conj top (map fashop l).

  Lemma simplSepSimpl_sound : forall (x:SepSimpl_Rep) a b c,
     simplSepSimpl x false = (a, b, c) ->
     interpSepSimpl x = all_true (map fashop a ++ (sep_collect2 (fash_conj b::c) :: nil)).

     induction x; intros.
     simpl in *.
     injection H; intros; subst.
     unfold fash_conj.
     simpl.
     refine (extensionality _ _ _); intros.
     apply prop_ext; intuition.
     rewrite sep_unit2; auto.


     rewrite sep_unit2 in H1; auto.

     simpl in H.
     case_eq (simplSepSimpl x1 false); intros p c1 ?.
     rewrite H0 in H.
     destruct p as [a1 b1].
     case_eq (simplSepSimpl x2 false); intros p c2 ?.
     rewrite H1 in H.
     destruct p as [a2 b2].
     injection H; intros; subst.
     
     simpl.
     rewrite map_app.
     repeat rewrite all_true_app.
     replace (map fashop b1 ++ c1 ++ map fashop b2 ++ c2) with
        ((map fashop b1 ++ c1) ++ (map fashop b2 ++ c2)).
     rewrite sep_collect2_app.
     rewrite all_true_one.
     rewrite (map_fashop_fashion a1).
     rewrite (map_fashop_fashion a2).
     rewrite <- fashion_conj.





     rewrite IHx1 with a1 b1 c1; auto.
     rewrite IHx2 with a2 b2 c2; auto.



     

     
*)     
    



     

    




(*
  Goal forall p q r s,
     (p * q) * (#(#r * q) * pure p) |-- s.

     intros.
     match goal with
      [ |- derives (?X :: ?G) ?t ] =>
        let X':=fresh in uninterpSepSimpl X X';
        change X with (interpSepSimpl X');
        subst X'
      end.
*)


End SepLogicTheory.
