(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import ModelSpec.

Module SimpleWorlds (WCWH:WORLD_CARRIER_WITH_HOLE) <: WORLDS.
  
  Module WC <: WORLD_CARRIER.
    Definition world := WCWH.world_functor nat.
    Definition sample_world := WCWH.sample_world nat 0.

    Definition field := WCWH.field.
    Definition field_enum := WCWH.field_enum.
    Definition field_enum_injective := WCWH.field_enum_injective.
    Definition field_type := WCWH.field_type.
    Definition field_eq_dec := WCWH.field_eq_dec.

    Definition w_upd := WCWH.w_upd nat.
    Definition w_proj := WCWH.w_proj nat.
  
    Lemma proj_upd_same : forall x a w,
      w_proj x (w_upd x a w) = a.

      intros; refine (WCWH.proj_upd_same nat _ _ _).
     Qed.

    Lemma upd_proj_same : forall x w,
      w_upd x (w_proj x w) w = w.

      intros; refine (WCWH.upd_proj_same _ _ _).
     Qed.

    Lemma proj_upd_diff : forall x y a w,
      x <> y ->
      w_proj x (w_upd y a w) = w_proj x w.

      intros; refine (WCWH.proj_upd_diff nat _ _ _ _ _); auto.
     Qed.

    Lemma upd_collapse : forall x a b w,
      w_upd x a (w_upd x b w) =
      w_upd x a w.

      intros; refine (WCWH.upd_collapse nat _ _ _ _).
     Qed.

    Lemma upd_injective : forall x a a' w w',
       w_upd x a w = w_upd x a' w' -> a = a'.

       intros; refine (WCWH.upd_injective nat x a a' w w' _); auto.
      Qed.

    Lemma upd_eq : forall x a b w w',
     w_upd x a w = w_upd x a w' ->
     w_upd x b w = w_upd x b w'.

       intros; refine (WCWH.upd_eq nat x a b _ _ _); auto.
      Qed.

    Lemma upd_commute : forall x y a b w,
      x <> y ->
      w_upd x a (w_upd y b w) =
      w_upd y b (w_upd x a w).

      intros; refine (WCWH.upd_commute nat _ _ _ _ _ _); auto.
     Qed.

  End WC.

  Module WT <: WORLD_TICK with Module WC:=WC.
    Module WC:=WC.

    Import WC.

    Definition world_ttl := WCWH.get_hole nat.
    Definition tick_world (w:WC.world) :=
      match world_ttl w with
        | 0 => None
        | S n => Some (WCWH.set_hole nat nat n w)
        end.

    Lemma tick_ttl : forall w w',
      tick_world w = Some w' -> 
      world_ttl w = S (world_ttl w').

      unfold tick_world, world_ttl; intros.
      case_eq (WCWH.get_hole nat w); intros; rewrite H0 in H; try discriminate.
      injection H; intros; subst w'.
      rewrite WCWH.get_set_hole; auto.
     Qed.

    Lemma tick_succeeds : forall w,
      world_ttl w > 0 ->
      exists w', tick_world w = Some w'.

      unfold tick_world, world_ttl; intros.
      destruct (WCWH.get_hole nat w).
      inversion H.
      exists (WCWH.set_hole nat nat n w).
      auto.
     Qed.

    Lemma untick : forall w', exists w, tick_world w = Some w'.
      
      unfold tick_world, world_ttl; intros.
      exists (WCWH.set_hole nat nat (S (WCWH.get_hole nat w')) w').
      rewrite WCWH.get_set_hole.
      rewrite WCWH.set_hole_collapse.
      rewrite WCWH.set_get_hole_same.
      auto.
     Qed.

    Lemma tick_inj : forall w1 w2 w,
      tick_world w1 = Some w ->
      tick_world w2 = Some w ->
      w1 = w2.

      unfold tick_world, world_ttl; intros.
      case_eq (WCWH.get_hole nat w1); intros;
        rewrite H1 in H; try discriminate.
      case_eq (WCWH.get_hole nat w2); intros;
        rewrite H2 in H0; try discriminate.
      injection H0; injection H.
      clear H0 H; intros.
      subst w.
      rewrite <- (WCWH.set_get_hole_same nat w1).      
      rewrite <- (WCWH.set_get_hole_same nat w2).
      rewrite H1.
      rewrite H2.
      assert (n = n0).
      eapply WCWH.set_hole_injective.
      symmetry; apply H0.
      subst n0.
      apply (WCWH.set_hole_eq nat nat n); auto.
     Qed.

  Lemma tick_upd_commute : forall w1 w2 x a,
    (exists w', tick_world w1 = Some w' /\ w_upd x a w' = w2) <->
    (tick_world (w_upd x a w1) = Some w2).

    intros; split.

    (* --> direction *)
    intros [w' [? ?]].
    unfold tick_world, world_ttl, w_upd.
    rewrite WCWH.get_hole_upd_collapse.
    unfold tick_world, world_ttl in H.
    case_eq (WCWH.get_hole nat w1); intros;
      rewrite H1 in H; try discriminate.
    unfold w_upd in H0.
    subst w2.
    injection H; clear H; intro H.
    subst w'.
    rewrite WCWH.set_hole_commute; auto.
 
    (* <-- direction *)
    unfold tick_world, w_upd, world_ttl.
    intros.
    rewrite WCWH.get_hole_upd_collapse in H.
    case_eq (WCWH.get_hole nat w1); intros;
      rewrite H0 in H; try discriminate.
    exists (WCWH.set_hole _ _ n w1); split; auto.
    rewrite WCWH.set_hole_commute.
    congruence.
   Qed.

  End WT.
End SimpleWorlds.


Module WorldTheory (W':WORLDS) : WORLD_THEORY with Module W:=W'.
  Module W := W'.

  Export W'.

  Fixpoint tick_world_n (n:nat) (w:world) { struct n } :=
    match n with
    | 0 => Some w
    | S n' =>
        match tick_world w with
        | Some w' => tick_world_n n' w'
        | None    => None
        end
    end.

  Inductive upd : Type :=
  | U : forall x:field, field_type x -> upd.

  Definition subst := list upd.

  Fixpoint upd_set (z:subst) : list field :=
   match z with
   | nil => nil
   | U x _ :: z' => x :: upd_set z'
   end.

  Fixpoint world_update (z:subst) : world -> world :=
   match z with
   | nil         => fun w => w
   | U x a :: z' => fun w => w_upd x a (world_update z' w)
   end.

  Lemma field_eq_dec : forall x y:field, {x=y}+{x<>y}.
    intros.
    destruct (eq_nat_dec (field_enum x) (field_enum y)).
    left; apply field_enum_injective; auto.
    right; red; intros; apply n; subst; auto.
   Defined.

  Lemma world_ttl_noninterference : forall w x a,
    world_ttl w = world_ttl (w_upd x a w).

    intros.
    set (n := world_ttl w).
    unfold n.
    generalize (refl_equal n); unfold n at 2; clearbody n.
    revert w x a; induction n; simpl; intros.
    case_eq (tick_world w); intros.
    rewrite (tick_ttl w w0) in H; auto.
    discriminate.
    case_eq (tick_world  (w_upd x a w)); intros.
    destruct (tick_upd_commute w w0 x a).
    destruct (H3 H1); clear H2 H3.
    rewrite H0 in H4; intuition; discriminate.
    rewrite <- H.    
    case_eq (world_ttl (w_upd x a w)); intros; auto.
    destruct (tick_succeeds (w_upd x a w)).
    rewrite H2; auto with arith.
    rewrite H1 in H3; discriminate.
    destruct (tick_succeeds w).
    rewrite <- H; auto with arith.
    rewrite (tick_ttl w x0); auto.
    destruct (tick_upd_commute w (w_upd x a x0) x a).
    generalize (H1 (@ex_intro _ _ _ (conj H0 (refl_equal _)))).
    clear H1 H2; intros.
    change w_upd with WC.w_upd.
    rewrite (tick_ttl _ _ H1).
    rewrite (IHn x0 x a); auto.
    cut (S n = S (world_ttl x0)); auto with arith.
    rewrite <- (tick_ttl w x0); auto.
   Qed.

  Lemma w_upd_noninterference : forall w w' x,
    tick_world w = Some w' ->
    w_proj x w = w_proj x w'.

    intros.
    destruct (tick_upd_commute w w' x (w_proj x w')).
    clear H1.
    assert (tick_world (w_upd x (w_proj x w') w) = Some w').
    apply H0.
    exists w'; intuition.
    change w_proj with WC.w_proj.
    apply upd_proj_same.
    clear H0.    
    replace w with (w_upd x (w_proj x w') w).
    rewrite proj_upd_same; auto.
    eapply tick_inj.
    apply H1.
    apply H.
   Qed.

  Lemma tick_world_n_lt : forall n w w',
     tick_world_n (S n) w = Some w' ->
     world_ttl w' < world_ttl w.

     intro n; induction n; simpl; intros.
     case_eq (tick_world w); intros.
     rewrite H0 in H.
     replace w' with w0; [ | congruence ].
     rewrite (tick_ttl w w0); auto.
     rewrite H0 in H; discriminate.

     case_eq (tick_world w); intros.
     rewrite H0 in H.
     apply lt_trans with (world_ttl w0).
     apply IHn.
     simpl; auto.
     rewrite (tick_ttl w w0); auto.
     rewrite H0 in H; discriminate.
    Qed.

  Lemma tick_congruence1 : forall n w1 w1' w2 w2',
     tick_world_n n w1 = Some w1' ->
     tick_world_n n w2 = Some w2' ->
     world_ttl w1 = world_ttl w2 ->
     world_ttl w1' = world_ttl w2'.

     intro n; induction n; simpl; intros.
     congruence.
     
     case_eq (tick_world w1); intros.
     case_eq (tick_world w2); intros.
     rewrite H2 in H.
     rewrite H3 in H0.
     eapply IHn.
     apply H.
     apply H0.    
     cut (S (world_ttl w) = S (world_ttl w0)).
     congruence.
     rewrite <- tick_ttl with w1 w; auto.
     rewrite <- tick_ttl with w2 w0; auto.
     rewrite H3 in H0.
     discriminate.
     rewrite H2 in H.
     discriminate.
    Qed.

  Lemma tick_congruence2 : forall n w1 w1' w2 w2',
     tick_world_n n w1 = Some w1' ->
     tick_world_n n w2 = Some w2' ->
     world_ttl w1' = world_ttl w2' ->
     world_ttl w1 = world_ttl w2.

     intro n; induction n; simpl; intros.
     congruence.
     case_eq (tick_world w1); intros.
     case_eq (tick_world w2); intros.
     rewrite H2 in H.
     rewrite H3 in H0.
     cut (world_ttl w = world_ttl w0).
     intros.
     rewrite (tick_ttl w1 w); auto.
     rewrite (tick_ttl w2 w0); auto.
     eapply IHn.
     apply H; auto.
     apply H0; auto.
     auto.
     rewrite H3 in H0; discriminate.
     rewrite H2 in H; discriminate.
    Qed.

  Lemma tick_n_ttl : forall n w w',
     tick_world_n n w = Some w' ->
     world_ttl w = world_ttl w' + n.

     intro n; induction n; simpl; intros.
     replace w with w'; [ auto | congruence ].
     case_eq (tick_world w); intros.
     rewrite H0 in H.
     rewrite (tick_ttl w w0); auto.
     cut (world_ttl w0 = world_ttl w' + n).
     omega.
     apply IHn; auto.
     rewrite H0 in H; discriminate.
    Qed.

  Lemma untick_n : forall n w', exists w, tick_world_n n w = Some w'.
     intro n; induction n; simpl; intros.
     exists w'; auto.
     destruct (IHn w').
     destruct (untick x).
     exists x0.
     rewrite H0.
     auto.
    Qed.

  Lemma tick_n_succeeds : forall n w,
     world_ttl w >= n ->
     exists w', tick_world_n n w = Some w'.

     intro n; induction n; intros.
     exists w; auto.
     case_eq (tick_world w); intros.
     cut (world_ttl w0 >= n).
     intros.
     destruct (IHn w0 H1) as [z Hz].
     exists z; simpl.
     rewrite H0; auto.
     rewrite (tick_ttl w w0) in H; auto.
     omega.
     destruct (tick_succeeds w).
     omega.
     rewrite H0 in H1.
     discriminate.
    Qed.


 Lemma tick_n_inj : forall n w1 w2 w,
     tick_world_n n w1 = Some w -> tick_world_n n w2 = Some w -> w1 = w2.

   induction n; intros.
   simpl in *; congruence.
   simpl in *.
   case_eq (tick_world w1); intros;
     rewrite H1 in H; try discriminate.
   case_eq (tick_world w2); intros;
     rewrite H2 in H0; try discriminate.
   cut (w0 = w3).
   intros.
   rewrite H3 in H1.
   apply tick_inj with w3; auto.
   apply IHn with w; auto.
  Qed.

  Lemma tick_n_split : forall n n1 n2 w1 w2,
    n = n1 + n2 ->
    tick_world_n n w1 = Some w2 ->
    exists w',
      tick_world_n n1 w1 = Some w' /\
      tick_world_n n2 w' = Some w2.

    intro n; induction n; simpl; intros.
    replace n1 with 0; [ | omega ].
    replace n2 with 0; [ | omega ].
    exists w1; simpl; auto.
    case_eq (tick_world w1); intros; rewrite H1 in H0; try discriminate.
    destruct n1.
    exists w1; intuition.
    replace n2 with (S n).
    simpl; rewrite H1; auto.
    destruct (IHn n1 n2 w w2); auto.
    exists x; intuition.
    simpl; rewrite H1; auto.
   Qed.

  Lemma upd_set_app : forall z1 z2,
    upd_set (z1++z2) = upd_set z1 ++ upd_set z2.

   intro z1; induction z1; simpl; auto.
   destruct a as [x a]; simpl; intros.
   rewrite IHz1; auto.
  Qed.
  
  Lemma world_update_app : forall w z1 z2,
    world_update (z1++z2) w = world_update z1 (world_update z2 w).

   intros w z1; induction z1; simpl; intros; auto.
   destruct a; simpl; rewrite IHz1; auto.
  Qed.
  
  Fixpoint insert_update (x:field) (a:field_type x) (z:subst) { struct z } : subst :=
  match z with
    | nil         => U x a :: nil 
    | U y b :: l' => if (le_lt_dec (field_enum y) (field_enum x))
      then if W.WC.field_eq_dec x y
        then U x a :: l'
        else U y b :: (insert_update x a l')
      else U x a :: U y b :: l'
  end.

  Fixpoint normalize_update (z:subst) : subst :=
    match z with
    | nil         => nil
    | U x a :: z' => insert_update x a (normalize_update z')
    end.

  Lemma change_world_insert : forall w l x a,
       world_update (U x a :: l) w =
       world_update (insert_update x a l) w.

    intros w l; induction l; simpl; intros; auto.
    destruct a as [y b]; simpl.
    destruct (le_lt_dec (field_enum y) (field_enum x)); simpl; auto.
    destruct (W.WC.field_eq_dec x y); subst; simpl.
    rewrite upd_collapse; auto.
    rewrite <- IHl; simpl.
    apply upd_commute; auto.
   Qed.

  Lemma world_update_normalize : forall w l,
     world_update l w = world_update (normalize_update l) w.

     intros w l; induction l; simpl; auto.
     destruct a.
     rewrite <- change_world_insert; simpl.
     rewrite <- IHl.
     auto.
    Qed.

  Definition w_upd_norm := fun l w => world_update (normalize_update l) w.

  Lemma world_update_normalize_rewrite : forall w l,
     world_update l w = w_upd_norm l w.

     apply world_update_normalize.
    Qed.

  Lemma proj_subst_notin : forall x z l w,
    incl (upd_set z) l ->
    ~In x l ->
    w_proj x (world_update z w) =
    w_proj x w.
  
    intros x z; induction z; simpl; intros.
    auto.
    destruct a; simpl.
    destruct (field_eq_dec x x0).
    contradiction H0.
    apply H.
    subst x0; simpl; auto.
    rewrite proj_upd_diff; auto.
    apply IHz with l; auto.
    red; intros.
    apply H; simpl; auto.
   Qed.

  Lemma commute_single : forall z1 z2 w
    (x:field) (a:field_type x),
    (~In x (upd_set z1)) ->
    world_update (U x a :: z1++z2) w = 
    world_update (z1 ++ (U x a):: z2) w.

    intro z1; induction z1; intros; auto.
    destruct a as [y b]; simpl.
    rewrite <- IHz1; simpl.
    rewrite upd_commute; auto.
    red; intro; subst; apply H.
    simpl; auto.
    red; intro; apply H.
    simpl; auto.
   Qed.  


  Lemma commute_subst : forall z1 z2 w,
    (forall x, In x (upd_set z1) -> In x (upd_set z2) -> False) ->
    world_update (z1++z2) w = world_update (z2++z1) w.
  
    intro z1; induction z1; intros.
    simpl; rewrite <- app_nil_end; auto.
    destruct a as [x a].
    simpl.
    rewrite <- commute_single; simpl.
    rewrite IHz1; auto.
    intros.  
    apply H with x0; auto.
    simpl; right; auto.
    red; intro; apply H with x; simpl; auto.
   Qed.

  Fixpoint drop_sel (x:field) (z:subst) {struct z}: subst :=
    match z with
    | nil => nil
    | U y a :: z' =>
       if W.WC.field_eq_dec x y
         then drop_sel x z'
         else U y a :: drop_sel x z'
    end.

  Lemma drop_sel_notin : forall x z,
    ~In x (upd_set (drop_sel x z)).

    intros x z; induction z; simpl; intros; auto.
    destruct a as [y b]; simpl.
    elim (W.WC.field_eq_dec x y); intro Hxy; simpl; auto.
    red; intros; destruct H.
    contradiction Hxy; auto.
    apply IHz; auto.
   Qed.  

  Lemma subst_drop_app : forall x z1 z2,
    drop_sel x (z1++z2) = drop_sel x z1 ++ drop_sel x z2.

    intros x z1; induction z1; simpl; intros; auto.
    destruct a as [y b]; simpl.
    elim (W.WC.field_eq_dec x y); intros; rewrite IHz1; auto.
   Qed.

  Lemma subst_drop_head : forall x a z w,
    world_update (U x a :: z) w =
    world_update (U x a :: drop_sel x z) w.

    intros; induction z; simpl; auto.
    destruct a0 as [y b].
    elim (W.WC.field_eq_dec x y); intro Hxy.
    subst x.
    rewrite upd_collapse; auto.
    simpl in *.
    rewrite upd_commute; auto.
    rewrite IHz.
    rewrite upd_commute; auto.
   Qed.

  Fixpoint float_sel' (x:field) (head:subst) (z:subst) {struct z} : subst :=
    match z with
    | nil => head
    | U y a :: z' =>
       if W.WC.field_eq_dec x y
          then U y a :: head ++ (drop_sel x z')
          else float_sel' x (head ++ (U y a::nil)) z'
    end.

  Definition float_sel (x:field) := float_sel' x nil.

  Lemma float_head_out : forall x y a head z w,
     x <> y -> 
     world_update (float_sel' x (U y a :: head) z) w =
     world_update (U y a :: float_sel' x head z) w.

    intros; revert x y a head H; induction z; simpl; intros; auto.
    destruct a as [v a].
    destruct (W.WC.field_eq_dec x v).
    subst v.
    simpl.
    rewrite upd_commute; auto.
    rewrite IHz; auto.
   Qed.  

  Lemma float_sel_eq : forall w z x,
    world_update z w = world_update (float_sel x z) w.

    intros w z x; induction z.
    unfold float_sel; simpl; auto.
    destruct a as [y a].
    unfold float_sel; simpl.
    destruct (W.WC.field_eq_dec x y); simpl; auto.
    change (w_upd y a (world_update (drop_sel x z) w))
       with (world_update (U y a :: drop_sel x z) w).
    subst y; rewrite <- (subst_drop_head x a z w).
    simpl; auto.
    rewrite float_head_out; auto.
    simpl; rewrite IHz; auto.
   Qed.   

  Lemma float_sel'_notin : forall z2 z1 z' x a,
    ~In x (upd_set z1) -> 
    float_sel' x z1 z2 = U x a :: z' ->
    ~In x (upd_set z').

    intro z2; induction z2; simpl; intros; auto.
    subst z1.
    red; intros; apply H; simpl; auto.

    destruct a as [y b].
    destruct (W.WC.field_eq_dec x y) as [Hxy|Hxy].
    subst y.
    injection H0; intros.
    subst z'.
    rewrite upd_set_app.
    red; intro.
    elim (in_app_or _ _ _ H1); auto.
    apply drop_sel_notin.
  
    red; intros.
    refine (IHz2 (z1 ++ U y b :: nil) z' x a0 _ _ _); auto.
    rewrite upd_set_app; red; intros.
    elim (in_app_or _ _ _ H2); auto.
    simpl; intuition.
   Qed.
  
  Lemma float_sel_notin : forall z z' x a,
    float_sel x z = U x a :: z' ->
    ~In x (upd_set z').

    unfold float_sel; intros.
    apply float_sel'_notin with z (nil:subst) a; auto.
   Qed.

  Fixpoint filter_subst (l:list field) (z:subst) { struct z } : subst :=
   match z with
   | nil => nil
   | U x a :: z' =>
     if In_dec W.WC.field_eq_dec x l
        then filter_subst l z'
        else U x a :: filter_subst l z'
   end.

  Lemma filter_subst_notin : forall x l z,
    In x l -> ~In x (upd_set (filter_subst l z)).
  
    intros x l z; induction z; simpl; intros; auto.
    destruct a as [y b].
    elim (In_dec W.WC.field_eq_dec y l); intros Hyl.
    apply IHz; auto.
    simpl; red; intros.
    destruct H0.
    subst y; contradiction.
    apply IHz; auto.
   Qed.  

  Lemma filter_subst_empty : forall l z,
    incl (upd_set z) l ->
    filter_subst l z = nil.

    intros l z; induction z; simpl; intros; auto.
    destruct a as [y b].
    elim (In_dec W.WC.field_eq_dec y l); intro Hyl.
    apply IHz.
    red; intros.
    apply H; simpl; auto.
    elimtype False; apply Hyl.
    red in H; apply H; simpl; auto.
   Qed.  

  Lemma proj_update_filter : forall x z l w,
     ~ In x l ->
     w_proj x (world_update z w) =
     w_proj x (world_update (filter_subst l z) w).

    intros x z; induction z; simpl; intros; auto.
    destruct a as [y b].
    elim (In_dec W.WC.field_eq_dec y l); intro Hyl.
    destruct (W.WC.field_eq_dec y x).
    elimtype False; subst y; contradiction.
    rewrite proj_upd_diff; auto.
    simpl.
    elim (W.WC.field_eq_dec x y); intros Hxy.
    subst y.
    repeat rewrite proj_upd_same; auto.
    repeat rewrite proj_upd_diff; auto.
   Qed.

  Lemma world_update_drop : forall z1 z2 w,
    incl (upd_set z2) (upd_set z1) ->
    world_update z1 (world_update z2 w) = world_update z1 w.

    induction z1.
    simpl; intros.
    red in H.
    destruct z2; simpl in *; auto.
    destruct u as [x a].
    elimtype False; apply H with x; simpl; auto.
    destruct a as [x a].  
    intros.
    rewrite <- world_update_app.
    rewrite (subst_drop_head x a (z1++z2)).
    rewrite subst_drop_app.
    simpl.
    rewrite world_update_app.
    change (world_update (U x a :: drop_sel x z1)
        (world_update (drop_sel x z2) w) = w_upd x a (world_update z1 w)).
    rewrite <- (subst_drop_head x a z1).
    simpl.
    rewrite IHz1; auto.
    unfold incl in *.
    clear IHz1.
    simpl in H.
    revert z1 H; clear.
    induction z2; simpl; intros.
    contradiction.
    destruct a as [y b].
    simpl in *.
    destruct (W.WC.field_eq_dec x y).
    apply IHz2; auto.
    simpl in *.
    destruct H0.
    subst a0.
    elim (H y); intros; intuition.
    apply IHz2; auto.
   Qed.  
  
  Lemma world_update_dup : forall z w,
    world_update z w = world_update z (world_update z w).

    intros; symmetry; apply world_update_drop.
    red; intros; auto.
   Qed.

End WorldTheory.
