(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Require Import Bool.

Require Export Maps.
Require Export Machine.

Module SimpleTC.

  Inductive typ : Set :=
  | ty_nil      : typ
  | ty_list     : typ -> typ
  | ty_listcons : typ -> typ.

  Definition env := map typ.
  Definition program_typing := map env.

  Fixpoint check_subtype (t1 t2: typ) {struct t1} : bool :=
    match t1, t2 with
      | ty_nil, ty_nil => true
      | ty_nil, ty_list t1 => true
      | ty_nil, ty_listcons t1 => false
      | ty_list s1, ty_nil => false
      | ty_list s1, ty_list t1 => check_subtype s1 t1
      | ty_list s1, ty_listcons t1 => false
      | ty_listcons s1, ty_nil => false
      | ty_listcons s1, ty_list t1 => check_subtype s1 t1
      | ty_listcons s1, ty_listcons t1 => check_subtype s1 t1
    end.

  Fixpoint lub (s t: typ) {struct s}: typ :=
    match s, t with
      | ty_nil, ty_nil => ty_nil
      | ty_nil, ty_list t1 => ty_list t1
      | ty_nil, ty_listcons t1 => ty_list t1
      | ty_list s1, ty_nil => ty_list s1
      | ty_list s1, ty_list t1 => ty_list (lub s1 t1)
      | ty_list s1, ty_listcons t1 => ty_list (lub s1 t1)
      | ty_listcons s1, ty_nil => ty_list s1
      | ty_listcons s1, ty_list t1 => ty_list (lub s1 t1)
      | ty_listcons s1, ty_listcons t1 => ty_listcons (lub s1 t1)
    end.

  Definition check_env_binding_sub (t: typ) (ot: option typ) :=
    match ot with
      | None => false
      | Some t' => check_subtype t' t
    end.

  Definition check_env_sub (e1 e2: env) : bool :=
    map_forall2 typ typ check_env_binding_sub e2 e1.

  Definition typecheck_branch (pt: program_typing) (vt: env) (l: label): bool :=
    match pt#l with
      | None => false
      | Some vt' => check_env_sub vt vt'
    end.

  Fixpoint typecheck_instr (pt: program_typing) (e: env) (i: instr)
    {struct i}: option env :=
    match i with
      | i1 ;; i2 =>
        match typecheck_instr pt e i1 with
          | None => None
          | Some e1 => typecheck_instr pt e1 i2
        end
      | instr_branch_if_nil v l =>
        match e#v with
          | None => None
          | Some (ty_list t) =>
            if typecheck_branch pt (e#v <- ty_nil) l
              then Some (e#v <- (ty_listcons t))
              else None
          | Some (ty_listcons t) =>
            Some e
          | Some ty_nil =>
            if typecheck_branch pt e l
              then Some e
              else None
        end
      | instr_fetch_field v1 0 v2 =>
        match e#v1 with
          | Some (ty_listcons t) => Some (e#v2 <- t)
          | _ => None
        end
      | instr_fetch_field v1 1 v2 =>
        match e#v1 with
          | Some (ty_listcons t) => Some (e#v2 <- (ty_list t))
          | _ => None
        end
      | instr_cons v1 v2 v3 =>
        match e#v1, e#v2 with
          | Some t1, Some t2 =>
            match lub (ty_list t1) t2 with
              | ty_list t3 => Some (e#v3 <- (ty_listcons t3))
              | _ => None
            end
          | _, _ => None
        end
      | instr_getlabel (1%positive) v => Some (e#v <- ty_nil)
      | _ =>
        None
    end.

  Fixpoint typecheck_block (pt: program_typing) (e: env) (i: instr)
    {struct i}: bool :=
    match i with
      | instr_halt => true
      | instr_getlabel l v ;; instr_jump v' =>
        andb
        (if var_eq v v' then true else false) 
        (typecheck_branch pt (unset typ e v) l)
      | i1 ;; i2 =>
        match typecheck_instr pt e i1 with
          | None => false
          | Some e1 => typecheck_block pt e1 i2
        end
      | _ => false
    end.

  Definition check_one_block (pt:program_typing) (i:instr) (oe:option env) :=
    match oe with
      | Some e => typecheck_block pt e i
      | None   => false
    end.

  Definition typecheck_blocks (p:program) (pt:program_typing) :=
    map_forall2 _ _ (check_one_block pt) p pt.

  Definition check_one_domain (e:env) (oi:option instr) :=
    match oi with
      | Some i => true
      | None   => false
    end.

  Definition check_pt_domain (p:program) (pt:program_typing) : bool :=
    map_forall2 _ _ check_one_domain pt p.

  Definition typecheck (p:program) (pt:program_typing) : bool :=
    (typecheck_blocks p pt &&
      check_pt_domain p pt &&
      match (pt#entry_label) with
        | Some e => is_empty _ e
        | _ => false
      end) % bool.

End SimpleTC.