(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import Bool.
Require Import CompleteInduction.
Require Import List.

Require Import simple_tc_safety.
Require Import Maps.
Require Import ModelSpec.
Require Import LMModel.
Require Import Listmachine.
Require Import LMHoare.

Module SimpleTCSemantic (LMH:LM_HOARE_RULES) : SIMPLE_TC_SAFETY.
  Import LMH.LM.MM.
  Import LMWorldCarrier.
  Import LMStep.
  Module LM := LMH.LM.

  Open Local Scope pred_scope.

  Definition ty := pred.
  Definition t_nil := LM.t_nil.
  Definition t_list := LM.t_list.
  Definition t_listcons := LM.t_listcons.

  Definition env := map ty.
  Definition program_typing := map env.

  Definition env_to_ty (e:env) :=
    All_ v:var, All_ t:ty,
      const (e#v = Some t) --> LM.t_slot v ([](LM.valuepred t)).

  Lemma env_to_ty_nec : forall e,
     [](env_to_ty e) = env_to_ty e.
  
     intros; apply derives_equal.
     elim_at 0; auto.
     apply prove_adj with necessary.
     unfold env_to_ty.
     apply_d adjective_forall; intro_d; intro.
     apply_d adjective_forall; intro_d; intro.
     apply_d adjective_impl_const; auto.
     apply_d LM.slot_necessary.
     apply_d adjective_modal; auto.
    Qed.

  Lemma env_to_ty_eqn : forall e x t,
     e#x = Some t ->
     |-- env_to_ty e >=> LM.t_slot x ([](LM.valuepred t)).

     intros; unfold env_to_ty.
     unfold sub; repeat apply modal_N; auto.
     intro_d.
     spec_at 0 x.
     spec_at 0 t.
     auto.
    Qed.

  Lemma env_to_ty_sub : forall e z,
    (forall x t, e#x = Some t -> |-- z >=> LM.t_slot x ([](LM.valuepred t))) ->
      |-- z >=> env_to_ty e.

    intros.
    unfold sub; repeat apply modal_N; auto.
    unfold env_to_ty; intuition_d.
    apply_d sub_project_herenow; [ | exact_at 0 ].
    apply_d H.
   Qed.

  Definition subtype (t1 t2:ty) :=
    |-- t1 >=> t2.

  Lemma sub_refl : forall t, subtype t t.
    unfold subtype; intros; apply sub_refl.
   Qed.

  Definition sub_trans' := sub_trans.

  Lemma sub_trans : forall t1 t2 t3,
     subtype t1 t2 ->
     subtype t2 t3 ->
     subtype t1 t3.

    unfold subtype; intros.
    apply_d sub_trans; [ apply H | apply H0 ].
   Qed.

  Lemma sub_nil  : forall t,
     subtype LM.t_nil (LM.t_list t).

    unfold subtype; simpl; intros.
    apply LM.subtype_nil.
   Qed.

  Lemma sub_list : forall t1 t2,
     subtype t1 t2 ->
     subtype (LM.t_list t1) (LM.t_list t2).

    unfold subtype; simpl; intros.
    apply_d LM.subtype_list; auto.
   Qed.

  Lemma sub_listmixed : forall t1 t2,
     subtype t1 t2 ->
     subtype (LM.t_listcons t1) (LM.t_list t2).

    unfold subtype; simpl; intros.
    apply_d LM.subtype_listmixed; auto.
   Qed.

  Lemma sub_listcons : forall t1 t2,
     subtype t1 t2 ->
     subtype (LM.t_listcons t1) (LM.t_listcons t2).
  
    unfold subtype; simpl; intros.
    apply_d LM.subtype_listcons; auto.
   Qed.

  Hint Resolve sub_refl sub_trans sub_nil
     sub_list sub_listmixed sub_listcons
     : typing.

  Definition env_sub (e e':env) :=
    forall v t', e'#v = Some t' ->
       exists t, e#v = Some t /\ subtype t t'.

  Lemma env_sub_interp_env : forall (e e':env),
     env_sub e e' ->
     |-- env_to_ty e >=> env_to_ty e'.

    intros; apply env_to_ty_sub; intros.
    destruct (H x t H0) as [t' ?]; intuition.
    red in H3.
    apply_d sub_trans'.
    instantiate (1 := LM.t_slot x ([](LM.valuepred t'))).
    apply env_to_ty_eqn; auto.
    apply_d LM.sub_slot; auto.
    apply_d sub_modal.
    apply_d sub_modal.
   Qed.

  Definition lm_pt (pt:program_typing) :=
      (map_fmap _ _ env_to_ty pt).

  Definition check_instr (p:program) (pt:program_typing) (e:env) (i:instr) (e':env) :=
    LM.check_instr p (lm_pt pt) (env_to_ty e) i (env_to_ty e').

  Definition check_block (p:program) (pt:program_typing)
    (e:env) (i:instr) := LM.check_block p (lm_pt pt) (env_to_ty e) i.

  Lemma env_to_ty_unset : forall e v,
    |-- env_to_ty e >=> env_to_ty (unset _ e v).

    unfold env_to_ty; intros.
    unfold sub; repeat apply modal_N; auto.

    intuition_d.
    spec_at 0 x.
    spec_at 0 x0.
    apply_at 0.
    apply const_intro.
    destruct (var_eq v x).
    subst x.
    rewrite unset_same in H; discriminate.
    rewrite unset_other in H; auto.
   Qed.

  Lemma unset_not_in_dom : forall e v,
    LM.not_in_dom v (env_to_ty (unset ty e v)).

    red; intros.
    unfold LM.with_var_upd.
    apply (with_projecting Sel_S); intro s.
    apply (field_pred_intro _ s); auto.
    in_goal unfold env_to_ty.
    rewrite pmodal_forall; auto.
    intro_d; intro x.
    rewrite pmodal_forall; auto.
    intro_d; intro t.
    unfold env_to_ty.
    spec_at 1 x.
    spec_at 0 t.
    rewrite pmodal_const_impl; auto.
    intuition_d.
    elim_at 0.
    apply const_intro; auto.
    unfold LM.t_slot.
    apply (field_pred_elim _ s); auto.
    case_eq (s#x); intros; auto.
    rewrite with_subst_field_pred.
    rewrite get_set_other.
    rewrite H0.
    rewrite with_subst_commute.
    2: simpl; intuition; subst; discriminate.
    apply derives_weaken_head.
    apply derives_modal.
    unfold LM.valuepred.
    rewrite <- respecting_nec_commute.
    apply_d (respecting_with_subst (Sel_V::nil)); auto.
    simpl; intuition; subst; discriminate.
    destruct (var_eq v x); auto.
    subst x.
    rewrite unset_same in H.
    discriminate.
   Qed.

  Lemma check_instr_getnil : forall p pt e v,
   (* -------------------------------------------------- *)
      check_instr p pt e (instr_getlabel (L 0) v) (e#v <- t_nil).

    unfold check_instr; intros.
    apply LMH.check_instr_weaken with
      (env_to_ty (unset ty e v))
      (env_to_ty (unset ty e v) && LM.t_slot v t_nil).
    apply_d env_to_ty_unset.
    apply_d env_to_ty_sub.
    intros.
    destruct (var_eq v x); subst.
    rewrite get_set_same in H.
    replace t with t_nil; [ | congruence ].
    unfold t_nil.
    apply sub_inter_left2.
    rewrite LM.nil_valuepred.
    pattern LM.t_nil at 1.
    rewrite <- LM.nil_nec.
    auto.
    apply sub_inter_left1.
    apply_d env_to_ty_eqn.
    rewrite unset_other; auto.
    rewrite get_set_other in H; auto.
    apply LMH.check_instr_getnil.
    rewrite env_to_ty_nec.
    refine (unset_not_in_dom _ _).
   Qed.

  Lemma check_instr_seq : forall p pt e e' e'' i1 i2,
    check_instr p pt e  i1 e' ->
    check_instr p pt e' i2 e'' ->
    check_instr p pt e  (i1;;i2) e''.

    unfold check_instr; intros.
    eapply LMH.check_instr_seq.
    apply H; auto.
    apply H0; auto.
   Qed.   

  Lemma check_instr_branch_list : forall p pt e e1 v l t,
    e#v = Some (t_list t) ->
    pt#l = Some e1 ->
    env_sub (e#v <- t_nil) e1 ->
    check_instr p pt
       e (instr_branch_if_nil v l) (e#v <- (t_listcons t)).

    unfold check_instr; intros.
    apply LMH.check_instr_weaken with
      (env_to_ty e && LM.t_slot v (t_nil || LM.t_anypair))
      (env_to_ty e && LM.t_slot v LM.t_anypair).

    apply sub_inter_right.
    apply_d sub_refl.
    apply_d sub_trans'.
    apply_d env_to_ty_eqn.
    apply H.
    rewrite LM.list_valuepred.
    rewrite LM.list_nec.
    apply_d LM.sub_slot.
    apply derives_axiom.
    apply LM.sub_list_nil_anypair.
    apply_d env_to_ty_sub; intros.
    destruct (var_eq v x).
    subst x.
    rewrite get_set_same in H2.
    replace t0 with (t_listcons t); [ | congruence ].
    simpl.
    apply_d sub_trans'.
    instantiate (1 := LM.t_slot v (LM.t_list t) && LM.t_slot v LM.t_anypair).
    apply sub_inter_right.
    apply sub_inter_left1.
    rewrite <- LM.list_nec.
    rewrite <- LM.list_valuepred.
    apply_d env_to_ty_eqn; auto.
    apply sub_inter_left2; auto.
    rewrite LM.slot_intersection.
    apply_d LM.sub_slot.
    rewrite LM.listcons_valuepred.
    rewrite LM.listcons_nec.
    apply LM.sub_list_anypair_listcons.
    rewrite get_set_other in H2; auto.
    apply sub_inter_left1.
    refine (env_to_ty_eqn _ _ _ _); auto.

    apply LMH.check_instr_branch with (env_to_ty e1).
    unfold lm_pt; apply fmap_eqn; auto.
    unfold sub; do 2 (apply modal_N; auto).
    unfold env_to_ty; intuition_d.
    red in H1.
    destruct (H1 x x0); auto.
    destruct (var_eq x v); subst; intuition.
    clear_at 0.
    rewrite get_set_same in H4.
    replace x1 with t_nil in H5; [ | congruence ].
    red in H5.
    change LMH.LM.t_nil with LM.t_nil.
    rewrite <- LM.nil_nec.
    rewrite <- LM.nil_valuepred.

    apply_d sub_project_herenow; [ | exact_at 0 ].
    clear_at 0.
    apply_d LM.sub_slot.
    apply sub_modal.
    apply sub_modal.
    auto.

    rewrite get_set_other in H4; auto.
    spec_at 0 x.
    spec_at 0 x1.
    elim_at 0; auto.
    apply_d sub_project_herenow; [ | exact_at 0 ].
    apply_d LM.sub_slot.
    apply_d sub_modal.
    apply_d sub_modal.
   Qed.

  Lemma check_instr_branch_listcons : forall p pt e l v t,
    e#v = Some (t_listcons t) ->
    check_instr p pt
       e (instr_branch_if_nil v l) e.

    red; intros.
    apply LMH.check_instr_weaken with
      (env_to_ty e && LM.t_slot v LM.t_anypair)
      (env_to_ty e).
    apply sub_inter_right.
    apply_d sub_refl.
    apply_d sub_trans'.
    instantiate (1 := LM.t_slot v (t_listcons t)).
    rewrite <- LM.listcons_nec.
    rewrite <- LM.listcons_valuepred.
    apply_d env_to_ty_eqn; auto.
    apply_d LM.sub_slot.
    apply_d LM.sub_listcons_anypair.
    apply_d sub_refl.
    apply LMH.check_instr_branch_pair.
   Qed.

  Lemma check_instr_branch_nil : forall p pt e v l e1 e2,
    e#v = Some t_nil ->
    pt#l = Some e1 ->
    env_sub e e1 ->
    check_instr p pt
        e (instr_branch_if_nil v l) e2.

    red; intros.
    apply LMH.check_instr_weaken with
      (env_to_ty e && LM.t_slot v t_nil) (env_to_ty e2).
    apply sub_inter_right.
    apply_d sub_refl.
    unfold t_nil.
    rewrite <- LM.nil_nec.
    rewrite <- LM.nil_valuepred.
    apply_d env_to_ty_eqn.
    apply_d sub_refl.
    apply LMH.check_instr_branch_nil with (env_to_ty e1).
    unfold lm_pt; apply fmap_eqn; auto.
    apply env_sub_interp_env; auto.
   Qed.

  Lemma check_instr_fetch_0: forall p pt e v1 v2 t,
    e#v1 = Some (t_listcons t) ->
    check_instr p pt
        e (instr_fetch_field v1 0 v2) (e#v2 <- t).

    red; intros.
    apply LMH.check_instr_weaken with
      (env_to_ty (unset _ e v2) && LM.t_slot v1 (LM.t_pair (LM.valuepred ([]t)) (LM.t_list t)))
      (env_to_ty (unset _ e v2) && LM.t_slot v2 (LM.valuepred ([]t))).

    apply sub_inter_right.
    apply_d env_to_ty_sub; intros.
    destruct (var_eq v2 x).
    subst x.
    rewrite unset_same in H0.
    discriminate.
    rewrite unset_other in H0; auto.
    refine (env_to_ty_eqn _ _ _ _); auto.
    apply_d sub_trans'.
    instantiate (1 := LM.t_slot v1 ([](LM.valuepred (LM.t_listcons t)))).
    apply_d env_to_ty_eqn.
    rewrite LM.listcons_valuepred.
    rewrite LM.listcons_nec.
    rewrite <- LM.list_valuepred.
    rewrite LM.pair_eat_valuepred.
    rewrite <- LM.list_nec.
    rewrite LM.pair_eat_nec.
    apply_d sub_refl.

    apply_d env_to_ty_sub; intros.
    destruct (var_eq v2 x).
    subst x.
    rewrite get_set_same in H0.
    replace t0 with t; [ | congruence ].
    apply sub_inter_left2.

    unfold LM.valuepred.
    rewrite respecting_nec_commute.
    apply sub_refl.
    rewrite get_set_other in H0; auto.
    apply sub_inter_left1.
    refine (env_to_ty_eqn _ _ _ _).
    rewrite unset_other; auto.
    
    apply LMH.check_instr_fetch_0.
    apply unset_not_in_dom.
  Qed.

  Lemma check_instr_fetch_1: forall p pt e v1 v2 t,
    e#v1 = Some (t_listcons t) ->
    check_instr p pt
        e (instr_fetch_field v1 1 v2) (e#v2 <- (t_list t)).

    red; intros.
    apply LMH.check_instr_weaken with
      (env_to_ty (unset _ e v2) && LM.t_slot v1 (LM.t_pair t (LM.t_list t)))
      (env_to_ty (unset _ e v2) && LM.t_slot v2 (LM.t_list t)).

    apply sub_inter_right.
    apply derives_axiom.
    apply env_to_ty_sub; intros.
    destruct (var_eq v2 x).
    subst x.
    rewrite unset_same in H0.
    discriminate.
    rewrite unset_other in H0; auto.
    apply env_to_ty_eqn; auto.
    apply_d sub_trans'.
    instantiate (1 := LM.t_slot v1 (LM.t_listcons t)).
    rewrite <- LM.listcons_nec.
    rewrite <- LM.listcons_valuepred.
    apply derives_axiom.
    refine (env_to_ty_eqn _ _ _ _).
    auto.
    apply_d sub_refl.
    apply_d env_to_ty_sub.
    intros.
    destruct (var_eq v2 x).
    subst x.
    rewrite get_set_same in H0.
    replace t0 with (t_list t); [ | congruence ].
    apply sub_inter_left2.
    rewrite LM.list_valuepred.
    rewrite LM.list_nec.
    apply sub_refl.
    rewrite get_set_other in H0; auto.
    apply sub_inter_left1.
    refine (env_to_ty_eqn _ _ _ _).
    rewrite unset_other; auto.
    
    apply LMH.check_instr_fetch_1.
    apply unset_not_in_dom.
  Qed.

  Lemma check_instr_cons: forall p pt e v1 v2 v3 t1 t2 t3,
    e#v1 = Some t1 ->
    e#v2 = Some t2 ->
    subtype t1 t3 ->
    subtype t2 (t_list t3) ->
    check_instr p pt
       e (instr_cons v1 v2 v3) (e#v3 <- (t_listcons t3)).

    red; intros.
    apply LMH.check_instr_weaken with
      (env_to_ty (unset _ e v3) && LM.t_slot v1 ([](LM.valuepred t3))
                                && LM.t_slot v2 ([](LM.valuepred (LM.t_list t3))))
      (env_to_ty (unset _ e v3) && LM.t_slot v3 (LM.t_pair t3 (LM.t_list t3))).

    apply derives_axiom.
    apply sub_inter_right.
    apply sub_inter_right.
    apply env_to_ty_unset.
    apply_d sub_trans'.
    instantiate (1 := LM.t_slot v1 ([](LM.valuepred t1))).
    apply env_to_ty_eqn; auto.
    apply_d LM.sub_slot.
    apply sub_modal.
    apply sub_modal.
    apply H1.
    apply_d sub_trans'.
    instantiate (1 := LM.t_slot v2 ([](LM.valuepred t2))).
    refine (env_to_ty_eqn _ _ _ _); auto.
    apply_d LM.sub_slot.
    apply sub_modal.
    apply sub_modal.
    apply H2.
    apply_d env_to_ty_sub; intros.
    destruct (var_eq v3 x).
    subst x.
    rewrite get_set_same in H3.
    replace t with (t_listcons t3); [ | congruence ].
    apply sub_inter_left2.
    apply_d LM.sub_slot.
    rewrite LM.listcons_valuepred.
    rewrite LM.listcons_nec.
    apply_d sub_refl.
    rewrite get_set_other in H3; auto.
    apply sub_inter_left1.
    refine (env_to_ty_eqn _ _ _ _).
    rewrite unset_other; auto.

    apply LMH.check_instr_cons.
    apply unset_not_in_dom.
  Qed.

  Lemma check_block_halt: forall p pt e,
    check_block p pt e instr_halt.

    red; intros; apply LMH.check_block_halt.
  Qed.

  Lemma check_block_seq: forall p pt e i1 i2 e',
    check_instr p pt e i1 e' ->
    check_block p pt e' i2 ->
    check_block p pt e (i1 ;; i2).

    red; intros; eapply LMH.check_block_seq.
    apply H.
    apply H0.
  Qed.

  Lemma check_block_jump: forall p pt e v l e1,
    pt#l = Some e1 ->
    env_sub (unset _ e v) e1 ->
    check_block p pt e (instr_getlabel l v ;; instr_jump v).

    red; intros; eapply LMH.check_block_seq;
      [ | apply LMH.check_block_jump ].
    instantiate (1 := env_to_ty e1).
    apply LMH.check_instr_weaken with
      (env_to_ty (unset _ e v))
      (env_to_ty (unset _ e v) && LM.t_slot v (LM.t_cont p (env_to_ty e1))).

    apply_d env_to_ty_unset.
    apply sub_inter_right.
    apply sub_inter_left1.
    apply_d (env_sub_interp_env _ _ H0).
    apply sub_inter_left2.
    apply_d sub_refl.
    
    apply LMH.check_instr_getlabel with (env_to_ty e1).
    unfold lm_pt; apply fmap_eqn; auto.
    apply sub_refl.
    rewrite env_to_ty_nec.
    apply unset_not_in_dom.
   Qed.

  Definition check_blocks (p:program) (pt:program_typing) :=
   forall l i,
     p#l = Some i ->
     exists e,
       pt#l = Some e /\
       check_block p pt e i.

  Definition typing_dom_match (p: program) (pt: program_typing) : Prop :=
    forall l e, pt#l = Some e -> exists i, p#l = Some i.

  Lemma lm_safety1 : forall (p:program) (pt: program_typing) i,
     check_blocks p pt ->
     typing_dom_match p pt ->
     pt#entry_label = Some (empty ty) ->
     p#entry_label = Some i ->
     |-- safe p i.

     intros.
     apply LM.program_safety with (lm_pt pt) (env_to_ty (empty _)).
     red; intros.
     destruct (H l i0); auto.
     exists (env_to_ty x); intuition.
     unfold lm_pt; apply fmap_eqn; auto.
     red; intros.
     destruct H3.
     case_eq (pt#l); intros.
     apply H0 with e; auto.
     elimtype False.
     unfold lm_pt in H3.
     rewrite fmap_none in H3; auto; discriminate.
     unfold lm_pt.
     apply fmap_eqn.
     auto.
     auto.     
     apply derives_equal; auto.
     unfold env_to_ty.
     intuition_d.
     destruct x; discriminate.
    Qed.

  Lemma safety : forall (p:program) (pt: program_typing),
     check_blocks p pt ->
     typing_dom_match p pt ->
     pt#entry_label = Some (empty ty) ->
     safe_prog p.

     intros.
     red; intros.
     red; intros.
     red in H2.
     assert (|-- safe p i).
     apply lm_safety1 with pt; auto.
     generalize (stepstar_stepN _ _ _ _ _ H3).
     intros [n Hn].
     set (w := World nat s (value_label (L 0)) n).
     set (w' := World nat s' (value_label (L 0)) n).
     destruct (H4 w I n i' w'); auto.
     subst w w'.
     set (n' := n).
     unfold n' at 1.
     generalize n'.
     revert Hn.
     clear.
     intro H; induction H; intros.
     apply stepN_O.
     apply stepN_S with i' (World nat s' (value_label (L 0)) n'); auto.
     hnf; simpl; intuition.
     hnf in H5; subst i'.
     apply step_or_halt_halt.
     destruct H5 as [k' [w'' H5]].
     hnf in H5; intuition.
     eapply step_or_halt_step.
     apply H6.
   Qed.

End SimpleTCSemantic.


Require Import Model.

Module MM : MODAL_MODEL 
    with Module W := LMSimpleWorlds
    with Module WS := LMStep
   := ModalModel LMSimpleWorlds LMStep.
Module LM  := ListMachine MM.
Module LMH := LMHoareRules LM.
Module SemSafe := SimpleTCSemantic LMH.
Module SemanticSoundness := SimpleTCSafetyTheorem SemSafe.

Check SemanticSoundness.typecheck_safeprog.
